Package: MetaHunt
Title: Privacy-Preserving Meta-Analysis via Low-Rank Basis Hunting
Version: 0.1.0
Authors@R: c(
    person("Wenqi", "Shi", , "wenqishi18@gmail.com", role = c("aut", "cre")),
    person("Kosuke", "Imai", , "imai@harvard.edu", role = "aut"),
    person("Yi", "Zhang", , "yizhang0017@gmail.com", role = "aut"))
Description: Tools for privacy-preserving meta-analysis of function-valued
    quantities across heterogeneous studies. Implements the 'MetaHunt'
    pipeline, including the denoised functional Successive Projection
    Algorithm (d-fSPA) for basis hunting, constrained weight estimation,
    Dirichlet regression of weights on study-level covariates, target
    prediction, and split/cross conformal prediction intervals. Operates
    on aggregate-level function evaluations, so individual-level data
    from source studies are not required. Methodology described in
    Shi, Imai, and Zhang (2026) <doi:10.48550/arXiv.2604.23847>.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.0)
Imports: quadprog, DirichletReg, stats, graphics, grDevices, withr
Suggests: grf, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
URL: https://github.com/WShi18/MetaHunt,
        https://wshi18.github.io/MetaHunt/,
        https://arxiv.org/abs/2604.23847
BugReports: https://github.com/WShi18/MetaHunt/issues
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-07 00:50:01 UTC; sophiashi
Author: Wenqi Shi [aut, cre],
  Kosuke Imai [aut],
  Yi Zhang [aut]
Maintainer: Wenqi Shi <wenqishi18@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-12 19:20:43 UTC
Built: R 4.4.3; ; 2026-05-12 19:58:41 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(coef,metahunt_weight_model)
S3method(plot,metahunt)
S3method(plot,metahunt_conformal)
S3method(predict,metahunt)
S3method(predict,metahunt_weight_model)
S3method(print,dfspa)
S3method(print,metahunt)
S3method(print,metahunt_conformal)
S3method(print,metahunt_denoising_search)
S3method(print,metahunt_weight_model)
S3method(print,minmax_regret)
S3method(print,summary.metahunt)
S3method(print,summary.metahunt_conformal)
S3method(summary,metahunt)
S3method(summary,metahunt_conformal)
export(apply_wrapper)
export(build_grid)
export(conformal_from_fit)
export(coverage)
export(cross_conformal)
export(cv_error_curve)
export(dfspa)
export(f_hat_from_models)
export(fit_weight_model)
export(metahunt)
export(minmax_regret)
export(predict_target)
export(project_to_simplex)
export(reconstruction_error_curve)
export(select_denoising_params)
export(split_conformal)

# MetaHunt <a href="https://wshi18.github.io/MetaHunt/"><img src="man/figures/logo.png" align="right" height="138" alt="MetaHunt website" /></a>

<!-- badges: start -->
[![R-CMD-check](https://github.com/WShi18/MetaHunt/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/WShi18/MetaHunt/actions/workflows/R-CMD-check.yaml)
[![pkgdown](https://github.com/WShi18/MetaHunt/actions/workflows/pkgdown.yaml/badge.svg)](https://wshi18.github.io/MetaHunt/)
[![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![License: MIT](https://img.shields.io/badge/license-MIT-yellow.svg)](https://opensource.org/licenses/MIT)
<!-- badges: end -->

**MetaHunt** is an R package for privacy-preserving meta-analysis of
function-valued quantities (e.g. regression functions, conditional
average treatment effect functions) across heterogeneous studies
(Shi, Imai, & Zhang 2024, [arXiv:2604.23847](https://arxiv.org/abs/2604.23847)).

Documentation: <https://wshi18.github.io/MetaHunt/>

The package implements the methodology of

> Shi, W., Imai, K., and Zhang, Y. *Privacy-preserving Meta-analysis
> through Low-Rank Basis Hunting.*

Given aggregate-level estimates of a function of interest from several
source studies, together with study-level covariates, MetaHunt recovers
a small set of latent basis functions that span all studies under a
low-rank assumption, models how the mixing weights depend on study-level
covariates via Dirichlet regression, and predicts the corresponding
function for a new target population with valid conformal prediction
intervals. Individual-level data from source studies are not required,
which makes the procedure suitable for federated and multi-site
settings.

The package also exports `minmax_regret()`, a covariate-free baseline
based on the worst-case-regret aggregator of Zhang, Huang, and Imai
(2024, [arXiv:2412.11136](https://arxiv.org/abs/2412.11136)). It is
useful when the number of source studies is small or the study-level
covariates are uninformative.

## Installation

```r
# install.packages("remotes")
remotes::install_github("WShi18/MetaHunt")
```

## Quick example

```r
library(MetaHunt)
set.seed(1)

# Simulate m studies, each represented by its function on a shared grid.
G <- 30; m <- 60
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- as.matrix(W) %*% cbind(c(1, -0.5), c(-0.4, 1), c(0, 0))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

# Fit and predict for a new study with W_new = (0, 0).
fit   <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
W_new <- data.frame(w1 = 0, w2 = 0)
pred  <- predict(fit, newdata = W_new, wrapper = mean)

# 95% conformal interval for the same target.
ci <- split_conformal(F_hat, W, W_new, K = 3,
                      wrapper = mean, alpha = 0.05, seed = 1,
                      dfspa_args = list(denoise = FALSE))
ci$prediction
c(ci$lower, ci$upper)
```

See `vignette("metahunt-intro")` for a full walkthrough including
data preparation from fitted model objects, rank selection, conformal
intervals (split / cross / pre-fit), and the minimax-regret baseline.

## Methods provided

1. **`metahunt()`** + `predict()`, `plot()`, `summary()` — end-to-end fitter.
2. **`split_conformal()`, `cross_conformal()`, `conformal_from_fit()`** — prediction intervals.
3. **`coverage()`, `summary.metahunt_conformal()`** — conformal diagnostics.
4. **`minmax_regret()`** — covariate-free worst-case-regret baseline (Zhang, Huang, & Imai 2024).
5. **`reconstruction_error_curve()`, `cv_error_curve()`** — rank-selection diagnostics.
6. **`select_denoising_params()`** — cross-validated tuning of `(N, Delta)`.
7. **`f_hat_from_models()`, `build_grid()`** — onramp from fitted-model lists to the package's matrix inputs.

The individual pipeline building blocks (`dfspa()`,
`project_to_simplex()`, `fit_weight_model()`, `predict_target()`,
`apply_wrapper()`) are also exported and can be composed independently.

## Citation

If you use MetaHunt in academic work, please cite:

> Shi, W., Imai, K., and Zhang, Y. (2024). *Privacy-preserving
> meta-analysis through low-rank basis hunting.* arXiv:2604.23847.
> <https://arxiv.org/abs/2604.23847>

## Status

`v0.1.0` is the first GitHub release. The API is stabilising; user
feedback is welcome via [issues](https://github.com/WShi18/MetaHunt/issues).

## License

MIT © Wenqi Shi, Kosuke Imai, Yi Zhang

## ----gw-knit-opts, include = FALSE--------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)

## ----setup--------------------------------------------------------------------
library(MetaHunt)
set.seed(1)

## ----sim-data-----------------------------------------------------------------
m <- 40
G <- 30
x <- seq(0, 1, length.out = G)

# True bases on the grid.
basis <- rbind(sin(pi * x), cos(pi * x), x)

# Mixing weights driven by two latent covariates.
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
eta  <- as.matrix(W) %*% beta
pi_true <- exp(eta) / rowSums(exp(eta))

F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
dim(F_hat)

## ----two-weights--------------------------------------------------------------
gw_uniform <- rep(1 / G, G)

# Up-weight x > 0.5 by ~5x. Renormalisation is optional but tidy.
gw_right   <- ifelse(x > 0.5, 5, 1)
gw_right   <- gw_right / sum(gw_right) * G   # comparable scale, optional

fit_unif  <- dfspa(F_hat, K = 3, grid_weights = gw_uniform, denoise = FALSE)
fit_right <- dfspa(F_hat, K = 3, grid_weights = gw_right,   denoise = FALSE)

# Index sets selected can differ.
fit_unif$original_indices
fit_right$original_indices

## ----overlay-plot-------------------------------------------------------------
matplot(
  x,
  cbind(fit_unif$bases[1, ], fit_right$bases[1, ]),
  type = "l", lty = 1, lwd = 2,
  col  = c("#0072B2", "#D55E00"),
  xlab = "x",
  ylab = "first recovered basis",
  main = expression(
    "Recovered first-vertex basis under uniform vs. target-weighted " ~ L^2 * (mu)
  )
)
abline(v = 0.5, lty = 2, col = "grey60")
legend("topright",
       legend = c("uniform", "up-weight x > 0.5", "x = 0.5 boundary"),
       col    = c("#0072B2", "#D55E00", "grey60"),
       lty = c(1, 1, 2), lwd = c(2, 2, 1), bty = "n")

## ----ws-knit-opts, include = FALSE--------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)

## ----setup--------------------------------------------------------------------
library(MetaHunt)
set.seed(1)

## ----ws-simulate-trials, eval = requireNamespace("grf", quietly = TRUE)-------
m <- 8
n_per_site <- 200
G <- 30

W <- data.frame(
  year        = sample(2010:2020, m, replace = TRUE),
  pct_treated = round(runif(m, 0.3, 0.6), 2)
)

site_data_list <- lapply(seq_len(m), function(i) {
  age <- runif(n_per_site, 30, 80)
  T   <- rbinom(n_per_site, 1, W$pct_treated[i])
  site_eff <- (W$year[i] - 2015) / 5   # site-level shift in CATE
  tau_age  <- 0.02 * (age - 50) + site_eff
  Y0  <- 0.01 * age + rnorm(n_per_site, sd = 0.5)
  Y1  <- Y0 + tau_age
  Y   <- ifelse(T == 1, Y1, Y0)
  data.frame(Y = Y, age = age, T = T)
})

grid <- data.frame(age = seq(30, 80, length.out = G))

## ----ws-fit-cf, eval = requireNamespace("grf", quietly = TRUE)----------------
cf_models <- lapply(site_data_list, function(d)
  grf::causal_forest(X = matrix(d$age, ncol = 1),
                     Y = d$Y,
                     W = d$T,
                     num.trees = 200))

## ----ws-build-fhat, eval = requireNamespace("grf", quietly = TRUE)------------
cate_predict <- function(model, grid) {
  as.numeric(stats::predict(model, newdata = matrix(grid$age, ncol = 1))$predictions)
}
F_hat <- f_hat_from_models(cf_models, grid, predict_fn = cate_predict)
dim(F_hat)

## ----ws-fit-metahunt, eval = requireNamespace("grf", quietly = TRUE)----------
fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
W_new <- data.frame(year = 2018, pct_treated = 0.45)
ate_pred <- predict(fit, newdata = W_new, wrapper = mean)
ate_pred

## ----wrapper-mean, eval = requireNamespace("grf", quietly = TRUE)-------------
predict(fit, newdata = W_new, wrapper = mean)

## ----wrapper-restricted, eval = requireNamespace("grf", quietly = TRUE)-------
restricted_pos_mean <- function(f) sum(pmax(f, 0)) / length(f)
predict(fit, newdata = W_new, wrapper = restricted_pos_mean)

## ----wrapper-endpoint, eval = requireNamespace("grf", quietly = TRUE)---------
endpoint_contrast <- function(f) f[length(f)] - f[1]
predict(fit, newdata = W_new, wrapper = endpoint_contrast)

## ----ws-split-scalar, eval = requireNamespace("grf", quietly = TRUE)----------
# Use 7 sites for training+calibration, predict for the held-out 8th
tr_cal <- 1:7; new <- 8
res <- split_conformal(
  F_hat[tr_cal, , drop = FALSE],
  W[tr_cal, , drop = FALSE],
  W[new, , drop = FALSE],
  K = 3, wrapper = mean, alpha = 0.1, cal_frac = 0.5, seed = 1,
  dfspa_args = list(denoise = FALSE)
)
data.frame(prediction = res$prediction,
           lower      = res$lower,
           upper      = res$upper)

## ----cp-knit-opts, include = FALSE--------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)

## ----setup--------------------------------------------------------------------
library(MetaHunt)
set.seed(1)

## ----cp-simulate--------------------------------------------------------------
# m = 80 is large enough that with cal_frac = 0.5 and alpha = 0.05 the conformal quantile is finite.
m <- 80; G <- 20; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(as.matrix(W) %*% beta); pi_true <- pi_true / rowSums(pi_true)
F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

## ----cp-wnew------------------------------------------------------------------
W_new <- data.frame(w1 = c(0, 1, -1), w2 = c(0, -0.5, 1))

## ----cp-split-pointwise-------------------------------------------------------
res_pw <- split_conformal(F_hat, W, W_new, K = K_true, alpha = 0.05,
                          cal_frac = 0.5, seed = 1,
                          dfspa_args = list(denoise = FALSE))
plot(res_pw, target_idx = 1, x_axis = x)

## ----cp-split-scalar----------------------------------------------------------
res_scalar <- split_conformal(F_hat, W, W_new, K = K_true,
                              wrapper = mean, alpha = 0.05,
                              cal_frac = 0.5, seed = 1,
                              dfspa_args = list(denoise = FALSE))
data.frame(prediction = res_scalar$prediction,
           lower      = res_scalar$lower,
           upper      = res_scalar$upper)

## ----cp-cross-----------------------------------------------------------------
res_cross <- cross_conformal(F_hat, W, W_new, K = K_true, n_folds = 4,
                             wrapper = mean, alpha = 0.1, seed = 1,
                             dfspa_args = list(denoise = FALSE))
res_cross

## ----cp-prefit----------------------------------------------------------------
fit <- metahunt(F_hat, W, K = K_true, dfspa_args = list(denoise = FALSE))
pi_hat <- project_to_simplex(F_hat, fit$dfspa_fit$bases)
res_pre <- conformal_from_fit(
  dfspa_fit = fit$dfspa_fit, weight_model = fit$weight_model,
  F_cal = F_hat, W_cal = W, W_new = W_new,
  wrapper = mean, alpha = 0.1
)
res_pre

## ----cp-small-m-warning, warning = TRUE---------------------------------------
m_small <- 30  # too small for alpha = 0.05 with cal_frac = 0.5
F_small <- F_hat[1:m_small, , drop = FALSE]
W_small <- W[1:m_small, , drop = FALSE]
res_inf <- split_conformal(F_small, W_small, W_new, K = K_true,
                           alpha = 0.05, cal_frac = 0.5, seed = 1,
                           dfspa_args = list(denoise = FALSE))
res_inf$quantile        # Inf — quantile is unbounded
range(res_inf$lower)    # -Inf
range(res_inf$upper)    #  Inf

## ----cdk-knit-opts, include = FALSE-------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)

## ----setup--------------------------------------------------------------------
library(MetaHunt)
set.seed(1)

## ----cdk-simulate-------------------------------------------------------------
m <- 30; G <- 20; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(as.matrix(W) %*% beta); pi_true <- pi_true / rowSums(pi_true)
F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

## ----cdk-elbow----------------------------------------------------------------
elbow <- reconstruction_error_curve(F_hat, K_range = 2:6,
                                    dfspa_args = list(denoise = FALSE))
plot(elbow$K, elbow$error, type = "b",
     xlab = "K", ylab = "reconstruction error",
     main = "Reconstruction error vs K",
     ylim = c(0, max(elbow$error, na.rm = TRUE) * 1.05))

## ----cdk-cv-------------------------------------------------------------------
cv <- cv_error_curve(F_hat, W, K_range = 2:6, n_folds = 4,
                     dfspa_args = list(denoise = FALSE), seed = 1)
plot(cv$K, cv$cv_error, type = "b",
     xlab = "K", ylab = "CV prediction error",
     main = "CV prediction error vs K",
     ylim = c(0, max(cv$cv_error, na.rm = TRUE) * 1.05))

## ----cdk-bypass---------------------------------------------------------------
fit_no <- metahunt(F_hat, W, K = K_true,
                   dfspa_args = list(denoise = FALSE))
fit_no

## ----cdk-dfspa-knobs----------------------------------------------------------
fit_no <- metahunt(F_hat, W, K = K_true,
                   dfspa_args = list(denoise = FALSE))

fit_manual <- metahunt(F_hat, W, K = K_true,
                       dfspa_args = list(N = 0.5 * log(nrow(F_hat)),
                                         Delta = 0.4))

## ----cdk-select-denoising, warning = FALSE------------------------------------
tune <- select_denoising_params(F_hat, W, K = K_true, n_folds = 4, seed = 1)
tune$best

## ----dp-knit-opts, include = FALSE--------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)

## ----setup--------------------------------------------------------------------
library(MetaHunt)
set.seed(1)

## ----build-grid-1d------------------------------------------------------------
ref <- data.frame(age = rnorm(500, 60, 10),
                  bp  = rnorm(500, 130, 15),
                  bmi = rnorm(500, 28, 4))
grid <- build_grid(ref, n_grid = 50, seed = 1)
dim(grid)
head(grid)

## ----ranger-example, eval = FALSE---------------------------------------------
# library(ranger)
# centre_models <- lapply(centre_data_list,
#                         function(d) ranger(y ~ ., data = d))
# F_hat <- f_hat_from_models(centre_models, grid)

## ----grf-example, eval = FALSE------------------------------------------------
# library(grf)
# centre_models <- lapply(centre_data_list,
#                         function(d) causal_forest(d$X, d$Y, d$W))
# F_hat <- f_hat_from_models(centre_models, grid)

## ----lm-onramp----------------------------------------------------------------
m <- 8
centre_meta <- data.frame(
  region     = factor(sample(c("N", "S", "E", "W"), m, replace = TRUE)),
  mean_age   = round(runif(m, 50, 70)),
  pct_female = round(runif(m, 0.4, 0.6), 2)
)

# Each centre fits a quadratic on a single covariate `x`.
make_centre_data <- function(i) {
  x <- runif(80, -1, 1)
  beta <- centre_meta$mean_age[i] / 60      # toy effect of metadata
  data.frame(x = x, y = beta * x + 0.3 * x^2 + rnorm(80, sd = 0.2))
}
centre_models <- lapply(seq_len(m), function(i)
  stats::lm(y ~ poly(x, 2), data = make_centre_data(i)))

# A 1-D grid in the centres' covariate space.
grid_centres  <- data.frame(x = seq(-1, 1, length.out = 30))
F_hat_centres <- f_hat_from_models(centre_models, grid_centres)

dim(F_hat_centres)        # 8 x 30
F_hat_centres[1:3, 1:5]

## ----predict-fn-demo----------------------------------------------------------
# Toy "model" that is just a list with a slope. predict_fn evaluates it.
fake_models <- lapply(seq_len(4), function(i)
  list(slope = i / 4, intercept = 0))

custom_predict <- function(model, grid) {
  model$intercept + model$slope * grid$x
}

F_hat_custom <- f_hat_from_models(fake_models, grid_centres,
                                  predict_fn = custom_predict)
dim(F_hat_custom)         # 4 x 30

## ----multi-d-grid-------------------------------------------------------------
# A 3-covariate reference dataset and a sub-sampled grid.
ref3   <- data.frame(age = rnorm(400, 60, 10),
                     bp  = rnorm(400, 130, 15),
                     bmi = rnorm(400, 28,  4))
grid3  <- build_grid(ref3, n_grid = 25, seed = 1)
dim(grid3)

# Each centre fits an lm on (age, bp, bmi); slopes vary across centres
# so there is genuine cross-centre heterogeneity to recover.
set.seed(2)
m3 <- 8
centre_data3 <- lapply(seq_len(m3), function(i) {
  n_i <- 60
  age <- rnorm(n_i, 60, 10)
  bp  <- rnorm(n_i, 130, 15)
  bmi <- rnorm(n_i, 28, 4)
  # slopes vary across centres
  beta_age <- 0.02 + 0.03 * (i / m3)
  beta_bp  <- -0.01 + 0.02 * cos(pi * i / m3)
  beta_bmi <- 0.05 - 0.04 * (i / m3)
  y <- beta_age * age + beta_bp * bp + beta_bmi * bmi + rnorm(n_i, sd = 0.3)
  data.frame(age = age, bp = bp, bmi = bmi, y = y)
})
centre_models3 <- lapply(centre_data3, function(d) stats::lm(y ~ age + bp + bmi, data = d))

F_hat3 <- f_hat_from_models(centre_models3, grid3)
dim(F_hat3)               # 8 x 25

## ----sanity-------------------------------------------------------------------
# Right shape: m studies x G grid points.
dim(F_hat3)

# Numeric, no NA.
is.numeric(F_hat3)
anyNA(F_hat3)

# Rows look like functions of similar magnitude (large outliers can
# dominate d-fSPA's `Delta`).
summary(apply(F_hat3, 1, function(r) c(min = min(r), max = max(r))))

## ----sanity-plot--------------------------------------------------------------
matplot(grid_centres$x, t(F_hat_centres), type = "l", lty = 1,
        col = "grey50",
        xlab = "x", ylab = expression(hat(f)(x)),
        main = "Per-centre fitted functions on the shared grid")

#' Predict the target function for new study-level covariates
#'
#' Given a fitted d-fSPA basis decomposition and a fitted weight model,
#' compute the predicted target function on the shared grid as
#' \deqn{\tilde f^{(0)}(\cdot) = \sum_{k=1}^{\hat K} \tilde\pi_{0k}\, \hat g_k(\cdot),
#'   \qquad \tilde\pi_0 = \widehat{\mathcal M}(\mathbf W_0).}
#'
#' @param dfspa_fit A `"dfspa"` object produced by [dfspa()]. Its `bases`
#'   slot (`K`-by-`G_grid` matrix) provides the basis functions on the grid.
#' @param weight_model A `"metahunt_weight_model"` object produced by
#'   [fit_weight_model()]. Must have `weight_model$K == nrow(dfspa_fit$bases)`.
#' @param W_new A matrix or data frame of study-level covariates for the new
#'   target studies, with columns matching those used to fit `weight_model`.
#'
#' @return An `nrow(W_new)`-by-`G_grid` numeric matrix; row `j` is the
#'   predicted target function on the grid for the `j`-th new study.
#'
#' @seealso [apply_wrapper()] to reduce predicted functions to scalars.
#'
#' @examples
#' set.seed(1)
#' G <- 40; m <- 60; K_true <- 3
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#'
#' # generate study-level covariates and softmax weights
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
#' eta  <- as.matrix(W) %*% beta
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.02), m, G)
#'
#' fit    <- dfspa(F_hat, K = K_true)
#' pi_hat <- project_to_simplex(F_hat, fit$bases)
#' wm     <- fit_weight_model(pi_hat, W)
#'
#' W_new   <- data.frame(w1 = c(0, 1), w2 = c(0, -1))
#' f_pred  <- predict_target(fit, wm, W_new)
#' dim(f_pred)   # 2 x G
#'
#' @export
predict_target <- function(dfspa_fit, weight_model, W_new) {
  if (!inherits(dfspa_fit, "dfspa")) {
    stop("`dfspa_fit` must be a `dfspa` object (from `dfspa()`).")
  }
  if (!inherits(weight_model, "metahunt_weight_model")) {
    stop("`weight_model` must be a `metahunt_weight_model` object (from `fit_weight_model()`).")
  }
  if (weight_model$K != nrow(dfspa_fit$bases)) {
    stop(sprintf(
      "Basis/weight-model mismatch: dfspa_fit has K = %d bases but weight_model has K = %d.",
      nrow(dfspa_fit$bases), weight_model$K))
  }

  pi_new <- stats::predict(weight_model, newdata = W_new)
  pi_new %*% dfspa_fit$bases
}

#' Reduce predicted functions to scalars via a user-supplied wrapper
#'
#' Many downstream quantities of interest (average treatment effect,
#' pointwise predictions, or other functionals of \eqn{f^{(0)}}) are scalar
#' summaries of the predicted function. `apply_wrapper()` applies any
#' user-supplied reduction to each row of a function matrix, with a default
#' of the weighted mean with respect to `grid_weights` (which coincides with
#' \eqn{\int f\,d\mu} when `grid_weights` represents \eqn{\mu}).
#'
#' @param F_mat An `n`-by-`G_grid` numeric matrix; each row a function
#'   on a shared grid (e.g. the output of [predict_target()]).
#' @param wrapper Either `NULL` (default, weighted mean) or a function that
#'   takes a single numeric vector of length `G_grid` and returns a scalar.
#'   Examples: `mean`, `median`, `max`, `function(f) f[17]` (point
#'   evaluation), `function(f) sum(f^2)`.
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector.
#'   Used only when `wrapper = NULL`. Defaults to uniform `1 / G_grid`.
#'
#' @return A length-`n` numeric vector of scalar summaries.
#'
#' @examples
#' F_mat <- matrix(1:12, nrow = 3, byrow = TRUE)  # 3 "functions" on a 4-point grid
#' apply_wrapper(F_mat)                            # row means (uniform grid weights)
#' apply_wrapper(F_mat, wrapper = max)             # row maxes
#' apply_wrapper(F_mat, wrapper = function(f) f[2]) # point evaluation at grid idx 2
#'
#' @export
apply_wrapper <- function(F_mat, wrapper = NULL, grid_weights = NULL) {
  if (!is.matrix(F_mat) || !is.numeric(F_mat)) {
    stop("`F_mat` must be a numeric matrix.")
  }
  G_grid <- ncol(F_mat)

  if (is.null(wrapper)) {
    if (is.null(grid_weights)) {
      grid_weights <- rep(1 / G_grid, G_grid)
    } else {
      if (!is.numeric(grid_weights) || length(grid_weights) != G_grid ||
          any(!is.finite(grid_weights)) || any(grid_weights < 0)) {
        stop("`grid_weights` must be a length-`ncol(F_mat)` non-negative numeric vector.")
      }
      if (sum(grid_weights) <= 0) {
        stop("`grid_weights` must sum to a positive value.")
      }
    }
    return(as.numeric(F_mat %*% grid_weights) / sum(grid_weights))
  }

  if (!is.function(wrapper)) {
    stop("`wrapper` must be a function or NULL.")
  }
  vals <- apply(F_mat, 1, wrapper)
  if (!is.numeric(vals) || length(vals) != nrow(F_mat)) {
    stop("`wrapper` must return a single numeric value for each row.")
  }
  as.numeric(vals)
}

test_that("predict_target reproduces the linear combination pi %*% bases", {
  set.seed(1)
  G <- 40; m <- 80; K <- 3
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  eta  <- as.matrix(W) %*% beta
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.01), m, G)

  fit     <- dfspa(F_hat, K = K)
  pi_hat  <- project_to_simplex(F_hat, fit$bases)
  wm      <- fit_weight_model(pi_hat, W)

  W_new   <- data.frame(w1 = c(0, 0.5), w2 = c(0.2, -0.3))
  f_pred  <- predict_target(fit, wm, W_new)
  pi_pred <- predict(wm, newdata = W_new)

  expect_equal(dim(f_pred), c(2L, G))
  expect_equal(f_pred, pi_pred %*% fit$bases)
})

test_that("predict_target errors on mismatched K between fit and weight model", {
  set.seed(2)
  G <- 20; m <- 40
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m))
  pi_true <- matrix(rgamma(m * 3, 1), m, 3); pi_true <- pi_true / rowSums(pi_true)
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.01), m, G)

  fit_K2 <- dfspa(F_hat, K = 2)
  pi_hat <- project_to_simplex(F_hat, dfspa(F_hat, K = 3)$bases)
  wm_K3  <- fit_weight_model(pi_hat, W)
  expect_error(predict_target(fit_K2, wm_K3, W), "mismatch")
})

test_that("apply_wrapper default equals weighted mean", {
  F_mat <- matrix(c(1, 2, 3, 4,
                    10, 20, 30, 40), nrow = 2, byrow = TRUE)
  expect_equal(apply_wrapper(F_mat), c(mean(1:4), mean(c(10, 20, 30, 40))))
  w <- c(0.1, 0.2, 0.3, 0.4)
  expect_equal(apply_wrapper(F_mat, grid_weights = w),
               as.numeric(F_mat %*% w) / sum(w))
})

test_that("apply_wrapper supports arbitrary wrappers", {
  F_mat <- matrix(c(1, 2, 3,
                    7, 5, 6), nrow = 2, byrow = TRUE)
  expect_equal(apply_wrapper(F_mat, wrapper = max), c(3, 7))
  expect_equal(apply_wrapper(F_mat, wrapper = function(f) f[2]), c(2, 5))
})

test_that("apply_wrapper errors on bad wrapper output", {
  F_mat <- matrix(1:6, 2, 3)
  bad_wrapper <- function(f) c(1, 2)
  expect_error(apply_wrapper(F_mat, wrapper = bad_wrapper), "single numeric")
})

#' Project study-level functions onto the simplex spanned by basis functions
#'
#' For each study `i`, solves the constrained projection
#' \deqn{\hat\pi_i = \arg\min_{\pi \in \Delta_{K-1}}
#'   \left\| \hat f^{(i)} - \sum_{k=1}^K \pi_k \hat g_k \right\|_{L^2(\mu)}}
#' where the norm is the weighted \eqn{L^2} norm defined by `grid_weights`.
#' This is Equation (3) of the paper and yields the study-specific weights
#' `hat pi_i` used downstream for weight-model fitting and prediction.
#'
#' @details
#' The projection reduces to the quadratic program
#' \deqn{\min_{\pi \in \mathbb R^K}\ \pi^\top D\,\pi - 2\,d^\top \pi
#'   \quad \text{s.t.} \quad \mathbf 1^\top \pi = 1,\ \pi \ge 0}
#' with \eqn{D = G W G^\top} and \eqn{d = G W f^{(i)}}, where
#' \eqn{G} is the `K`-by-`G_grid` basis matrix, \eqn{W = \mathrm{diag}(}`grid_weights`\eqn{)},
#' and \eqn{f^{(i)}} is the `i`-th row of `F_hat`. Solved via
#' [quadprog::solve.QP()]. A tiny ridge is added to `D` for numerical stability.
#'
#' @param F_hat An `m`-by-`G_grid` numeric matrix; row `i` is the study
#'   function \eqn{\hat f^{(i)}} evaluated on the shared grid.
#' @param bases A `K`-by-`G_grid` numeric matrix of basis functions on the
#'   same grid, typically the `bases` slot of a [dfspa()] result.
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector of
#'   grid weights defining the \eqn{L^2(\mu)} inner product. Defaults to
#'   uniform weights `1 / G_grid`.
#' @param ridge Small non-negative scalar added to the diagonal of the QP
#'   Hessian for numerical stability. Defaults to `1e-10`.
#'
#' @return An `m`-by-`K` numeric matrix of simplex weights; rows sum to 1
#'   and entries are non-negative.
#'
#' @examples
#' set.seed(1)
#' G <- 40
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' true_pi <- rbind(diag(3), c(0.4, 0.3, 0.3), c(0.1, 0.7, 0.2))
#' F_hat <- true_pi %*% basis
#' fit <- dfspa(F_hat, K = 3, denoise = FALSE)
#' pi_hat <- project_to_simplex(F_hat, fit$bases)
#' round(pi_hat, 3)
#'
#' @export
project_to_simplex <- function(F_hat, bases, grid_weights = NULL,
                               ridge = 1e-10) {
  if (!is.matrix(F_hat) || !is.numeric(F_hat)) {
    stop("`F_hat` must be a numeric matrix with studies as rows.")
  }
  if (!is.matrix(bases) || !is.numeric(bases)) {
    stop("`bases` must be a numeric matrix with bases as rows.")
  }
  if (anyNA(F_hat) || anyNA(bases)) {
    stop("`F_hat` and `bases` must not contain NA values.")
  }
  if (ncol(F_hat) != ncol(bases)) {
    stop("`F_hat` and `bases` must have the same number of columns (grid size).")
  }
  if (!is.numeric(ridge) || length(ridge) != 1L || ridge < 0 || !is.finite(ridge)) {
    stop("`ridge` must be a non-negative finite scalar.")
  }

  m <- nrow(F_hat)
  G_grid <- ncol(F_hat)
  K <- nrow(bases)

  if (is.null(grid_weights)) {
    grid_weights <- rep(1 / G_grid, G_grid)
  } else {
    if (!is.numeric(grid_weights) || length(grid_weights) != G_grid ||
        any(!is.finite(grid_weights)) || any(grid_weights < 0)) {
      stop("`grid_weights` must be a length-`ncol(F_hat)` non-negative numeric vector.")
    }
  }

  if (K == 1L) {
    return(matrix(1, nrow = m, ncol = 1L))
  }

  Gw   <- sweep(bases, 2, grid_weights, `*`)   # K x G_grid
  Dmat <- tcrossprod(Gw, bases)                # K x K, = G W G^T
  Dmat <- Dmat + ridge * diag(K)

  Amat <- cbind(rep(1, K), diag(K))            # K x (K + 1)
  bvec <- c(1, rep(0, K))

  pi_hat <- matrix(0, m, K)
  for (i in seq_len(m)) {
    dvec <- as.numeric(Gw %*% F_hat[i, ])
    sol  <- quadprog::solve.QP(Dmat = Dmat, dvec = dvec,
                               Amat = Amat, bvec = bvec, meq = 1L)
    pi_i <- pmax(sol$solution, 0)
    s <- sum(pi_i)
    if (s <= 0) {
      warning("project_to_simplex(): solver returned all-zero weights for row ",
              i, "; falling back to uniform weights.", call. = FALSE)
      pi_hat[i, ] <- 1 / K
    } else {
      pi_hat[i, ] <- pi_i / s
    }
  }
  pi_hat
}

#' Fit a weight model mapping study-level covariates to simplex weights
#'
#' Given a matrix of simplex-valued weights \eqn{\hat\pi_1,\ldots,\hat\pi_m}
#' (e.g. from [project_to_simplex()]) and associated study-level covariates
#' \eqn{\mathbf W_1,\ldots,\mathbf W_m}, fit a model
#' \eqn{\widehat{\mathcal M}:\mathbf W \mapsto \boldsymbol\pi}.
#' The default method is Dirichlet regression via the `DirichletReg` package.
#'
#' @details
#' Dirichlet regression cannot handle weights exactly at the simplex boundary
#' (`0` or `1`), which frequently arise after constrained projection. Before
#' fitting, rows of `pi_hat` are shrunk toward the barycenter via
#' \eqn{\tilde\pi = (\pi + \varepsilon) / (1 + K\varepsilon)}, with
#' \eqn{\varepsilon} set by `boundary_eps`.
#'
#' @param pi_hat An `m`-by-`K` numeric matrix of simplex weights; rows must
#'   be non-negative and sum to 1 (up to tolerance `1e-6`).
#' @param W An `m`-by-`p` matrix or data frame of study-level covariates.
#' @param method Weight-model method. Currently only `"dirichlet"` is
#'   supported.
#' @param boundary_eps Small positive scalar used to shrink weights away
#'   from the simplex boundary before Dirichlet fitting. Defaults to `1e-4`.
#' @param formula Optional RHS-only formula (e.g. `~ x1 + I(x2^2)`) describing
#'   the covariate part of the Dirichlet regression. Defaults to `~ .`
#'   (all columns of `W`).
#' @param ... Passed through to [DirichletReg::DirichReg()].
#'
#' @return An object of class `"metahunt_weight_model"`: a list with the
#'   fitted model, formula, method, `K`, and training covariate names.
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' m <- 80; K <- 3; p <- 2
#' W <- matrix(rnorm(m * p), m, p); colnames(W) <- c("w1", "w2")
#' # generate simplex weights driven by W
#' eta <- cbind(0.5 * W[, 1], -0.3 * W[, 2], rep(0, m))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' pi_hat <- pi_true + matrix(rnorm(m * K, sd = 0.01), m, K)
#' pi_hat <- pmax(pi_hat, 0); pi_hat <- pi_hat / rowSums(pi_hat)
#' model <- fit_weight_model(pi_hat, W)
#' predict(model, newdata = matrix(c(0, 0), 1, 2, dimnames = list(NULL, c("w1","w2"))))
#' }
#'
#' @export
fit_weight_model <- function(pi_hat, W, method = c("dirichlet"),
                             boundary_eps = 1e-4, formula = NULL, ...) {
  method <- match.arg(method)

  if (!is.numeric(pi_hat)) stop("`pi_hat` must be numeric.")
  pi_hat <- as.matrix(pi_hat)
  if (anyNA(pi_hat)) stop("`pi_hat` must not contain NA values.")
  if (any(pi_hat < -1e-8)) stop("`pi_hat` must have non-negative entries.")
  pi_hat <- pmax(pi_hat, 0)
  rs <- rowSums(pi_hat)
  if (any(abs(rs - 1) > 1e-6)) {
    stop("Rows of `pi_hat` must sum to 1 (tolerance 1e-6).")
  }
  m <- nrow(pi_hat); K <- ncol(pi_hat)

  if (is.matrix(W)) {
    W <- as.data.frame(W)
  }
  if (!is.data.frame(W)) {
    stop("`W` must be a matrix or data frame.")
  }
  if (nrow(W) != m) {
    stop("`W` must have the same number of rows as `pi_hat`.")
  }
  if ("..Y.." %in% names(W)) {
    stop("`W` must not contain a column named '..Y..' (reserved).")
  }

  if (!is.numeric(boundary_eps) || length(boundary_eps) != 1L ||
      !is.finite(boundary_eps) || boundary_eps < 0 || boundary_eps >= 1 / K) {
    stop("`boundary_eps` must be a scalar in [0, 1/K).")
  }

  switch(method,
    dirichlet = fit_dirichlet_(pi_hat, W, boundary_eps, formula, ...)
  )
}

fit_dirichlet_ <- function(pi_hat, W, boundary_eps, formula, ...) {
  K <- ncol(pi_hat)
  pi_adj <- (pi_hat + boundary_eps) / (1 + K * boundary_eps)

  predictor_names <- names(W)
  if (is.null(formula)) {
    rhs <- if (length(predictor_names) == 0L) "1"
           else paste(predictor_names, collapse = " + ")
  } else {
    if (!inherits(formula, "formula") || length(formula) != 2L) {
      stop("`formula` must be a one-sided formula, e.g. ~ x1 + x2.")
    }
    rhs <- paste(deparse(formula[[2]]), collapse = " ")
  }

  metahunt_Y <- DirichletReg::DR_data(pi_adj)
  fmla <- stats::as.formula(paste("metahunt_Y ~", rhs))
  environment(fmla) <- environment()

  # Inline the formula into the call so that DirichReg's internal match.call()
  # sees the formula object, not the symbol `fmla`.
  fit_call <- substitute(DirichletReg::DirichReg(FMLA, data = W),
                         list(FMLA = fmla))
  fit <- eval(fit_call)

  structure(
    list(
      method          = "dirichlet",
      fit             = fit,
      formula         = fmla,
      K               = K,
      predictor_names = predictor_names,
      boundary_eps    = boundary_eps
    ),
    class = "metahunt_weight_model"
  )
}

#' Predict simplex weights for new study-level covariates
#'
#' @param object A fitted `"metahunt_weight_model"` from [fit_weight_model()].
#' @param newdata A matrix or data frame of new study-level covariates with
#'   the same columns used at fitting.
#' @param ... Ignored.
#'
#' @return An `nrow(newdata)`-by-`K` numeric matrix of predicted simplex
#'   weights (component means); rows sum to 1.
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' m <- 40; K <- 3
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- cbind(0.5 * W$w1, -0.3 * W$w2, rep(0, m))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' pi_hat <- pi_true + matrix(rnorm(m * K, sd = 0.01), m, K)
#' pi_hat <- pmax(pi_hat, 0); pi_hat <- pi_hat / rowSums(pi_hat)
#' model <- fit_weight_model(pi_hat, W)
#' predict(model, newdata = data.frame(w1 = c(0, 1), w2 = c(0, -1)))
#' }
#'
#' @export
predict.metahunt_weight_model <- function(object, newdata, ...) {
  if (is.matrix(newdata)) newdata <- as.data.frame(newdata)
  if (!is.data.frame(newdata)) {
    stop("`newdata` must be a matrix or data frame.")
  }
  missing_cols <- setdiff(object$predictor_names, names(newdata))
  if (length(missing_cols)) {
    stop("`newdata` is missing required columns: ",
         paste(missing_cols, collapse = ", "))
  }
  preds <- stats::predict(object$fit, newdata = newdata)
  preds <- unname(as.matrix(preds))
  preds
}

#' @export
print.metahunt_weight_model <- function(x, ...) {
  cat("MetaHunt weight model\n")
  cat("  method:     ", x$method, "\n")
  cat("  K:          ", x$K, "\n")
  cat("  predictors: ", paste(x$predictor_names, collapse = ", "), "\n")
  invisible(x)
}

#' Extract coefficients from a MetaHunt weight model
#'
#' Returns the regression coefficients from the underlying weight-model fit.
#' For the default `"dirichlet"` method this delegates to
#' [DirichletReg::DirichReg()]'s `coef()` method.
#'
#' @param object A fitted `"metahunt_weight_model"` from [fit_weight_model()].
#' @param ... Passed through to [stats::coef()].
#'
#' @return The coefficient vector / matrix returned by
#'   `DirichletReg::DirichReg()`'s `coef()` method (numeric vector or matrix
#'   depending on the parametrisation used by the underlying fit).
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' m <- 60; K <- 3
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- cbind(0.5 * W$w1, -0.3 * W$w2, rep(0, m))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' pi_hat <- pi_true + matrix(rnorm(m * K, sd = 0.01), m, K)
#' pi_hat <- pmax(pi_hat, 0); pi_hat <- pi_hat / rowSums(pi_hat)
#' model <- fit_weight_model(pi_hat, W)
#' coef(model)
#' }
#'
#' @export
coef.metahunt_weight_model <- function(object, ...) {
  if (!inherits(object, "metahunt_weight_model")) {
    stop("`object` must be a `metahunt_weight_model` object.")
  }
  stats::coef(object$fit, ...)
}

test_that("project_to_simplex recovers exact weights in the noiseless case", {
  set.seed(1)
  G <- 40
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  true_pi <- rbind(
    diag(3),
    c(0.5, 0.3, 0.2),
    c(0.1, 0.7, 0.2),
    c(0.25, 0.25, 0.5)
  )
  F_hat <- true_pi %*% basis

  pi_hat <- project_to_simplex(F_hat, basis)

  expect_equal(dim(pi_hat), c(6L, 3L))
  expect_true(all(pi_hat >= -1e-10))
  expect_equal(rowSums(pi_hat), rep(1, 6), tolerance = 1e-8)
  expect_equal(pi_hat, true_pi, tolerance = 1e-6)
})

test_that("project_to_simplex clips external points to the simplex boundary", {
  G <- 30
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x))
  # a function living outside the convex hull: 1.5 * g1
  f_ext <- 1.5 * basis[1, , drop = FALSE]
  pi_hat <- project_to_simplex(f_ext, basis)
  # closest point on the simplex is (1, 0) (pure g1)
  expect_equal(as.numeric(pi_hat), c(1, 0), tolerance = 1e-6)
})

test_that("project_to_simplex handles K = 1", {
  G <- 20
  F_hat <- matrix(rnorm(4 * G), 4, G)
  basis <- matrix(rnorm(G), 1, G)
  pi_hat <- project_to_simplex(F_hat, basis)
  expect_equal(pi_hat, matrix(1, 4, 1))
})

test_that("project_to_simplex rejects mismatched grids", {
  expect_error(
    project_to_simplex(matrix(0, 3, 10), matrix(0, 2, 12)),
    "same number of columns"
  )
})

test_that("fit_weight_model + predict give simplex outputs close to truth", {
  set.seed(123)
  m <- 150; K <- 3; p <- 2
  W <- matrix(rnorm(m * p), m, p)
  colnames(W) <- c("w1", "w2")
  # weights driven by W via softmax of linear predictor
  beta <- cbind(c(1.0, -0.8), c(-0.5, 1.2), c(0, 0))
  eta <- W %*% beta
  pi_true <- exp(eta) / rowSums(exp(eta))
  # add small simplex noise via a Dirichlet draw centered at pi_true
  pi_hat <- t(apply(pi_true, 1, function(p) {
    a <- 80 * p + 1e-3
    g <- stats::rgamma(K, shape = a)
    g / sum(g)
  }))

  model <- fit_weight_model(pi_hat, W)
  expect_s3_class(model, "metahunt_weight_model")
  expect_equal(model$K, K)

  # predict on held-out points
  W_new <- matrix(c(0, 0, 1, -1, -1, 1), ncol = 2, byrow = TRUE)
  colnames(W_new) <- c("w1", "w2")
  pi_new <- predict(model, newdata = W_new)
  expect_equal(dim(pi_new), c(3L, K))
  expect_equal(rowSums(pi_new), rep(1, 3), tolerance = 1e-6)
  expect_true(all(pi_new >= 0))

  # compare to the true softmax weights at the same points
  eta_new <- W_new %*% beta
  pi_true_new <- exp(eta_new) / rowSums(exp(eta_new))
  expect_true(max(abs(pi_new - pi_true_new)) < 0.15)
})

test_that("fit_weight_model handles boundary weights via shrinkage", {
  set.seed(7)
  m <- 60
  W <- data.frame(w1 = rnorm(m))
  # include pure vertex rows (boundary!)
  pi_hat <- rbind(
    matrix(c(1, 0, 0), 5, 3, byrow = TRUE),
    matrix(c(0, 1, 0), 5, 3, byrow = TRUE),
    matrix(c(0, 0, 1), 5, 3, byrow = TRUE)
  )
  # fill the rest with interior Dirichlet draws
  rest <- t(replicate(m - 15, {
    g <- stats::rgamma(3, shape = 1); g / sum(g)
  }))
  pi_hat <- rbind(pi_hat, rest)
  # DirichletReg may emit harmless convergence warnings on some BLAS
  # implementations (e.g. Windows); we only require that the fit completes
  # and returns a model of the right class.
  suppressWarnings(model <- fit_weight_model(pi_hat, W))
  expect_s3_class(model, "metahunt_weight_model")
})

test_that("fit_weight_model rejects malformed pi_hat", {
  W <- data.frame(w1 = rnorm(5))
  bad_pi <- matrix(c(0.5, 0.2, 0.1), 5, 3, byrow = TRUE)  # rows do not sum to 1
  expect_error(fit_weight_model(bad_pi, W), "sum to 1")
  neg_pi <- matrix(c(-0.1, 0.5, 0.6), 5, 3, byrow = TRUE)
  expect_error(fit_weight_model(neg_pi, W), "non-negative")
})

test_that("metahunt fits and predicts end-to-end", {
  set.seed(1)
  G <- 30; m <- 60
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  fit <- metahunt(F_hat, W, K = 3)

  expect_s3_class(fit, "metahunt")
  expect_equal(fit$m, m)
  expect_equal(fit$G_grid, G)
  expect_equal(fit$K, 3L)
  expect_equal(dim(fit$pi_hat), c(m, 3L))

  # predict with no wrapper returns matrix
  W_new <- W[1:5, , drop = FALSE]
  f_pred <- predict(fit, newdata = W_new)
  expect_equal(dim(f_pred), c(5L, G))

  # predict with wrapper returns vector
  s_pred <- predict(fit, newdata = W_new, wrapper = mean)
  expect_length(s_pred, 5L)
  expect_equal(s_pred, rowMeans(f_pred))
})

test_that("metahunt matches manual pipeline", {
  set.seed(2)
  G <- 25; m <- 50
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m))
  eta <- cbind(0.5 * W$w1, -0.3 * W$w1, rep(0, m))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.03), m, G)

  fit <- metahunt(F_hat, W, K = 3)

  manual_fit <- dfspa(F_hat, K = 3)
  manual_pi  <- project_to_simplex(F_hat, manual_fit$bases)
  manual_wm  <- fit_weight_model(manual_pi, W)
  manual_pred <- predict_target(manual_fit, manual_wm, W[1:3, , drop = FALSE])
  wrapped_pred <- predict(fit, newdata = W[1:3, , drop = FALSE])

  expect_equal(wrapped_pred, manual_pred)
})

test_that("metahunt fit integrates cleanly with conformal_from_fit", {
  set.seed(3)
  G <- 25; m <- 80
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m))
  eta <- cbind(0.5 * W$w1, -0.3 * W$w1, rep(0, m))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  tr <- 1:50; cal <- 51:70; new <- 71:80
  fit <- metahunt(F_hat[tr, ], W[tr, , drop = FALSE], K = 3,
                  dfspa_args = list(denoise = FALSE))

  res <- conformal_from_fit(
    fit$dfspa_fit, fit$weight_model,
    F_cal = F_hat[cal, , drop = FALSE],
    W_cal = W[cal, , drop = FALSE],
    W_new = W[new, , drop = FALSE],
    wrapper = mean
  )
  expect_s3_class(res, "metahunt_conformal")
  expect_length(res$prediction, length(new))
})

test_that("metahunt rejects invalid inputs", {
  G <- 20; m <- 30
  F_hat <- matrix(rnorm(m * G), m, G)
  W <- data.frame(w1 = rnorm(m))
  expect_error(metahunt(F_hat[-1, ], W, K = 3), "same number of rows")
  expect_error(metahunt(F_hat, W, K = 0), "positive integer")
  expect_error(metahunt(F_hat, W, K = 3, dfspa_args = "nope"), "must be named lists")
})

test_that("print.metahunt runs", {
  set.seed(4)
  F_hat <- matrix(rnorm(30 * 20), 30, 20)
  W <- data.frame(w1 = rnorm(30))
  fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
  expect_output(print(fit), "MetaHunt fit")
})

make_toy <- function(seed = 1, m = 100, G = 30) {
  set.seed(seed)
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
  list(F_hat = F_hat, W = W)
}

test_that("conformal_from_fit matches split_conformal with the same split", {
  d <- make_toy(m = 100)
  m <- nrow(d$F_hat)
  # deterministic split
  cal_idx <- 1:30
  tr_idx  <- setdiff(seq_len(m), cal_idx)
  W_new <- d$W[1:5, , drop = FALSE]

  # via high-level split_conformal with explicit calibration_idx
  a <- split_conformal(d$F_hat, d$W, W_new, K = 3,
                       calibration_idx = cal_idx, wrapper = mean)

  # via low-level conformal_from_fit with manual fit
  fit <- dfspa(d$F_hat[tr_idx, , drop = FALSE], K = 3)
  pih <- project_to_simplex(d$F_hat[tr_idx, , drop = FALSE], fit$bases)
  wm  <- fit_weight_model(pih, d$W[tr_idx, , drop = FALSE])
  b <- conformal_from_fit(fit, wm,
                          F_cal = d$F_hat[cal_idx, , drop = FALSE],
                          W_cal = d$W[cal_idx, , drop = FALSE],
                          W_new = W_new,
                          wrapper = mean)

  expect_equal(a$prediction, b$prediction)
  expect_equal(a$lower, b$lower)
  expect_equal(a$upper, b$upper)
  expect_equal(a$quantile, b$quantile)
})

test_that("conformal_from_fit pointwise: shapes", {
  d <- make_toy(m = 60)
  tr  <- 1:40; cal <- 41:55; new <- 56:60

  fit <- dfspa(d$F_hat[tr, ], K = 3)
  pih <- project_to_simplex(d$F_hat[tr, ], fit$bases)
  wm  <- fit_weight_model(pih, d$W[tr, , drop = FALSE])

  res <- suppressWarnings(conformal_from_fit(
    fit, wm,
    F_cal = d$F_hat[cal, , drop = FALSE],
    W_cal = d$W[cal, , drop = FALSE],
    W_new = d$W[new, , drop = FALSE]
  ))
  expect_equal(dim(res$prediction), c(length(new), ncol(d$F_hat)))
  expect_length(res$quantile, ncol(d$F_hat))
  expect_true(all(res$lower <= res$prediction))
  expect_true(all(res$upper >= res$prediction))
})

test_that("conformal_from_fit rejects basis/weight-model K mismatch", {
  d <- make_toy(m = 50)
  fit2 <- dfspa(d$F_hat[1:30, ], K = 2)
  fit3 <- dfspa(d$F_hat[1:30, ], K = 3)
  pih3 <- project_to_simplex(d$F_hat[1:30, ], fit3$bases)
  wm3  <- fit_weight_model(pih3, d$W[1:30, , drop = FALSE])
  expect_error(
    conformal_from_fit(fit2, wm3,
                       F_cal = d$F_hat[31:50, ], W_cal = d$W[31:50, , drop = FALSE],
                       W_new = d$W[1:2, , drop = FALSE]),
    "mismatch"
  )
})

#' Fit the full MetaHunt pipeline
#'
#' End-to-end convenience wrapper that runs the three training-time steps
#' of the MetaHunt pipeline in sequence:
#' (1) [dfspa()] for basis hunting,
#' (2) [project_to_simplex()] for per-study weight recovery, and
#' (3) [fit_weight_model()] for modelling the weight-to-covariate map.
#' The result supports [predict.metahunt()] for generating target-function
#' predictions on new study-level covariates.
#'
#' For uncertainty quantification, pair a `"metahunt"` fit with
#' [conformal_from_fit()] (requires a separate calibration set). The
#' high-level [split_conformal()] and [cross_conformal()] functions
#' perform their own fitting and do not consume a pre-fit `"metahunt"`
#' object.
#'
#' @param F_hat An `m`-by-`G_grid` numeric matrix of study-level function
#'   evaluations on a shared grid; row `i` is \eqn{\hat f^{(i)}}.
#' @param W An `m`-by-`p` matrix or data frame of study-level covariates.
#' @param K Integer number of basis functions.
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector
#'   defining the \eqn{L^2(\mu)} inner product; defaults to uniform.
#' @param dfspa_args Named list of extra arguments for [dfspa()].
#' @param weight_model_args Named list of extra arguments for
#'   [fit_weight_model()].
#'
#' @return An object of class `"metahunt"`: a list with the `dfspa_fit`,
#'   `weight_model`, training `pi_hat`, `K`, and a stored copy of
#'   `grid_weights`.
#'
#' @seealso [predict.metahunt()], [split_conformal()],
#'   [cross_conformal()], [conformal_from_fit()],
#'   [reconstruction_error_curve()], [cv_error_curve()].
#'
#' @examples
#' set.seed(1)
#' G <- 40; m <- 80
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' fit <- metahunt(F_hat, W, K = 3)
#' fit
#' f_pred <- predict(fit, newdata = W[1:3, ])
#' dim(f_pred)                                  # 3 x G: predicted functions
#' predict(fit, newdata = W[1:3, ], wrapper = mean)  # scalar summaries
#'
#' @export
metahunt <- function(F_hat, W, K,
                     grid_weights = NULL,
                     dfspa_args = list(),
                     weight_model_args = list()) {
  call <- match.call()

  if (!is.matrix(F_hat) || !is.numeric(F_hat)) {
    stop("`F_hat` must be a numeric matrix with studies as rows.")
  }
  if (is.matrix(W)) W <- as.data.frame(W)
  if (!is.data.frame(W)) stop("`W` must be a matrix or data frame.")
  if (nrow(W) != nrow(F_hat)) {
    stop("`W` must have the same number of rows as `F_hat`.")
  }
  if (!is.numeric(K) || length(K) != 1L || K < 1L || K != as.integer(K)) {
    stop("`K` must be a positive integer.")
  }
  K <- as.integer(K)
  if (!is.list(dfspa_args) || !is.list(weight_model_args)) {
    stop("`dfspa_args` and `weight_model_args` must be named lists.")
  }

  fit <- do.call(dfspa,
                 c(list(F_hat = F_hat, K = K,
                        grid_weights = grid_weights), dfspa_args))
  pi_hat <- project_to_simplex(F_hat, fit$bases,
                               grid_weights = grid_weights)
  wm <- do.call(fit_weight_model,
                c(list(pi_hat = pi_hat, W = W), weight_model_args))

  structure(
    list(
      dfspa_fit    = fit,
      weight_model = wm,
      pi_hat       = pi_hat,
      K            = K,
      grid_weights = grid_weights,
      m            = nrow(F_hat),
      G_grid       = ncol(F_hat),
      call         = call
    ),
    class = "metahunt"
  )
}

#' Predict target functions (or scalar summaries) from a MetaHunt fit
#'
#' @param object A `"metahunt"` object from [metahunt()].
#' @param newdata A matrix or data frame of new study-level covariates.
#' @param wrapper Optional reduction function. If `NULL`, returns the full
#'   predicted function matrix (`nrow(newdata)`-by-`G_grid`). If a function,
#'   applied to each predicted function to return a scalar per new target;
#'   see [apply_wrapper()].
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector
#'   used only when `wrapper = NULL` with the default weighted-mean path.
#'   Defaults to the `grid_weights` stored in `object`.
#' @param ... Ignored.
#'
#' @return Either an `nrow(newdata)`-by-`G_grid` matrix of predicted
#'   functions (when `wrapper = NULL`) or a length-`nrow(newdata)` numeric
#'   vector of scalar summaries.
#'
#' @examples
#' set.seed(1)
#' G <- 25; m <- 40
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
#' f_pred <- predict(fit, newdata = W[1:3, ])
#' dim(f_pred)                                       # 3 x G
#' predict(fit, newdata = W[1:3, ], wrapper = mean)  # scalar summaries
#'
#' @export
predict.metahunt <- function(object, newdata, wrapper = NULL,
                             grid_weights = NULL, ...) {
  f_pred <- predict_target(object$dfspa_fit, object$weight_model, newdata)
  if (is.null(wrapper)) return(f_pred)
  gw <- if (is.null(grid_weights)) object$grid_weights else grid_weights
  apply_wrapper(f_pred, wrapper = wrapper, grid_weights = gw)
}

#' Plot recovered basis functions from a MetaHunt fit
#'
#' @param x A `"metahunt"` object from [metahunt()].
#' @param x_axis Optional numeric vector of length `G_grid` giving the
#'   x-axis values. Defaults to `seq_len(G_grid)`.
#' @param ... Passed to [graphics::matplot()].
#'
#' @return Invisibly returns `x`.
#'
#' @examples
#' set.seed(1)
#' G <- 25; m <- 40
#' x_grid <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x_grid), cos(pi * x_grid), x_grid)
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
#' plot(fit)
#' plot(fit, x_axis = x_grid)
#'
#' @export
plot.metahunt <- function(x, x_axis = NULL, ...) {
  bases <- x$dfspa_fit$bases
  K <- nrow(bases); G <- ncol(bases)
  if (is.null(x_axis)) x_axis <- seq_len(G)
  if (length(x_axis) != G) {
    stop("`x_axis` must have length equal to ncol(bases).")
  }
  args <- list(x = x_axis, y = t(bases), type = "l", lty = 1,
               xlab = "grid", ylab = expression(g[k](x)),
               main = sprintf("Recovered basis functions (K = %d)", K))
  user <- list(...)
  args[names(user)] <- user
  do.call(graphics::matplot, args)
  graphics::legend("topright",
                   legend = paste("basis", seq_len(K)),
                   col = seq_len(K), lty = 1, bty = "n")
  invisible(x)
}

#' @export
print.metahunt <- function(x, ...) {
  cat("MetaHunt fit\n")
  cat("  m (studies):   ", x$m, "\n")
  cat("  G (grid size): ", x$G_grid, "\n")
  cat("  K (bases):     ", x$K, "\n")
  cat("  weight method: ", x$weight_model$method, "\n")
  cat("  predictors:    ",
      paste(x$weight_model$predictor_names, collapse = ", "), "\n")
  invisible(x)
}

#' Summarise a MetaHunt fit
#'
#' Produces a compact summary of a `"metahunt"` object, including study/grid
#' sizes, the weight-model method, per-basis summary statistics for the
#' training simplex weights `pi_hat`, and denoising bookkeeping from the
#' underlying [dfspa()] fit.
#'
#' @param object A `"metahunt"` object from [metahunt()].
#' @param ... Ignored.
#'
#' @return An object of class `"summary.metahunt"`: a list with components
#'   \describe{
#'     \item{`m`}{Number of studies.}
#'     \item{`G_grid`}{Grid size.}
#'     \item{`K`}{Number of basis functions.}
#'     \item{`weight_method`}{Method used by the weight model.}
#'     \item{`predictor_names`}{Character vector of covariate names.}
#'     \item{`pi_summary`}{A `K`-by-5 numeric matrix; each row gives `min`,
#'       `mean`, `median`, `max`, and `sd` of the corresponding column of
#'       `object$pi_hat`.}
#'     \item{`n_kept`}{Number of studies retained after denoising.}
#'     \item{`n_dropped`}{Number of studies dropped (`m - n_kept`).}
#'     \item{`denoising`}{List with `N` and `Delta` from the `dfspa` fit
#'       (both `NA` when `denoise = FALSE`).}
#'   }
#'
#' @examples
#' set.seed(1)
#' G <- 25; m <- 40
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
#' summary(fit)
#'
#' @export
summary.metahunt <- function(object, ...) {
  if (!inherits(object, "metahunt")) {
    stop("`object` must be a `metahunt` object.")
  }

  pi_hat <- object$pi_hat
  K <- object$K
  pi_summary <- matrix(NA_real_, nrow = K, ncol = 5L,
                       dimnames = list(paste0("basis_", seq_len(K)),
                                       c("min", "mean", "median",
                                         "max", "sd")))
  for (k in seq_len(K)) {
    col_k <- pi_hat[, k]
    pi_summary[k, ] <- c(min(col_k), mean(col_k), stats::median(col_k),
                         max(col_k), stats::sd(col_k))
  }

  n_kept <- length(object$dfspa_fit$kept)
  n_dropped <- object$m - n_kept

  denoising <- list(
    N     = object$dfspa_fit$N,
    Delta = object$dfspa_fit$Delta
  )

  out <- list(
    m               = object$m,
    G_grid          = object$G_grid,
    K               = K,
    weight_method   = object$weight_model$method,
    predictor_names = object$weight_model$predictor_names,
    pi_summary      = pi_summary,
    n_kept          = n_kept,
    n_dropped       = n_dropped,
    denoising       = denoising
  )
  class(out) <- "summary.metahunt"
  out
}

#' Print a `summary.metahunt` object
#'
#' @param x A `"summary.metahunt"` object from [summary.metahunt()].
#' @param ... Ignored.
#'
#' @return Invisibly returns `x`.
#'
#' @examples
#' set.seed(1)
#' G <- 25; m <- 40
#' x_grid <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x_grid), cos(pi * x_grid), x_grid)
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
#' print(summary(fit))
#'
#' @export
print.summary.metahunt <- function(x, ...) {
  cat("MetaHunt fit summary\n")
  cat("  m (studies):   ", x$m, "\n")
  cat("  G (grid size): ", x$G_grid, "\n")
  cat("  K (bases):     ", x$K, "\n")
  cat("  weight method: ", x$weight_method, "\n")
  cat("  predictors:    ",
      paste(x$predictor_names, collapse = ", "), "\n")
  cat("  studies kept:  ", x$n_kept, "\n")
  cat("  studies dropped:", x$n_dropped, "\n")
  N_val     <- x$denoising$N
  Delta_val <- x$denoising$Delta
  cat("  denoising N:   ",
      if (is.finite(N_val)) signif(N_val, 4) else "NA", "\n")
  cat("  denoising Delta:",
      if (is.finite(Delta_val)) signif(Delta_val, 4) else "NA", "\n")
  cat("\nPer-basis pi_hat summary:\n")
  print(round(x$pi_summary, 4))
  invisible(x)
}

test_that("summary.metahunt returns the right class and fields", {
  set.seed(1)
  G <- 25; m <- 40
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  fit <- metahunt(F_hat, W, K = 3,
                  dfspa_args = list(denoise = FALSE))
  s <- summary(fit)

  expect_s3_class(s, "summary.metahunt")
  expected_fields <- c("m", "G_grid", "K", "weight_method",
                       "predictor_names", "pi_summary",
                       "n_kept", "n_dropped", "denoising")
  expect_true(all(expected_fields %in% names(s)))

  expect_equal(s$m, m)
  expect_equal(s$G_grid, G)
  expect_equal(s$K, 3L)
  expect_equal(s$weight_method, fit$weight_model$method)
  expect_equal(s$predictor_names, fit$weight_model$predictor_names)
  expect_true(is.list(s$denoising))
  expect_true(all(c("N", "Delta") %in% names(s$denoising)))
})

test_that("print.summary.metahunt runs without error", {
  set.seed(2)
  G <- 25; m <- 40
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  fit <- metahunt(F_hat, W, K = 3,
                  dfspa_args = list(denoise = FALSE))
  s <- summary(fit)

  expect_output(print(s), "MetaHunt fit summary")
  expect_output(print(s), "Per-basis pi_hat summary")
})

test_that("pi_summary has shape K x 5 with correct column names", {
  set.seed(3)
  G <- 30; m <- 50
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  fit <- metahunt(F_hat, W, K = 3,
                  dfspa_args = list(denoise = FALSE))
  s <- summary(fit)

  expect_true(is.matrix(s$pi_summary))
  expect_equal(dim(s$pi_summary), c(3L, 5L))
  expect_equal(colnames(s$pi_summary),
               c("min", "mean", "median", "max", "sd"))
})

test_that("n_kept + n_dropped equals m", {
  set.seed(4)
  G <- 25; m <- 40
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  fit <- metahunt(F_hat, W, K = 3,
                  dfspa_args = list(denoise = FALSE))
  s <- summary(fit)

  expect_equal(s$n_kept + s$n_dropped, s$m)
})

test_that("coef.metahunt_weight_model returns a non-empty numeric object", {
  set.seed(5)
  G <- 25; m <- 50
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  fit <- metahunt(F_hat, W, K = 3,
                  dfspa_args = list(denoise = FALSE))
  cf <- coef(fit$weight_model)

  # DirichletReg::coef returns a list of numeric vectors (one per category
  # in the "common" parametrization), or a numeric vector / matrix in
  # other parametrizations.
  expect_true(is.numeric(cf) || is.matrix(cf) || is.list(cf))
  expect_true(length(cf) > 0L)
  expect_true(all(is.finite(unlist(cf))))
})

test_that("plot.metahunt runs without error", {
  set.seed(1)
  G <- 20; m <- 30
  F_hat <- matrix(rnorm(m * G), m, G)
  W <- data.frame(w1 = rnorm(m))
  fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
  pdf(NULL)
  on.exit(dev.off())
  expect_invisible(plot(fit))
  expect_invisible(plot(fit, x_axis = seq(0, 1, length.out = G)))
  expect_error(plot(fit, x_axis = 1:5), "length")
})

test_that("plot.metahunt_conformal runs for pointwise and scalar objects", {
  set.seed(2)
  G <- 25; m <- 100
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.5), c(-0.4, 1), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  res_pw <- split_conformal(F_hat, W, W[1:3, ], K = 3, seed = 1)
  res_sc <- split_conformal(F_hat, W, W[1:3, ], K = 3,
                            wrapper = mean, seed = 1)
  pdf(NULL)
  on.exit(dev.off())
  expect_invisible(plot(res_pw))
  expect_invisible(plot(res_pw, target_idx = 2))
  expect_invisible(plot(res_sc))
  expect_error(plot(res_pw, target_idx = 99), "out of range")
})

#' Minimax-regret aggregator for multisite function-valued estimands
#'
#' Implements the minimax-regret estimator of Zhang, Huang, and Imai
#' (\href{https://arxiv.org/abs/2412.11136}{arXiv:2412.11136}) for
#' aggregating site-level function estimates. Given study-level functions
#' \eqn{\hat f^{(i)}}, the estimator is
#' \deqn{\hat q = \arg\min_{q \in \Delta_{m-1}}\
#'   q^\top \hat\Gamma\, q - \hat d^\top q,
#'   \qquad
#'   \hat\Gamma_{ij} = \sum_g w_g\, \hat f^{(i)}(x_g)\hat f^{(j)}(x_g),\ \
#'   \hat d_i = \sum_g w_g\, (\hat f^{(i)}(x_g))^2,}
#' yielding the predicted target function
#' \eqn{\tilde f(x) = \sum_{i=1}^m \hat q_i\, \hat f^{(i)}(x)}.
#' Unlike [metahunt()], this method does not use study-level covariates;
#' the target is the worst-case-regret aggregator over the convex hull of
#' source functions.
#'
#' @details
#' The simplex-constrained QP is solved with [quadprog::solve.QP()].
#' A small ridge is added to \eqn{\hat\Gamma} for numerical stability when
#' source functions are highly collinear. The resulting `q` is clipped to
#' be non-negative and renormalised to sum to 1 to absorb floating-point
#' drift.
#'
#' @param F_hat An `m`-by-`G_grid` numeric matrix of source-site function
#'   evaluations on a shared grid; row `i` is \eqn{\hat f^{(i)}}.
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector
#'   defining the target measure used to compute \eqn{\hat\Gamma} and
#'   \eqn{\hat d}. Defaults to uniform weights `1 / G_grid`.
#' @param ridge Non-negative scalar; replaces \eqn{\hat\Gamma} with
#'   \eqn{\hat\Gamma + \text{ridge}\cdot I} for numerical stability.
#'   Defaults to `1e-10`.
#' @param wrapper Optional reduction function (see [apply_wrapper()]). If
#'   `NULL`, `prediction` is the length-`G_grid` predicted target function;
#'   if a function, `prediction` is the scalar `wrapper(prediction)`
#'   (e.g. an ATE when `wrapper = mean` and rows of `F_hat` are CATE
#'   evaluations).
#'
#' @return An object of class `"minmax_regret"`: a list with
#'   \describe{
#'     \item{`prediction`}{Predicted target. Length-`G_grid` vector when
#'       `wrapper = NULL`; scalar otherwise.}
#'     \item{`q`}{Length-`m` simplex weights from the minimax-regret QP.}
#'     \item{`Gamma`}{The `m`-by-`m` Gram matrix used (post-ridge).}
#'     \item{`d`}{Length-`m` linear coefficient vector.}
#'     \item{`grid_weights`}{Grid weights used.}
#'     \item{`ridge`}{Ridge value used.}
#'     \item{`wrapper`}{Wrapper function or `NULL`.}
#'   }
#'
#' @references
#' Zhang, Y., Huang, M., and Imai, K. (2024). Minimax regret estimation
#' for generalizing heterogeneous treatment effects with multisite data.
#' \href{https://arxiv.org/abs/2412.11136}{arXiv:2412.11136}.
#'
#' @examples
#' set.seed(1)
#' G <- 30; m <- 6
#' x <- seq(0, 1, length.out = G)
#' F_hat <- rbind(
#'   sin(pi * x),
#'   cos(pi * x),
#'   x,
#'   0.5 * sin(pi * x) + 0.5 * x,
#'   0.3 * cos(pi * x) + 0.7 * x,
#'   0.4 * sin(pi * x) + 0.4 * cos(pi * x) + 0.2 * x
#' )
#'
#' fit <- minmax_regret(F_hat)
#' fit$q                                        # simplex weights over sources
#' length(fit$prediction)                       # G: predicted target function
#'
#' # ATE-style scalar via wrapper
#' minmax_regret(F_hat, wrapper = mean)$prediction
#'
#' @export
minmax_regret <- function(F_hat, grid_weights = NULL,
                          ridge = 1e-10, wrapper = NULL) {
  if (!is.matrix(F_hat) || !is.numeric(F_hat)) {
    stop("`F_hat` must be a numeric matrix with sources as rows.")
  }
  if (anyNA(F_hat)) stop("`F_hat` must not contain NA or NaN values.")
  m <- nrow(F_hat); G_grid <- ncol(F_hat)
  if (m < 1L || G_grid < 1L) stop("`F_hat` must be non-empty.")
  if (!is.numeric(ridge) || length(ridge) != 1L ||
      ridge < 0 || !is.finite(ridge)) {
    stop("`ridge` must be a non-negative finite scalar.")
  }

  if (is.null(grid_weights)) {
    grid_weights <- rep(1 / G_grid, G_grid)
  } else {
    if (!is.numeric(grid_weights) || length(grid_weights) != G_grid ||
        any(!is.finite(grid_weights)) || any(grid_weights < 0)) {
      stop("`grid_weights` must be a length-`ncol(F_hat)` non-negative numeric vector.")
    }
    if (sum(grid_weights) <= 0) {
      stop("`grid_weights` must have strictly positive total mass.")
    }
  }

  # Gram matrix: Gamma[i,j] = sum_g w_g F[i,g] F[j,g]
  Fw    <- sweep(F_hat, 2, grid_weights, `*`)        # m x G
  Gamma <- tcrossprod(Fw, F_hat)                     # m x m
  Gamma <- (Gamma + t(Gamma)) / 2                    # symmetrise
  d_vec <- rowSums(F_hat * Fw)                       # m

  # Single-source short circuit
  if (m == 1L) {
    q <- 1
    pred <- as.numeric(F_hat[1, ])
  } else {
    Dmat <- 2 * (Gamma + ridge * diag(m))
    Amat <- cbind(rep(1, m), diag(m))
    bvec <- c(1, rep(0, m))
    sol  <- quadprog::solve.QP(Dmat = Dmat, dvec = d_vec,
                               Amat = Amat, bvec = bvec, meq = 1L)
    q <- pmax(sol$solution, 0)
    if (sum(q) <= 0) stop("Solver returned an all-zero weight vector.")
    q <- q / sum(q)
    pred <- as.numeric(crossprod(q, F_hat))
  }

  if (!is.null(wrapper)) {
    if (!is.function(wrapper)) stop("`wrapper` must be a function or NULL.")
    pred <- apply_wrapper(matrix(pred, nrow = 1L),
                          wrapper = wrapper,
                          grid_weights = grid_weights)
    pred <- as.numeric(pred)
  }

  structure(
    list(
      prediction   = pred,
      q            = q,
      Gamma        = Gamma + ridge * diag(m),
      d            = d_vec,
      grid_weights = grid_weights,
      ridge        = ridge,
      wrapper      = wrapper
    ),
    class = "minmax_regret"
  )
}

#' @export
print.minmax_regret <- function(x, ...) {
  cat("Minimax-regret aggregator (Zhang, Huang & Imai 2024)\n")
  cat("  m (sources):  ", length(x$q), "\n")
  cat("  G (grid size):", length(x$grid_weights), "\n")
  cat("  ridge:        ", signif(x$ridge, 4), "\n")
  if (is.null(x$wrapper)) {
    cat("  prediction:    function on grid (length", length(x$prediction), ")\n")
  } else {
    cat("  prediction:    scalar =", signif(x$prediction, 4), "\n")
  }
  ord <- order(x$q, decreasing = TRUE)
  top <- min(5L, length(x$q))
  cat("  top weights:\n")
  for (k in seq_len(top)) {
    cat(sprintf("    source %d: q = %.4f\n", ord[k], x$q[ord[k]]))
  }
  invisible(x)
}

test_that("minmax_regret returns a simplex weight and a function on the grid", {
  set.seed(1)
  G <- 30; m <- 5
  x <- seq(0, 1, length.out = G)
  F_hat <- rbind(
    sin(pi * x),
    cos(pi * x),
    x,
    0.4 * sin(pi * x) + 0.6 * x,
    0.3 * cos(pi * x) + 0.7 * x
  )
  fit <- minmax_regret(F_hat)

  expect_s3_class(fit, "minmax_regret")
  expect_length(fit$q, m)
  expect_equal(sum(fit$q), 1, tolerance = 1e-8)
  expect_true(all(fit$q >= -1e-10))
  expect_length(fit$prediction, G)

  # prediction is exactly q^T F_hat
  expect_equal(fit$prediction,
               as.numeric(crossprod(fit$q, F_hat)),
               tolerance = 1e-8)
})

test_that("minmax_regret with one source returns it unchanged", {
  G <- 20
  f1 <- sin(seq(0, 2 * pi, length.out = G))
  fit <- minmax_regret(matrix(f1, nrow = 1))
  expect_equal(fit$q, 1)
  expect_equal(fit$prediction, f1)
})

test_that("minmax_regret wrapper returns scalar matching grid mean", {
  G <- 25; m <- 4
  set.seed(2)
  F_hat <- matrix(rnorm(m * G), m, G)
  fit_fun <- minmax_regret(F_hat)
  fit_avg <- minmax_regret(F_hat, wrapper = mean)
  expect_length(fit_avg$prediction, 1L)
  expect_equal(fit_avg$prediction, mean(fit_fun$prediction), tolerance = 1e-10)
})

test_that("minmax_regret achieves zero objective when target is in the convex hull", {
  set.seed(3)
  G <- 20; m <- 4
  x <- seq(0, 1, length.out = G)
  F_hat <- rbind(sin(pi * x), cos(pi * x), x, 0.3 * x + 0.7 * sin(pi * x))
  # the 4th row is in the convex hull of rows 1 and 3, so q should put
  # mass on the row representing the worst-case-regret minimiser.
  fit <- minmax_regret(F_hat)
  expect_equal(sum(fit$q), 1, tolerance = 1e-8)
  # the QP objective at the solution should be finite and not larger than
  # the value at any vertex
  Q_sol <- as.numeric(crossprod(fit$q, fit$Gamma) %*% fit$q -
                        crossprod(fit$d, fit$q))
  for (k in seq_len(m)) {
    e_k <- numeric(m); e_k[k] <- 1
    Q_k <- as.numeric(crossprod(e_k, fit$Gamma) %*% e_k -
                        crossprod(fit$d, e_k))
    expect_lte(Q_sol, Q_k + 1e-8)
  }
})

test_that("minmax_regret rejects malformed inputs", {
  expect_error(minmax_regret(as.data.frame(matrix(1:6, 2, 3))),
               "numeric matrix")
  expect_error(minmax_regret(matrix(c(1, NA, 2, 3), 2, 2)),
               "NA or NaN")
  expect_error(minmax_regret(matrix(1:6, 2, 3), grid_weights = c(1, -1, 1)),
               "non-negative")
  expect_error(minmax_regret(matrix(1:6, 2, 3), ridge = -1),
               "non-negative")
})

test_that("print.minmax_regret runs without error", {
  G <- 20
  F_hat <- matrix(rnorm(60), 3, 20)
  fit <- minmax_regret(F_hat)
  expect_output(print(fit), "Minimax-regret aggregator")
})

#' @keywords internal
#' @noRd
.row_l2 <- function(F_mat, grid_weights = NULL) {
  if (is.null(grid_weights)) {
    grid_weights <- rep(1 / ncol(F_mat), ncol(F_mat))
  }
  sqrt(as.numeric((F_mat^2) %*% grid_weights))
}

#' Reconstruction-error curve for basis-rank selection
#'
#' For each candidate number of bases `K`, run [dfspa()] followed by
#' [project_to_simplex()] and report the average projection residual
#' \deqn{\mathcal E(K) = \frac{1}{m}\sum_{i=1}^m
#'        \left\|\hat f^{(i)} - \sum_{k=1}^K \hat\pi_{ik}\hat g_k\right\|_{L^2(\mu)}.}
#' Plotting `error` against `K` typically shows an elbow.
#'
#' @details
#' This is the unsupervised rank-selection criterion of Section 3.2 of the
#' paper (Equation for \eqn{\mathcal E(K)}). It does not require
#' study-level covariates.
#'
#' @param F_hat An `m`-by-`G_grid` numeric matrix of study-level function
#'   evaluations on the shared grid.
#' @param K_range Integer vector of candidate `K` values. Defaults to
#'   `2:min(nrow(F_hat) - 1, 10)`.
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector
#'   used for the \eqn{L^2(\mu)} norm; defaults to uniform.
#' @param dfspa_args Named list of extra arguments passed to [dfspa()],
#'   e.g. `list(denoise = FALSE, N = 2)`.
#'
#' @return A data frame with columns `K` (integer) and `error` (numeric).
#'   Rows where `dfspa()` or the projection fails are reported with
#'   `error = NA` and a single warning summarising the failures.
#'
#' @examples
#' set.seed(1)
#' G <- 40
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' m <- 50
#' pi_mat <- matrix(stats::rgamma(m * 3, shape = 0.5), m, 3)
#' pi_mat <- pi_mat / rowSums(pi_mat)
#' F_hat  <- pi_mat %*% basis + matrix(stats::rnorm(m * G, sd = 0.02), m, G)
#'
#' elbow <- reconstruction_error_curve(F_hat, K_range = 2:6)
#' elbow
#'
#' @export
reconstruction_error_curve <- function(F_hat, K_range = NULL,
                                       grid_weights = NULL,
                                       dfspa_args = list()) {
  if (!is.matrix(F_hat) || !is.numeric(F_hat)) {
    stop("`F_hat` must be a numeric matrix with studies as rows.")
  }
  m <- nrow(F_hat)
  if (is.null(K_range)) {
    K_range <- 2L:min(m - 1L, 10L)
  }
  K_range <- as.integer(K_range)
  if (any(K_range < 1L) || any(K_range > m)) {
    stop("`K_range` values must lie in 1:nrow(F_hat).")
  }
  if (!is.list(dfspa_args)) stop("`dfspa_args` must be a named list.")

  errs <- rep(NA_real_, length(K_range))
  failures <- character(0)
  for (j in seq_along(K_range)) {
    K <- K_range[j]
    fit <- tryCatch(
      do.call(dfspa, c(list(F_hat = F_hat, K = K,
                            grid_weights = grid_weights), dfspa_args)),
      error = function(e) e
    )
    if (inherits(fit, "error")) {
      failures <- c(failures, sprintf("K = %d: dfspa: %s", K, conditionMessage(fit)))
      next
    }
    pi_hat <- tryCatch(
      project_to_simplex(F_hat, fit$bases, grid_weights = grid_weights),
      error = function(e) e
    )
    if (inherits(pi_hat, "error")) {
      failures <- c(failures, sprintf("K = %d: projection: %s", K, conditionMessage(pi_hat)))
      next
    }
    resid <- F_hat - pi_hat %*% fit$bases
    errs[j] <- mean(.row_l2(resid, grid_weights))
  }

  if (length(failures)) {
    warning(sprintf("%d of %d K values failed:\n%s",
                    length(failures), length(K_range),
                    paste0("  ", failures, collapse = "\n")))
  }

  data.frame(K = K_range, error = errs)
}

#' Cross-validated prediction-error curve for basis-rank selection
#'
#' For each candidate `K`, perform k-fold cross-validation at the study
#' level. Within each fold, the full MetaHunt pipeline (d-fSPA + constrained
#' projection + weight model) is refit on the training studies, and the
#' held-out studies' functions are predicted from their study-level
#' covariates. The prediction error is
#' \eqn{\|\hat f^{(i)} - \tilde f^{(i)}\|_{L^2(\mu)}} averaged over
#' held-out studies and then over folds.
#'
#' @details
#' This is the supervised rank-selection criterion of Section 3.2 of the
#' paper. Each held-out study is excluded from both basis hunting and
#' weight-model fitting.
#'
#' @param F_hat An `m`-by-`G_grid` numeric matrix of study-level function
#'   evaluations.
#' @param W An `m`-by-`p` matrix or data frame of study-level covariates.
#' @param K_range Integer vector of candidate `K` values. Defaults to
#'   `2:min(nrow(F_hat) - 1, 10)`.
#' @param n_folds Integer number of CV folds (default `5`).
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector.
#' @param dfspa_args Named list of extra arguments for [dfspa()].
#' @param weight_model_args Named list of extra arguments for
#'   [fit_weight_model()].
#' @param seed Optional integer seed for reproducible fold assignment; if
#'   `NULL` no seeding is performed.
#'
#' @return A data frame with columns `K`, `cv_error` (mean over folds),
#'   and `cv_se` (standard error across folds). The per-fold error matrix
#'   is attached as the attribute `"fold_errors"`
#'   (`length(K_range)`-by-`n_folds`). Folds where the pipeline fails
#'   contribute `NA` and are summarised in a single warning.
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' G <- 40; m <- 80; K_true <- 3
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.02), m, G)
#' cv <- cv_error_curve(F_hat, W, K_range = 2:5, n_folds = 4, seed = 1)
#' cv
#' }
#'
#' @export
cv_error_curve <- function(F_hat, W, K_range = NULL, n_folds = 5L,
                           grid_weights = NULL,
                           dfspa_args = list(),
                           weight_model_args = list(),
                           seed = NULL) {
  if (!is.matrix(F_hat) || !is.numeric(F_hat)) {
    stop("`F_hat` must be a numeric matrix with studies as rows.")
  }
  m <- nrow(F_hat)
  if (is.matrix(W)) W <- as.data.frame(W)
  if (!is.data.frame(W)) stop("`W` must be a matrix or data frame.")
  if (nrow(W) != m) stop("`W` must have the same number of rows as `F_hat`.")
  if (!is.list(dfspa_args) || !is.list(weight_model_args)) {
    stop("`dfspa_args` and `weight_model_args` must be named lists.")
  }
  if (!is.numeric(n_folds) || length(n_folds) != 1L ||
      n_folds < 2L || n_folds != as.integer(n_folds)) {
    stop("`n_folds` must be an integer >= 2.")
  }
  n_folds <- as.integer(n_folds)
  if (n_folds > m) stop("`n_folds` cannot exceed the number of studies.")

  if (is.null(K_range)) {
    K_range <- 2L:min(m - 1L, 10L)
  }
  K_range <- as.integer(K_range)

  if (!is.null(seed)) withr::local_seed(seed)
  fold_id <- sample(rep(seq_len(n_folds), length.out = m))

  fold_errors <- matrix(NA_real_, nrow = length(K_range), ncol = n_folds,
                        dimnames = list(NULL, paste0("fold", seq_len(n_folds))))
  failures <- character(0)

  for (j in seq_along(K_range)) {
    K <- K_range[j]
    for (f in seq_len(n_folds)) {
      train <- which(fold_id != f)
      test  <- which(fold_id == f)
      result <- tryCatch({
        fit_f <- do.call(dfspa,
                         c(list(F_hat = F_hat[train, , drop = FALSE], K = K,
                                grid_weights = grid_weights), dfspa_args))
        pi_tr <- project_to_simplex(F_hat[train, , drop = FALSE],
                                    fit_f$bases, grid_weights = grid_weights)
        wm <- do.call(fit_weight_model,
                      c(list(pi_hat = pi_tr, W = W[train, , drop = FALSE]),
                        weight_model_args))
        f_pred <- predict_target(fit_f, wm, W[test, , drop = FALSE])
        mean(.row_l2(F_hat[test, , drop = FALSE] - f_pred, grid_weights))
      }, error = function(e) e)
      if (inherits(result, "error")) {
        failures <- c(failures, sprintf("K = %d, fold %d: %s",
                                        K, f, conditionMessage(result)))
      } else {
        fold_errors[j, f] <- result
      }
    }
  }

  if (length(failures)) {
    warning(sprintf("%d (K, fold) combinations failed:\n%s",
                    length(failures),
                    paste0("  ", failures, collapse = "\n")))
  }

  n_ok     <- rowSums(!is.na(fold_errors))
  cv_error <- ifelse(n_ok > 0L, rowMeans(fold_errors, na.rm = TRUE), NA_real_)
  cv_sd    <- ifelse(n_ok > 1L, apply(fold_errors, 1, stats::sd, na.rm = TRUE),
                     NA_real_)
  cv_se    <- ifelse(n_ok > 1L, cv_sd / sqrt(n_ok), NA_real_)

  if (any(n_ok == 0L)) {
    warning(sprintf(
      "K = %s had no successful folds; cv_error and cv_se reported as NA.",
      paste(K_range[n_ok == 0L], collapse = ", ")
    ), call. = FALSE)
  }

  out <- data.frame(K = K_range, cv_error = cv_error, cv_se = cv_se)
  attr(out, "fold_errors") <- fold_errors
  out
}

test_that("reconstruction_error_curve is non-increasing for noiseless rank-K data", {
  set.seed(1)
  G <- 40; K_true <- 3; m <- 30
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  pi_mat <- matrix(stats::rgamma(m * K_true, 0.5), m, K_true)
  pi_mat <- pi_mat / rowSums(pi_mat)
  # include near-pure studies to guarantee basis recovery
  pi_mat[1:3, ] <- diag(K_true)
  F_hat <- pi_mat %*% basis

  res <- reconstruction_error_curve(F_hat, K_range = 2:K_true,
                                    dfspa_args = list(denoise = FALSE))

  expect_s3_class(res, "data.frame")
  expect_named(res, c("K", "error"))
  expect_equal(res$K, 2:K_true)
  expect_true(all(diff(res$error) <= 1e-8))
  # at K = K_true, error should be ~0 in the noiseless case
  expect_lt(res$error[res$K == K_true], 1e-6)
})

test_that("reconstruction_error_curve respects dfspa_args and grid_weights", {
  set.seed(2)
  G <- 20; m <- 15
  F_hat <- matrix(stats::rnorm(m * G), m, G)
  w <- runif(G, min = 0.1)

  res <- reconstruction_error_curve(
    F_hat, K_range = 2:4, grid_weights = w,
    dfspa_args = list(denoise = FALSE)
  )
  expect_true(all(is.finite(res$error)))
  expect_equal(nrow(res), 3L)
})

test_that("cv_error_curve returns the expected structure", {
  set.seed(3)
  G <- 30; m <- 40; K_true <- 3
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(0.8, -0.5), c(-0.3, 0.9), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.02), m, G)

  res <- cv_error_curve(F_hat, W, K_range = 2:4, n_folds = 4L,
                       dfspa_args = list(denoise = FALSE), seed = 42)

  expect_s3_class(res, "data.frame")
  expect_named(res, c("K", "cv_error", "cv_se"))
  expect_equal(res$K, 2:4)
  fe <- attr(res, "fold_errors")
  expect_equal(dim(fe), c(3L, 4L))
  expect_true(all(is.finite(res$cv_error)))
})

test_that("cv_error_curve is reproducible with a seed", {
  set.seed(4)
  G <- 20; m <- 30
  F_hat <- matrix(rnorm(m * G), m, G)
  W <- data.frame(w1 = rnorm(m))

  a <- cv_error_curve(F_hat, W, K_range = 2:3, n_folds = 3L,
                      dfspa_args = list(denoise = FALSE), seed = 99)
  b <- cv_error_curve(F_hat, W, K_range = 2:3, n_folds = 3L,
                      dfspa_args = list(denoise = FALSE), seed = 99)
  expect_equal(a, b)
})

test_that("cv_error_curve surfaces failures as NA with a warning", {
  set.seed(5)
  G <- 20; m <- 10
  F_hat <- matrix(rnorm(m * G), m, G)
  W <- data.frame(w1 = rnorm(m))

  # Force failures by asking for K too large relative to fold training size
  expect_warning(
    res <- cv_error_curve(F_hat, W, K_range = 9:10, n_folds = 2L,
                          dfspa_args = list(denoise = FALSE), seed = 1),
    "failed"
  )
  expect_true(anyNA(res$cv_error) || anyNA(attr(res, "fold_errors")))
})

#' MetaHunt: Privacy-Preserving Meta-Analysis via Low-Rank Basis Hunting
#'
#' Tools for privacy-preserving meta-analysis of function-valued quantities
#' (e.g. regression, conditional average treatment effect functions) across
#' heterogeneous studies. Implements the MetaHunt pipeline of Shi, Imai,
#' and Zhang: denoised functional Successive Projection Algorithm (d-fSPA)
#' for basis hunting, constrained projection of study functions onto the
#' recovered simplex, Dirichlet regression of mixing weights on study-level
#' covariates, target prediction, and split or cross conformal prediction
#' intervals.
#'
#' @section Main entry points:
#' * [metahunt()] — fit the full pipeline end-to-end.
#' * [predict.metahunt()] — predict target functions for new study-level covariates.
#' * [split_conformal()], [cross_conformal()], [conformal_from_fit()] —
#'   prediction intervals.
#' * [minmax_regret()] — covariate-free worst-case-regret aggregator
#'   (Zhang, Huang, and Imai 2024) as a baseline.
#' * [f_hat_from_models()], [build_grid()] — onramp from fitted-model
#'   lists and reference data to the package's matrix inputs.
#'
#' @section Pipeline building blocks:
#' [dfspa()], [project_to_simplex()], [fit_weight_model()],
#' [predict_target()], [apply_wrapper()],
#' [reconstruction_error_curve()], [cv_error_curve()],
#' [select_denoising_params()].
#'
#' @keywords internal
"_PACKAGE"

#' @keywords internal
#' @noRd
.conformal_quantile <- function(scores, alpha) {
  scores <- scores[is.finite(scores)]
  n <- length(scores)
  if (n == 0L) return(Inf)
  k <- ceiling((1 - alpha) * (n + 1))
  if (k > n) return(Inf)
  sort(scores)[k]
}

#' @keywords internal
#' @noRd
.run_pipeline <- function(F_hat_tr, W_tr, K, grid_weights,
                          dfspa_args, weight_model_args) {
  fit <- do.call(dfspa,
                 c(list(F_hat = F_hat_tr, K = K,
                        grid_weights = grid_weights), dfspa_args))
  pi_tr <- project_to_simplex(F_hat_tr, fit$bases,
                              grid_weights = grid_weights)
  wm <- do.call(fit_weight_model,
                c(list(pi_hat = pi_tr, W = W_tr), weight_model_args))
  list(fit = fit, wm = wm)
}

#' Split conformal intervals from a pre-fit MetaHunt pipeline
#'
#' Lower-level entry point that builds split conformal intervals for new
#' target covariates using an already-fitted d-fSPA basis decomposition,
#' an already-fitted weight model, and a user-supplied calibration set.
#' Use this when you have independently tuned `K` or want to reuse a
#' pipeline fit; otherwise the high-level [split_conformal()] is usually
#' more convenient.
#'
#' @param dfspa_fit A `"dfspa"` object from [dfspa()].
#' @param weight_model A `"metahunt_weight_model"` object from
#'   [fit_weight_model()]. Must satisfy
#'   `weight_model$K == nrow(dfspa_fit$bases)`.
#' @param F_cal An `n_cal`-by-`G_grid` numeric matrix of observed study-level
#'   function evaluations for the calibration set. Calibration studies must
#'   be independent of both `dfspa_fit` and `weight_model` for valid
#'   coverage.
#' @param W_cal A matrix or data frame of study-level covariates for the
#'   calibration set, with the same columns used to fit `weight_model`.
#' @param W_new A matrix or data frame of study-level covariates for new
#'   target studies.
#' @param alpha Miscoverage level; default `0.05`.
#' @param wrapper Optional reduction function (see [apply_wrapper()]). If
#'   `NULL`, intervals are pointwise at every grid point.
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector
#'   used for the default wrapper and for weighted inner products.
#'
#' @return An object of class `"metahunt_conformal"`; see [split_conformal()]
#'   for a description of its fields.
#'
#' @seealso [split_conformal()] for the high-level version that splits and
#'   fits internally, [cross_conformal()] for the K-fold variant.
#'
#' @examples
#' set.seed(1)
#' G <- 30; m <- 80
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m))
#' eta <- cbind(0.8 * W$w1, -0.4 * W$w1, rep(0, m))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' # user-controlled split and fit
#' tr  <- 1:50; cal <- 51:70; new <- 71:80
#' fit <- dfspa(F_hat[tr, ], K = 3)
#' pih <- project_to_simplex(F_hat[tr, ], fit$bases)
#' wm  <- fit_weight_model(pih, W[tr, , drop = FALSE])
#'
#' res <- conformal_from_fit(
#'   fit, wm,
#'   F_cal = F_hat[cal, ], W_cal = W[cal, , drop = FALSE],
#'   W_new = W[new, , drop = FALSE],
#'   wrapper = mean
#' )
#' res
#'
#' @export
conformal_from_fit <- function(dfspa_fit, weight_model,
                               F_cal, W_cal, W_new,
                               alpha = 0.05,
                               wrapper = NULL,
                               grid_weights = NULL) {
  if (!inherits(dfspa_fit, "dfspa")) {
    stop("`dfspa_fit` must be a `dfspa` object (from `dfspa()`).")
  }
  if (!inherits(weight_model, "metahunt_weight_model")) {
    stop("`weight_model` must be a `metahunt_weight_model` object.")
  }
  if (weight_model$K != nrow(dfspa_fit$bases)) {
    stop(sprintf(
      "Basis/weight-model mismatch: dfspa_fit has K = %d bases but weight_model has K = %d.",
      nrow(dfspa_fit$bases), weight_model$K))
  }
  if (!is.matrix(F_cal) || !is.numeric(F_cal)) {
    stop("`F_cal` must be a numeric matrix with calibration studies as rows.")
  }
  if (ncol(F_cal) != ncol(dfspa_fit$bases)) {
    stop("`F_cal` must have the same number of columns as `dfspa_fit$bases`.")
  }
  if (is.matrix(W_cal)) W_cal <- as.data.frame(W_cal)
  if (!is.data.frame(W_cal)) stop("`W_cal` must be a matrix or data frame.")
  if (nrow(W_cal) != nrow(F_cal)) {
    stop("`W_cal` and `F_cal` must have the same number of rows.")
  }
  if (nrow(F_cal) < 2L) {
    stop("Calibration set must have at least 2 studies.")
  }
  if (is.matrix(W_new)) W_new <- as.data.frame(W_new)
  if (!is.data.frame(W_new)) stop("`W_new` must be a matrix or data frame.")
  if (!is.numeric(alpha) || length(alpha) != 1L || alpha <= 0 || alpha >= 1) {
    stop("`alpha` must be a single number in (0, 1).")
  }

  pred_cal <- predict_target(dfspa_fit, weight_model, W_cal)

  .build_conformal_output(
    obs_cal = F_cal, pred_cal = pred_cal,
    pipe = list(fit = dfspa_fit, wm = weight_model),
    W_new = W_new,
    alpha = alpha, wrapper = wrapper, grid_weights = grid_weights,
    method = "split",
    n_cal_label = nrow(F_cal)
  )
}

#' Split conformal prediction intervals for target-function predictions
#'
#' Implements Algorithm 2 of the paper (split conformal prediction) over
#' the MetaHunt pipeline. Studies are partitioned into training and
#' calibration sets. The training set is used to fit d-fSPA, the constrained
#' projection, and the weight model; the calibration set supplies
#' conformity scores, which determine the width of the intervals.
#'
#' @details
#' Given a target function, one can either construct intervals **pointwise
#' at every grid point** (when `wrapper = NULL`) or for a **scalar summary**
#' of the target function (when `wrapper` is a function).
#'
#' * **Pointwise** (`wrapper = NULL`): for each grid point `g` the
#'   conformity score is \eqn{R_{i,g} = |\hat f^{(i)}(x_g) - \tilde
#'   f^{(i)}(x_g)|} across calibration studies `i`. A separate
#'   \eqn{(1-\alpha)}-quantile \eqn{q_g} is computed per grid point, and
#'   the interval at grid point `g` for target `j` is
#'   \eqn{[\tilde f^{(j)}(x_g)-q_g,\ \tilde f^{(j)}(x_g)+q_g]}.
#' * **Scalar** (`wrapper` supplied): conformity scores are
#'   \eqn{R_i = |s(\hat f^{(i)}) - s(\tilde f^{(i)})|} with
#'   `s = wrapper`, and the interval for each target is
#'   \eqn{[s(\tilde f^{(j)})-q,\ s(\tilde f^{(j)})+q]} with a single
#'   shared quantile \eqn{q}.
#'
#' The finite-sample quantile is
#' \eqn{q = R_{(k)}} with \eqn{k = \lceil(1-\alpha)(n_\mathrm{cal}+1)\rceil};
#' if \eqn{k > n_\mathrm{cal}}, `q = Inf` and intervals are
#' \eqn{(-\infty, \infty)}.
#'
#' @param F_hat An `m`-by-`G_grid` numeric matrix of study-level function
#'   evaluations.
#' @param W An `m`-by-`p` matrix or data frame of study-level covariates.
#' @param W_new A matrix or data frame of new target covariates. Must
#'   contain columns matching `W`.
#' @param K Integer number of basis functions.
#' @param alpha Miscoverage level; interval has nominal coverage
#'   \eqn{1-\alpha}. Default `0.05`.
#' @param cal_frac Numeric in `(0, 1)` giving the fraction of studies in
#'   the calibration set. Default `0.3`. Ignored if `calibration_idx` is
#'   supplied.
#' @param wrapper Optional reduction function (see [apply_wrapper()]). If
#'   `NULL`, intervals are constructed pointwise at every grid point.
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector
#'   used for the \eqn{L^2(\mu)} norm and for the default wrapper.
#' @param calibration_idx Optional integer vector of row indices in `F_hat`
#'   to use as the calibration set. If supplied, `cal_frac` is ignored.
#' @param dfspa_args,weight_model_args Named lists passed to [dfspa()] and
#'   [fit_weight_model()] respectively.
#' @param seed Optional integer seed for reproducible train/calibration
#'   splits.
#'
#' @return An object of class `"metahunt_conformal"`: a list with
#'   \describe{
#'     \item{`prediction`}{Point predictions for `W_new`. A numeric vector
#'       of length `nrow(W_new)` in the scalar case, or an
#'       `nrow(W_new)`-by-`G_grid` matrix in the pointwise case.}
#'     \item{`lower`, `upper`}{Interval endpoints, same shape as
#'       `prediction`.}
#'     \item{`alpha`}{Miscoverage level used.}
#'     \item{`method`}{`"split"`.}
#'     \item{`n_cal`}{Calibration sample size.}
#'     \item{`quantile`}{The conformal quantile: a scalar (scalar case)
#'       or a length-`G_grid` vector (pointwise case).}
#'     \item{`wrapper`}{The wrapper used, or `NULL`.}
#'   }
#'
#' @examples
#' set.seed(1)
#' G <- 40; m <- 80; K_true <- 3
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' W_new <- data.frame(w1 = c(0, 1), w2 = c(0, -1))
#' # pointwise intervals at every grid point
#' pi_grid <- split_conformal(F_hat, W, W_new, K = 3, seed = 1)
#' dim(pi_grid$lower)  # 2 x 40
#' # scalar intervals for the grid-weighted mean (ATE-style)
#' pi_ate  <- split_conformal(F_hat, W, W_new, K = 3, wrapper = mean, seed = 1)
#' pi_ate$prediction
#'
#' @export
split_conformal <- function(F_hat, W, W_new, K,
                            alpha = 0.05, cal_frac = 0.3,
                            wrapper = NULL, grid_weights = NULL,
                            calibration_idx = NULL,
                            dfspa_args = list(),
                            weight_model_args = list(),
                            seed = NULL) {
  .validate_conformal_inputs(F_hat, W, W_new, K, alpha, grid_weights,
                             dfspa_args, weight_model_args)
  m <- nrow(F_hat)
  if (is.matrix(W))     W     <- as.data.frame(W)
  if (is.matrix(W_new)) W_new <- as.data.frame(W_new)

  if (is.null(calibration_idx)) {
    if (!is.numeric(cal_frac) || length(cal_frac) != 1L ||
        cal_frac <= 0 || cal_frac >= 1) {
      stop("`cal_frac` must be a single number in (0, 1).")
    }
    if (!is.null(seed)) withr::local_seed(seed)
    n_cal <- max(1L, floor(cal_frac * m))
    calibration_idx <- sort(sample(seq_len(m), n_cal))
  } else {
    calibration_idx <- as.integer(calibration_idx)
    if (any(calibration_idx < 1L) || any(calibration_idx > m) ||
        anyDuplicated(calibration_idx)) {
      stop("`calibration_idx` must be unique integers in 1:nrow(F_hat).")
    }
  }
  training_idx <- setdiff(seq_len(m), calibration_idx)
  if (length(training_idx) < K) {
    stop(sprintf("Training set has only %d studies but K = %d.",
                 length(training_idx), K))
  }
  if (length(calibration_idx) < 2L) {
    stop("Calibration set must have at least 2 studies.")
  }

  pipe <- .run_pipeline(F_hat[training_idx, , drop = FALSE],
                        W[training_idx, , drop = FALSE],
                        K, grid_weights, dfspa_args, weight_model_args)

  conformal_from_fit(
    dfspa_fit    = pipe$fit,
    weight_model = pipe$wm,
    F_cal        = F_hat[calibration_idx, , drop = FALSE],
    W_cal        = W[calibration_idx, , drop = FALSE],
    W_new        = W_new,
    alpha        = alpha,
    wrapper      = wrapper,
    grid_weights = grid_weights
  )
}

#' Cross-conformal prediction intervals (pooled K-fold scores)
#'
#' Computes K-fold split conformal intervals in which the calibration scores
#' are pooled across folds, while point predictions for new targets are
#' produced from a final pipeline fit on all studies. Equivalent to running
#' [split_conformal()] `n_folds` times with different calibration sets and
#' pooling all conformity scores into a single empirical distribution.
#'
#' @details
#' For each fold, the MetaHunt pipeline is fit on the out-of-fold studies,
#' and conformity scores are computed on the in-fold studies. After all
#' folds complete, the pooled scores yield a single
#' \eqn{(1-\alpha)}-quantile (or one per grid point). Point predictions for
#' `W_new` use a pipeline refit on the full dataset. The interval at grid
#' point `g` for target `j` is
#' \eqn{[\tilde f^{(j)}(x_g) - q_g, \tilde f^{(j)}(x_g) + q_g]} (or the
#' scalar analogue when `wrapper` is supplied).
#'
#' This differs from Vovk's original cross-conformal predictor for
#' classification. For regression, pooling scores across folds is a
#' common practical extension of split conformal and reduces the variance
#' due to the single calibration split. Exact finite-sample coverage is
#' not guaranteed; see Barber et al. (2021, Jackknife+) for more
#' conservative alternatives.
#'
#' @inheritParams split_conformal
#' @param n_folds Integer number of folds (>= 2). Default `5`.
#'
#' @return An object of class `"metahunt_conformal"` (see
#'   [split_conformal()] for fields). `method` is `"cross"` and `n_cal` is
#'   the number of pooled scores.
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' G <- 30; m <- 60; K_true <- 3
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m))
#' eta <- cbind(0.8 * W$w1, -0.3 * W$w1, rep(0, m))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#' W_new <- data.frame(w1 = c(0, 1))
#'
#' res <- cross_conformal(F_hat, W, W_new, K = 3, n_folds = 4,
#'                        dfspa_args = list(denoise = FALSE), seed = 1)
#' res
#' }
#'
#' @export
cross_conformal <- function(F_hat, W, W_new, K,
                            alpha = 0.05, n_folds = 5L,
                            wrapper = NULL, grid_weights = NULL,
                            dfspa_args = list(),
                            weight_model_args = list(),
                            seed = NULL) {
  .validate_conformal_inputs(F_hat, W, W_new, K, alpha, grid_weights,
                             dfspa_args, weight_model_args)
  m <- nrow(F_hat); G_grid <- ncol(F_hat)
  if (is.matrix(W))     W     <- as.data.frame(W)
  if (is.matrix(W_new)) W_new <- as.data.frame(W_new)

  if (!is.numeric(n_folds) || length(n_folds) != 1L ||
      n_folds < 2L || n_folds != as.integer(n_folds)) {
    stop("`n_folds` must be an integer >= 2.")
  }
  n_folds <- as.integer(n_folds)
  if (n_folds > m) stop("`n_folds` cannot exceed the number of studies.")

  if (!is.null(seed)) withr::local_seed(seed)
  fold_id <- sample(rep(seq_len(n_folds), length.out = m))

  if (is.null(wrapper)) {
    pooled_scores <- matrix(NA_real_, m, G_grid)
  } else {
    pooled_scores <- rep(NA_real_, m)
  }
  failures <- character(0)
  for (f in seq_len(n_folds)) {
    train <- which(fold_id != f)
    test  <- which(fold_id == f)
    result <- tryCatch({
      pipe_f <- .run_pipeline(F_hat[train, , drop = FALSE],
                              W[train, , drop = FALSE],
                              K, grid_weights, dfspa_args, weight_model_args)
      pred_f <- predict_target(pipe_f$fit, pipe_f$wm,
                               W[test, , drop = FALSE])
      if (is.null(wrapper)) {
        abs(F_hat[test, , drop = FALSE] - pred_f)
      } else {
        abs(apply_wrapper(F_hat[test, , drop = FALSE], wrapper, grid_weights) -
              apply_wrapper(pred_f, wrapper, grid_weights))
      }
    }, error = function(e) e)
    if (inherits(result, "error")) {
      failures <- c(failures, sprintf("fold %d: %s", f,
                                      conditionMessage(result)))
      next
    }
    if (is.null(wrapper)) {
      pooled_scores[test, ] <- result
    } else {
      pooled_scores[test]   <- result
    }
  }
  if (length(failures)) {
    warning(sprintf("%d of %d folds failed:\n%s",
                    length(failures), n_folds,
                    paste0("  ", failures, collapse = "\n")))
  }

  pipe_all <- .run_pipeline(F_hat, W, K, grid_weights,
                            dfspa_args, weight_model_args)

  # Fake observed/predicted matrices to reuse output builder; we already
  # have the scores, so we'll inject them directly.
  .build_conformal_output(
    obs_cal = NULL, pred_cal = NULL,
    pipe = pipe_all, W_new = W_new,
    alpha = alpha, wrapper = wrapper, grid_weights = grid_weights,
    method = "cross",
    n_cal_label = NA_integer_,
    precomputed_scores = pooled_scores
  )
}

#' @keywords internal
#' @noRd
.validate_conformal_inputs <- function(F_hat, W, W_new, K, alpha,
                                       grid_weights,
                                       dfspa_args, weight_model_args) {
  if (!is.matrix(F_hat) || !is.numeric(F_hat)) {
    stop("`F_hat` must be a numeric matrix with studies as rows.")
  }
  if (!is.data.frame(W) && !is.matrix(W)) {
    stop("`W` must be a matrix or data frame.")
  }
  if (!is.data.frame(W_new) && !is.matrix(W_new)) {
    stop("`W_new` must be a matrix or data frame.")
  }
  if (nrow(W) != nrow(F_hat)) {
    stop("`W` must have the same number of rows as `F_hat`.")
  }
  if (!is.numeric(K) || length(K) != 1L || K < 1L || K != as.integer(K)) {
    stop("`K` must be a positive integer.")
  }
  if (!is.numeric(alpha) || length(alpha) != 1L || alpha <= 0 || alpha >= 1) {
    stop("`alpha` must be a single number in (0, 1).")
  }
  if (!is.list(dfspa_args) || !is.list(weight_model_args)) {
    stop("`dfspa_args` and `weight_model_args` must be named lists.")
  }
  invisible(NULL)
}

#' @keywords internal
#' @noRd
.build_conformal_output <- function(obs_cal, pred_cal, pipe, W_new, alpha,
                                    wrapper, grid_weights, method,
                                    n_cal_label,
                                    precomputed_scores = NULL) {
  if (is.null(precomputed_scores)) {
    if (is.null(wrapper)) {
      scores <- abs(obs_cal - pred_cal)
    } else {
      scores <- abs(apply_wrapper(obs_cal, wrapper, grid_weights) -
                      apply_wrapper(pred_cal, wrapper, grid_weights))
    }
  } else {
    scores <- precomputed_scores
  }

  if (is.null(wrapper)) {
    q <- apply(scores, 2, .conformal_quantile, alpha = alpha)
    n_cal_used <- sum(stats::complete.cases(scores))
    if (any(is.infinite(q))) {
      n_inf <- sum(is.infinite(q))
      warning(sprintf(
        paste0("With n_cal = %d and alpha = %g, the conformal quantile is infinite ",
               "at %d of %d grid points; intervals are unbounded there. Increase ",
               "calibration size (raise `cal_frac` or supply more studies) or use ",
               "a larger `alpha`."),
        n_cal_used, alpha, n_inf, length(q)))
    }
    pred_new <- predict_target(pipe$fit, pipe$wm, W_new)
    lower <- sweep(pred_new, 2, q, `-`)
    upper <- sweep(pred_new, 2, q, `+`)
    result <- list(
      prediction = pred_new, lower = lower, upper = upper,
      alpha = alpha, method = method,
      n_cal = if (is.na(n_cal_label)) n_cal_used else n_cal_label,
      quantile = q, wrapper = NULL
    )
  } else {
    scores_vec <- as.numeric(scores)
    q <- .conformal_quantile(scores_vec, alpha)
    n_cal_used <- sum(is.finite(scores_vec))
    if (is.infinite(q)) {
      warning(sprintf(
        paste0("With n_cal = %d and alpha = %g, the conformal quantile is infinite; ",
               "intervals are unbounded. Increase calibration size or use a larger `alpha`."),
        n_cal_used, alpha))
    }
    pred_new <- predict_target(pipe$fit, pipe$wm, W_new)
    s_new <- apply_wrapper(pred_new, wrapper, grid_weights)
    result <- list(
      prediction = s_new, lower = s_new - q, upper = s_new + q,
      alpha = alpha, method = method,
      n_cal = if (is.na(n_cal_label)) n_cal_used else n_cal_label,
      quantile = q, wrapper = wrapper
    )
  }
  class(result) <- "metahunt_conformal"
  result
}

#' Plot a conformal prediction-interval object
#'
#' For pointwise objects (no `wrapper` was used), draws the predicted
#' function for one target study together with its pointwise band.
#' For scalar objects, draws point predictions with whisker error bars
#' over all targets.
#'
#' @param x A `"metahunt_conformal"` object.
#' @param target_idx For pointwise objects, the integer index of the
#'   target study to plot (default `1`). Ignored for scalar objects.
#' @param x_axis Optional numeric vector of length `G_grid` giving the
#'   x-axis values for pointwise plotting. Defaults to `seq_len(G_grid)`.
#' @param fill Polygon fill colour for the band. Default semi-transparent
#'   steel blue.
#' @param line_col Line colour for the predicted function (or points in
#'   scalar mode). Default steel blue.
#' @param ... Additional graphical parameters passed to the underlying
#'   plotting calls.
#'
#' @return Invisibly returns `x`.
#'
#' @export
plot.metahunt_conformal <- function(x, target_idx = 1L, x_axis = NULL,
                                    fill = grDevices::adjustcolor("steelblue", 0.2),
                                    line_col = "steelblue", ...) {
  alpha <- x$alpha
  if (is.matrix(x$prediction)) {
    if (target_idx < 1L || target_idx > nrow(x$prediction)) {
      stop("`target_idx` is out of range for this object.")
    }
    G <- ncol(x$prediction)
    if (is.null(x_axis)) x_axis <- seq_len(G)
    if (length(x_axis) != G) {
      stop("`x_axis` must have length equal to ncol(x$prediction).")
    }
    pred <- x$prediction[target_idx, ]
    lo   <- x$lower[target_idx, ]
    up   <- x$upper[target_idx, ]
    finite_y <- c(pred, lo[is.finite(lo)], up[is.finite(up)])
    plot(x_axis, pred, type = "n", ylim = range(finite_y),
         xlab = "grid", ylab = expression(tilde(f)(x)),
         main = sprintf("%.0f%% pointwise band (target %d)",
                        100 * (1 - alpha), target_idx), ...)
    if (all(is.finite(c(lo, up)))) {
      graphics::polygon(c(x_axis, rev(x_axis)), c(lo, rev(up)),
                        col = fill, border = NA)
    }
    graphics::lines(x_axis, pred, lwd = 2, col = line_col)
  } else {
    n <- length(x$prediction)
    yrange <- range(c(x$prediction, x$lower, x$upper),
                    finite = TRUE)
    plot(seq_len(n), x$prediction, pch = 16, col = line_col,
         ylim = yrange,
         xlab = "target index", ylab = "scalar summary",
         main = sprintf("%.0f%% intervals (n = %d)",
                        100 * (1 - alpha), n), ...)
    graphics::arrows(seq_len(n), x$lower, seq_len(n), x$upper,
                     length = 0.05, angle = 90, code = 3,
                     col = line_col)
  }
  invisible(x)
}

#' Summarise a conformal prediction-interval object
#'
#' Produces a small list of descriptive statistics about a
#' `"metahunt_conformal"` object: interval widths, quantile summaries, and
#' calibration diagnostics. Returns an object of class
#' `"summary.metahunt_conformal"` with a matching `print` method.
#'
#' @param object A `"metahunt_conformal"` object from [split_conformal()],
#'   [cross_conformal()], or [conformal_from_fit()].
#' @param ... Unused; present for S3 generic consistency.
#'
#' @return A list of class `"summary.metahunt_conformal"`. In pointwise mode
#'   (no wrapper) the list contains `n_targets`, `G_grid`, `n_cal`, `alpha`,
#'   `method`, `mean_interval_width`, `frac_finite_quantile`,
#'   `quantile_summary`, and `wrapper`. In scalar mode (wrapper supplied)
#'   the list contains `n_targets`, `n_cal`, `alpha`, `method`,
#'   `mean_interval_width`, `quantile`, `quantile_finite`, and `wrapper`.
#'
#' @examples
#' set.seed(1)
#' G <- 25; m <- 60
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m))
#' eta <- cbind(0.6 * W$w1, -0.3 * W$w1, rep(0, m))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#' W_new <- data.frame(w1 = c(0, 1))
#' res <- split_conformal(F_hat, W, W_new, K = 3,
#'                        dfspa_args = list(denoise = FALSE), seed = 1)
#' summary(res)
#'
#' @export
summary.metahunt_conformal <- function(object, ...) {
  if (!inherits(object, "metahunt_conformal")) {
    stop("`object` must be a `metahunt_conformal` object.")
  }
  if (is.null(object$wrapper)) {
    widths <- object$upper - object$lower
    finite_widths <- widths[is.finite(widths)]
    qfinite <- object$quantile[is.finite(object$quantile)]
    if (length(qfinite)) {
      qsum <- c(min = min(qfinite), median = stats::median(qfinite),
                mean = mean(qfinite), max = max(qfinite))
    } else {
      qsum <- c(min = NA_real_, median = NA_real_,
                mean = NA_real_, max = NA_real_)
    }
    out <- list(
      mode = "pointwise",
      n_targets = nrow(object$prediction),
      G_grid = ncol(object$prediction),
      n_cal = object$n_cal,
      alpha = object$alpha,
      method = object$method,
      mean_interval_width = if (length(finite_widths))
        mean(finite_widths) else NA_real_,
      frac_finite_quantile = mean(is.finite(object$quantile)),
      quantile_summary = qsum,
      wrapper = NULL
    )
  } else {
    widths <- object$upper - object$lower
    finite_widths <- widths[is.finite(widths)]
    out <- list(
      mode = "scalar",
      n_targets = length(object$prediction),
      n_cal = object$n_cal,
      alpha = object$alpha,
      method = object$method,
      mean_interval_width = if (length(finite_widths))
        mean(finite_widths) else NA_real_,
      quantile = object$quantile,
      quantile_finite = is.finite(object$quantile),
      wrapper = object$wrapper
    )
  }
  class(out) <- "summary.metahunt_conformal"
  out
}

#' @export
print.summary.metahunt_conformal <- function(x, ...) {
  cat("Summary of MetaHunt conformal prediction\n")
  cat("  method:        ", x$method, "\n")
  cat("  alpha:         ", x$alpha, "\n")
  cat("  n calibration: ", x$n_cal, "\n")
  if (identical(x$mode, "pointwise")) {
    cat("  mode:           pointwise (per grid point)\n")
    cat("  n targets:     ", x$n_targets, "\n")
    cat("  grid size:     ", x$G_grid, "\n")
    cat("  mean interval width:", signif(x$mean_interval_width, 4), "\n")
    cat("  fraction finite quantiles:",
        signif(x$frac_finite_quantile, 4), "\n")
    qs <- x$quantile_summary
    if (all(is.na(qs))) {
      cat("  quantile:       all infinite (insufficient calibration)\n")
    } else {
      cat(sprintf(
        "  quantile (finite):  min=%.4g, median=%.4g, mean=%.4g, max=%.4g\n",
        qs["min"], qs["median"], qs["mean"], qs["max"]))
    }
  } else {
    cat("  mode:           scalar (via wrapper)\n")
    cat("  n targets:     ", x$n_targets, "\n")
    if (isTRUE(x$quantile_finite)) {
      cat("  quantile:      ", signif(x$quantile, 4), "\n")
      cat("  mean interval width:", signif(x$mean_interval_width, 4), "\n")
    } else {
      cat("  quantile:       Inf (insufficient calibration)\n")
    }
  }
  invisible(x)
}

#' Empirical coverage of a conformal prediction-interval object
#'
#' Computes empirical coverage indicators of a fitted
#' `"metahunt_conformal"` object against held-out observed study-level
#' functions. The held-out studies must correspond positionally to the
#' targets used to build `object` (i.e. `F_obs[i, ]` is the observed
#' function for the same target whose prediction is in
#' `object$prediction[i, ]` or `object$prediction[i]`).
#'
#' @details
#' In **pointwise** mode each entry `(i, g)` of `F_obs` is compared against
#' `[object$lower[i, g], object$upper[i, g]]`. In **scalar** mode `F_obs`
#' is first reduced to a length-`n_target` vector via
#' [apply_wrapper()] using `object$wrapper` and the supplied
#' `grid_weights`, and then compared against
#' `[object$lower, object$upper]`.
#'
#' Coverage is a finite-sample diagnostic. Nominal coverage is
#' `1 - object$alpha`; empirical coverage will fluctuate around this
#' value due to sampling.
#'
#' @param object A `"metahunt_conformal"` object from [split_conformal()],
#'   [cross_conformal()], or [conformal_from_fit()].
#' @param F_obs An `n_target`-by-`G_grid` numeric matrix of observed
#'   study-level functions for the target studies, in the same row order
#'   as `object$prediction`.
#' @param grid_weights Optional length-`G_grid` non-negative numeric
#'   vector. Only used in scalar mode (passed to [apply_wrapper()]).
#'
#' @return A list. In **pointwise** mode the list contains
#'   \describe{
#'     \item{`pointwise`}{`n_target`-by-`G_grid` logical matrix of
#'       coverage indicators.}
#'     \item{`per_target`}{Length-`n_target` numeric vector of mean
#'       coverage across the grid for each target.}
#'     \item{`per_grid_point`}{Length-`G_grid` numeric vector of mean
#'       coverage across targets at each grid point.}
#'     \item{`overall`}{Scalar mean coverage across all entries.}
#'     \item{`nominal`}{Nominal coverage `1 - object$alpha`.}
#'   }
#'   In **scalar** mode the list contains
#'   \describe{
#'     \item{`pointwise`}{Length-`n_target` logical vector of coverage
#'       indicators.}
#'     \item{`overall`}{Scalar mean coverage.}
#'     \item{`nominal`}{Nominal coverage `1 - object$alpha`.}
#'   }
#'
#' @examples
#' set.seed(1)
#' G <- 25; m <- 80
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m))
#' eta <- cbind(0.6 * W$w1, -0.3 * W$w1, rep(0, m))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' # held-out test set: same data-generating process, same W
#' test_idx <- 1:10
#' train_idx <- setdiff(seq_len(m), test_idx)
#' res <- split_conformal(F_hat[train_idx, ], W[train_idx, , drop = FALSE],
#'                        W[test_idx, , drop = FALSE], K = 3,
#'                        dfspa_args = list(denoise = FALSE), seed = 1)
#' cov <- coverage(res, F_obs = F_hat[test_idx, , drop = FALSE])
#' cov$overall
#'
#' @export
coverage <- function(object, F_obs, grid_weights = NULL) {
  if (!inherits(object, "metahunt_conformal")) {
    stop("`object` must be a `metahunt_conformal` object (from `split_conformal()`, `cross_conformal()`, or `conformal_from_fit()`).")
  }
  if (!is.matrix(F_obs) || !is.numeric(F_obs)) {
    stop("`F_obs` must be a numeric matrix with target studies as rows.")
  }

  if (is.null(object$wrapper)) {
    n_target <- nrow(object$prediction)
    G_grid   <- ncol(object$prediction)
    if (nrow(F_obs) != n_target) {
      stop(sprintf(
        "`F_obs` must have %d rows (one per target in `object$prediction`); received %d.",
        n_target, nrow(F_obs)))
    }
    if (ncol(F_obs) != G_grid) {
      stop(sprintf(
        "`F_obs` must have %d columns (matching `ncol(object$prediction)`); received %d.",
        G_grid, ncol(F_obs)))
    }
    pw <- (F_obs >= object$lower) & (F_obs <= object$upper)
    list(
      pointwise = pw,
      per_target = rowMeans(pw),
      per_grid_point = colMeans(pw),
      overall = mean(pw),
      nominal = 1 - object$alpha
    )
  } else {
    n_target <- length(object$prediction)
    if (nrow(F_obs) != n_target) {
      stop(sprintf(
        "`F_obs` must have %d rows (one per target in `object$prediction`); received %d.",
        n_target, nrow(F_obs)))
    }
    s_obs <- apply_wrapper(F_obs, object$wrapper, grid_weights)
    pw <- (s_obs >= object$lower) & (s_obs <= object$upper)
    list(
      pointwise = pw,
      overall = mean(pw),
      nominal = 1 - object$alpha
    )
  }
}

#' @export
print.metahunt_conformal <- function(x, ...) {
  cat("MetaHunt conformal prediction\n")
  cat("  method:       ", x$method, "\n")
  cat("  alpha:        ", x$alpha, "\n")
  cat("  n calibration:", x$n_cal, "\n")
  if (is.null(x$wrapper)) {
    cat("  mode:          pointwise (per grid point)\n")
    cat("  n targets:    ", nrow(x$prediction), "\n")
    cat("  grid size:    ", ncol(x$prediction), "\n")
    qfinite <- x$quantile[is.finite(x$quantile)]
    if (length(qfinite)) {
      cat("  quantile range:", signif(min(qfinite), 4),
          "to", signif(max(qfinite), 4), "\n")
    } else {
      cat("  quantile:      Inf (insufficient calibration)\n")
    }
  } else {
    cat("  mode:          scalar (via wrapper)\n")
    cat("  n targets:    ", length(x$prediction), "\n")
    cat("  quantile:     ", signif(x$quantile, 4), "\n")
  }
  invisible(x)
}

make_toy <- function(seed = 1, m = 100, G = 30, K_true = 3, sd_noise = 0.05) {
  set.seed(seed)
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = sd_noise), m, G)
  list(F_hat = F_hat, W = W, basis = basis, pi_true = pi_true, x = x)
}

test_that("split_conformal pointwise: shapes and containment", {
  d <- make_toy()
  W_new <- d$W[1:5, , drop = FALSE]
  res <- split_conformal(d$F_hat, d$W, W_new, K = 3, seed = 1)

  expect_s3_class(res, "metahunt_conformal")
  expect_equal(res$method, "split")
  expect_equal(dim(res$prediction), c(5L, ncol(d$F_hat)))
  expect_equal(dim(res$lower), dim(res$prediction))
  expect_equal(dim(res$upper), dim(res$prediction))
  expect_length(res$quantile, ncol(d$F_hat))
  expect_true(all(res$lower <= res$prediction))
  expect_true(all(res$upper >= res$prediction))
  expect_equal(res$alpha, 0.05)
})

test_that("split_conformal scalar: shapes and containment", {
  d <- make_toy()
  W_new <- d$W[1:7, , drop = FALSE]
  res <- split_conformal(d$F_hat, d$W, W_new, K = 3,
                         wrapper = mean, seed = 1)

  expect_length(res$prediction, 7L)
  expect_length(res$lower, 7L)
  expect_length(res$upper, 7L)
  expect_length(res$quantile, 1L)
  expect_true(all(res$lower <= res$prediction))
  expect_true(all(res$upper >= res$prediction))
})

test_that("split_conformal achieves roughly nominal coverage on held-out data", {
  d <- make_toy(m = 200)
  # leave-out 50 studies as test, use the rest to fit + calibrate
  set.seed(2)
  test_idx <- sample(nrow(d$F_hat), 50)
  train_idx <- setdiff(seq_len(nrow(d$F_hat)), test_idx)

  res <- split_conformal(
    d$F_hat[train_idx, , drop = FALSE], d$W[train_idx, , drop = FALSE],
    d$W[test_idx, , drop = FALSE], K = 3,
    wrapper = mean, alpha = 0.1, seed = 3
  )
  obs_scalar <- apply_wrapper(d$F_hat[test_idx, , drop = FALSE],
                              wrapper = mean)
  covered <- obs_scalar >= res$lower & obs_scalar <= res$upper
  emp_cov <- mean(covered)
  # with n_cal ~ 45 and alpha = 0.1, expect ~90% empirical coverage;
  # be generous (0.75-1.0) since we only have 50 test points.
  expect_gt(emp_cov, 0.75)
})

test_that("split_conformal is reproducible with seed", {
  d <- make_toy(m = 80)
  W_new <- d$W[1:3, , drop = FALSE]
  a <- split_conformal(d$F_hat, d$W, W_new, K = 3, wrapper = mean, seed = 42)
  b <- split_conformal(d$F_hat, d$W, W_new, K = 3, wrapper = mean, seed = 42)
  expect_equal(a$prediction, b$prediction)
  expect_equal(a$lower, b$lower)
  expect_equal(a$upper, b$upper)
})

test_that("split_conformal honours calibration_idx", {
  d <- make_toy(m = 80)
  W_new <- d$W[1:3, , drop = FALSE]
  idx <- 1:20
  res <- split_conformal(d$F_hat, d$W, W_new, K = 3, wrapper = mean,
                         calibration_idx = idx)
  expect_equal(res$n_cal, length(idx))
})

test_that("split_conformal: infinite quantile when n_cal too small for alpha", {
  d <- make_toy(m = 30)
  W_new <- d$W[1:2, , drop = FALSE]
  expect_warning(
    res <- split_conformal(d$F_hat, d$W, W_new, K = 3,
                           alpha = 0.01, cal_frac = 0.1, wrapper = mean,
                           seed = 1),
    "infinite"
  )
  # with n_cal = 3 and alpha = 0.01, k = ceiling(0.99 * 4) = 4 > 3 → Inf
  expect_true(is.infinite(res$quantile))
  expect_true(all(is.infinite(res$lower) | is.infinite(res$upper)))
})

test_that("cross_conformal: shapes and basic correctness", {
  d <- make_toy(m = 60)
  W_new <- d$W[1:4, , drop = FALSE]
  res <- cross_conformal(d$F_hat, d$W, W_new, K = 3,
                         n_folds = 4L, wrapper = mean, seed = 7)

  expect_s3_class(res, "metahunt_conformal")
  expect_equal(res$method, "cross")
  expect_length(res$prediction, 4L)
  expect_true(res$n_cal <= nrow(d$F_hat))
  expect_true(all(res$lower <= res$prediction))
  expect_true(all(res$upper >= res$prediction))
})

test_that("cross_conformal pointwise: shapes match grid", {
  d <- make_toy(m = 60)
  W_new <- d$W[1:3, , drop = FALSE]
  res <- cross_conformal(d$F_hat, d$W, W_new, K = 3,
                         n_folds = 4L, seed = 9)
  expect_equal(dim(res$prediction), c(3L, ncol(d$F_hat)))
  expect_length(res$quantile, ncol(d$F_hat))
})

test_that("conformal input validation catches malformed inputs", {
  d <- make_toy(m = 40)
  expect_error(split_conformal(d$F_hat, d$W, d$W[1:2, ], K = 3, alpha = 1.5),
               "alpha.*\\(0, 1\\)")
  expect_error(split_conformal(d$F_hat, d$W, d$W[1:2, ], K = 3,
                               cal_frac = 1.2),
               "cal_frac.*\\(0, 1\\)")
  expect_error(cross_conformal(d$F_hat, d$W, d$W[1:2, ], K = 3,
                               n_folds = 1),
               "n_folds")
  expect_error(split_conformal(d$F_hat, d$W[-1, ], d$W[1:2, ], K = 3),
               "same number of rows")
})

test_that("print.metahunt_conformal runs without error", {
  d <- make_toy(m = 40)
  suppressWarnings({
    res1 <- split_conformal(d$F_hat, d$W, d$W[1:2, ], K = 3, seed = 1)
    res2 <- split_conformal(d$F_hat, d$W, d$W[1:2, ], K = 3,
                            wrapper = mean, seed = 1)
  })
  expect_output(print(res1), "MetaHunt conformal prediction")
  expect_output(print(res2), "MetaHunt conformal prediction")
})

make_toy_cs <- function(seed = 1, m = 80, G = 25, sd_noise = 0.05) {
  set.seed(seed)
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = sd_noise), m, G)
  list(F_hat = F_hat, W = W, basis = basis, pi_true = pi_true, x = x)
}

test_that("summary.metahunt_conformal: pointwise object has expected fields", {
  d <- make_toy_cs(m = 80, G = 25)
  W_new <- d$W[1:5, , drop = FALSE]
  res <- split_conformal(d$F_hat, d$W, W_new, K = 3,
                         dfspa_args = list(denoise = FALSE), seed = 1)

  s <- summary(res)
  expect_s3_class(s, "summary.metahunt_conformal")
  expect_true(all(c("n_targets", "G_grid", "n_cal", "alpha", "method",
                    "mean_interval_width", "frac_finite_quantile",
                    "quantile_summary") %in% names(s)))
  expect_equal(s$n_targets, 5L)
  expect_equal(s$G_grid, ncol(d$F_hat))
  expect_equal(s$method, "split")
  expect_true(s$frac_finite_quantile >= 0 && s$frac_finite_quantile <= 1)
  expect_output(print(s), "Summary of MetaHunt conformal prediction")
})

test_that("summary.metahunt_conformal: scalar object has expected fields", {
  d <- make_toy_cs(m = 80, G = 25)
  W_new <- d$W[1:5, , drop = FALSE]
  res <- split_conformal(d$F_hat, d$W, W_new, K = 3, wrapper = mean,
                         dfspa_args = list(denoise = FALSE), seed = 1)

  s <- summary(res)
  expect_s3_class(s, "summary.metahunt_conformal")
  expect_true(all(c("n_targets", "n_cal", "alpha", "method",
                    "mean_interval_width", "quantile",
                    "quantile_finite") %in% names(s)))
  expect_equal(s$n_targets, 5L)
  expect_equal(s$method, "split")
  expect_output(print(s), "Summary of MetaHunt conformal prediction")
})

test_that("coverage(): pointwise object returns correct shapes and ranges", {
  d <- make_toy_cs(m = 80, G = 25)
  test_idx <- 1:8
  train_idx <- setdiff(seq_len(nrow(d$F_hat)), test_idx)
  res <- split_conformal(d$F_hat[train_idx, , drop = FALSE],
                         d$W[train_idx, , drop = FALSE],
                         d$W[test_idx, , drop = FALSE], K = 3,
                         dfspa_args = list(denoise = FALSE), seed = 1)

  cov <- coverage(res, F_obs = d$F_hat[test_idx, , drop = FALSE])
  expect_true(all(c("pointwise", "per_target", "per_grid_point",
                    "overall", "nominal") %in% names(cov)))
  expect_equal(dim(cov$pointwise),
               c(length(test_idx), ncol(d$F_hat)))
  expect_length(cov$per_target, length(test_idx))
  expect_length(cov$per_grid_point, ncol(d$F_hat))
  expect_true(cov$overall >= 0 && cov$overall <= 1)
  expect_true(all(cov$per_target >= 0 & cov$per_target <= 1))
  expect_true(all(cov$per_grid_point >= 0 & cov$per_grid_point <= 1))
  expect_equal(cov$nominal, 1 - res$alpha)
})

test_that("coverage(): scalar object returns correct shapes and ranges", {
  d <- make_toy_cs(m = 80, G = 25)
  test_idx <- 1:8
  train_idx <- setdiff(seq_len(nrow(d$F_hat)), test_idx)
  res <- split_conformal(d$F_hat[train_idx, , drop = FALSE],
                         d$W[train_idx, , drop = FALSE],
                         d$W[test_idx, , drop = FALSE], K = 3,
                         wrapper = mean,
                         dfspa_args = list(denoise = FALSE), seed = 1)

  cov <- coverage(res, F_obs = d$F_hat[test_idx, , drop = FALSE])
  expect_true(all(c("pointwise", "overall", "nominal") %in% names(cov)))
  expect_length(cov$pointwise, length(test_idx))
  expect_true(is.logical(cov$pointwise))
  expect_true(cov$overall >= 0 && cov$overall <= 1)
  expect_equal(cov$nominal, 1 - res$alpha)
})

test_that("coverage(): rejects mismatched F_obs shape", {
  d <- make_toy_cs(m = 80, G = 25)
  test_idx <- 1:5
  train_idx <- setdiff(seq_len(nrow(d$F_hat)), test_idx)
  res_pt <- split_conformal(d$F_hat[train_idx, , drop = FALSE],
                            d$W[train_idx, , drop = FALSE],
                            d$W[test_idx, , drop = FALSE], K = 3,
                            dfspa_args = list(denoise = FALSE), seed = 1)
  # wrong number of rows
  expect_error(
    coverage(res_pt, F_obs = d$F_hat[1:3, , drop = FALSE]),
    "rows"
  )
  # wrong number of columns
  expect_error(
    coverage(res_pt, F_obs = d$F_hat[test_idx, 1:10, drop = FALSE]),
    "columns"
  )

  res_sc <- split_conformal(d$F_hat[train_idx, , drop = FALSE],
                            d$W[train_idx, , drop = FALSE],
                            d$W[test_idx, , drop = FALSE], K = 3,
                            wrapper = mean,
                            dfspa_args = list(denoise = FALSE), seed = 1)
  expect_error(
    coverage(res_sc, F_obs = d$F_hat[1:2, , drop = FALSE]),
    "rows"
  )
})

test_that("coverage(): rejects non-metahunt_conformal input", {
  expect_error(coverage(list(prediction = 1), F_obs = matrix(0, 1, 1)),
               "metahunt_conformal")
  expect_error(coverage(42, F_obs = matrix(0, 1, 1)),
               "metahunt_conformal")
})

test_that("coverage(): empirical coverage is roughly close to nominal", {
  d <- make_toy_cs(m = 200, G = 25)
  set.seed(11)
  test_idx <- sample(nrow(d$F_hat), 50)
  train_idx <- setdiff(seq_len(nrow(d$F_hat)), test_idx)
  res <- split_conformal(d$F_hat[train_idx, , drop = FALSE],
                         d$W[train_idx, , drop = FALSE],
                         d$W[test_idx, , drop = FALSE], K = 3,
                         wrapper = mean, alpha = 0.1,
                         dfspa_args = list(denoise = FALSE), seed = 3)
  cov <- coverage(res, F_obs = d$F_hat[test_idx, , drop = FALSE])
  expect_gt(cov$overall, 0.6)
})

#' Build the `F_hat` matrix from a list of fitted study-level models
#'
#' Many users arrive at MetaHunt with one fitted model per study (e.g. a
#' `ranger::ranger()` random forest or a `grf::causal_forest()`) and a
#' chosen evaluation grid. `f_hat_from_models()` evaluates each model on
#' the shared grid and stacks the predictions into the `m`-by-`G_grid`
#' matrix the rest of the package expects.
#'
#' @details
#' By default the function dispatches on the model class:
#' * `ranger` objects are evaluated as `predict(model, data = grid)$predictions`.
#' * Objects inheriting from `causal_forest` or `grf` are evaluated as
#'   `predict(model, newdata = grid)$predictions`.
#' * All other classes fall through to
#'   `as.numeric(predict(model, newdata = grid))`, which works for `lm`,
#'   `glm`, `randomForest`, and most other R model objects.
#'
#' Override the dispatch with `predict_fn = function(model, grid) ...` if
#' your models need bespoke handling. The function must return a length-`G_grid`
#' numeric vector for each model.
#'
#' All rows of the returned matrix must have the same length (`G_grid`)
#' and contain no `NA` values; the function errors otherwise.
#'
#' @param models A non-empty list of fitted model objects, one per study.
#' @param grid A data frame (or matrix) of grid points at which to evaluate
#'   each model. Same columns as the data each model was fitted on.
#' @param predict_fn Optional `function(model, grid)` returning a numeric
#'   vector; defaults to the class-aware dispatcher above.
#'
#' @return An `length(models)`-by-`nrow(grid)` numeric matrix; row `i` is
#'   model `i` evaluated at every row of `grid`.
#'
#' @seealso [build_grid()] to construct `grid` from a reference dataset.
#'
#' @examples
#' # Toy example: each "centre" fits a polynomial regression
#' set.seed(1)
#' make_centre_data <- function(slope) {
#'   x <- runif(60)
#'   data.frame(x = x, y = slope * x + rnorm(60, sd = 0.1))
#' }
#' models <- lapply(c(-1, 0, 1, 0.5, -0.5), function(s)
#'   stats::lm(y ~ poly(x, 2), data = make_centre_data(s)))
#'
#' grid  <- data.frame(x = seq(0, 1, length.out = 30))
#' F_hat <- f_hat_from_models(models, grid)
#' dim(F_hat)   # 5 x 30
#'
#' @export
f_hat_from_models <- function(models, grid, predict_fn = NULL) {
  if (!is.list(models) || length(models) < 1L) {
    stop("`models` must be a non-empty list.")
  }
  if (is.matrix(grid)) grid <- as.data.frame(grid)
  if (!is.data.frame(grid)) {
    stop("`grid` must be a matrix or data frame.")
  }
  if (nrow(grid) < 1L) stop("`grid` must have at least one row.")

  if (is.null(predict_fn)) {
    predict_fn <- .predict_dispatch
  } else if (!is.function(predict_fn)) {
    stop("`predict_fn` must be a function or NULL.")
  }

  m <- length(models)
  G_grid <- nrow(grid)
  out <- matrix(NA_real_, nrow = m, ncol = G_grid)

  for (i in seq_len(m)) {
    pred <- tryCatch(predict_fn(models[[i]], grid),
                     error = function(e)
                       stop(sprintf("predict_fn failed on model %d: %s",
                                    i, conditionMessage(e))))
    pred <- as.numeric(pred)
    if (length(pred) != G_grid) {
      stop(sprintf("Model %d returned %d predictions but grid has %d rows.",
                   i, length(pred), G_grid))
    }
    if (anyNA(pred)) {
      stop(sprintf("Model %d returned NA predictions.", i))
    }
    out[i, ] <- pred
  }
  out
}

#' @keywords internal
#' @noRd
.predict_dispatch <- function(model, grid) {
  if (inherits(model, "ranger")) {
    pred <- stats::predict(model, data = grid)$predictions
  } else if (inherits(model, "causal_forest") ||
             inherits(model, "regression_forest") ||
             inherits(model, "grf")) {
    pred <- stats::predict(model, newdata = grid)$predictions
  } else {
    pred <- stats::predict(model, newdata = grid)
  }
  as.numeric(pred)
}

#' Build a shared evaluation grid from a reference dataset
#'
#' Constructs a data frame of grid points suitable for [f_hat_from_models()]
#' from any reference patient-level dataset. This is convenient when the
#' patient-level covariate space is multidimensional and there is no
#' obvious one-dimensional grid.
#'
#' @details
#' If `n_grid` is `NULL` or at least `nrow(reference_data)`, the reference
#' data is returned unchanged. Otherwise `n_grid` rows are sampled uniformly
#' at random (without replacement). The reference data should be on the
#' same scale and have the same columns as the data each centre's model
#' was fitted on.
#'
#' The empirical distribution of the returned grid implicitly defines the
#' \eqn{\mu} measure used downstream. Pass uniform `grid_weights` (the
#' default) to weight each grid point equally; pass non-uniform
#' `grid_weights` to weight by an external reference distribution.
#'
#' @param reference_data A data frame (or matrix) of reference patient-level
#'   covariates. May be a held-out target-population sample, the pooled
#'   source covariates, or any plausible reference distribution.
#' @param n_grid Optional integer giving the desired grid size. If `NULL`
#'   or `>= nrow(reference_data)`, the full reference data is returned.
#' @param seed Optional integer seed for reproducibility (used only when
#'   sub-sampling).
#'
#' @return A data frame of grid points.
#'
#' @examples
#' set.seed(1)
#' ref <- data.frame(age = rnorm(500, 60, 10),
#'                   bp  = rnorm(500, 130, 15),
#'                   sex = sample(c("F", "M"), 500, replace = TRUE))
#' grid <- build_grid(ref, n_grid = 50, seed = 1)
#' nrow(grid)
#' head(grid)
#'
#' @export
build_grid <- function(reference_data, n_grid = NULL, seed = NULL) {
  if (is.matrix(reference_data)) {
    reference_data <- as.data.frame(reference_data)
  }
  if (!is.data.frame(reference_data)) {
    stop("`reference_data` must be a matrix or data frame.")
  }
  N <- nrow(reference_data)
  if (N < 1L) stop("`reference_data` is empty.")

  if (is.null(n_grid) || n_grid >= N) {
    return(reference_data)
  }
  if (!is.numeric(n_grid) || length(n_grid) != 1L ||
      n_grid < 1L || n_grid != as.integer(n_grid)) {
    stop("`n_grid` must be a positive integer or NULL.")
  }
  if (!is.null(seed)) withr::local_seed(seed)
  reference_data[sample(N, n_grid), , drop = FALSE]
}

test_that("f_hat_from_models stacks lm predictions correctly", {
  set.seed(1)
  make <- function(slope) {
    x <- runif(60)
    data.frame(x = x, y = slope * x + rnorm(60, sd = 0.05))
  }
  models <- lapply(c(-1, 0, 1, 0.5), function(s) stats::lm(y ~ poly(x, 2),
                                                           data = make(s)))
  grid <- data.frame(x = seq(0, 1, length.out = 25))
  F_hat <- f_hat_from_models(models, grid)
  expect_equal(dim(F_hat), c(4L, 25L))
  expect_equal(F_hat[1, ],
               as.numeric(stats::predict(models[[1]], newdata = grid)))
})

test_that("f_hat_from_models honours custom predict_fn", {
  models <- list(
    function(x) sin(x),
    function(x) cos(x),
    function(x) x^2
  )
  grid <- data.frame(x = seq(0, 1, length.out = 10))
  F_hat <- f_hat_from_models(
    models, grid,
    predict_fn = function(model, grid) model(grid$x)
  )
  expect_equal(F_hat[1, ], sin(grid$x))
  expect_equal(F_hat[2, ], cos(grid$x))
  expect_equal(F_hat[3, ], grid$x^2)
})

test_that("f_hat_from_models errors when predict lengths disagree", {
  models <- list(
    function(x) x[1:5],
    function(x) x
  )
  grid <- data.frame(x = 1:10)
  expect_error(
    f_hat_from_models(models, grid,
                      predict_fn = function(m, g) m(g$x)),
    "different-length|returned 5 predictions"
  )
})

test_that("f_hat_from_models errors on NA predictions", {
  models <- list(function(x) c(NA_real_, x[-1]))
  grid <- data.frame(x = 1:5)
  expect_error(
    f_hat_from_models(models, grid,
                      predict_fn = function(m, g) m(g$x)),
    "NA"
  )
})

test_that("build_grid returns reference unchanged when n_grid is NULL or large", {
  ref <- data.frame(a = rnorm(40), b = rnorm(40))
  expect_identical(build_grid(ref), ref)
  expect_identical(build_grid(ref, n_grid = 100), ref)
})

test_that("build_grid samples without replacement and is reproducible", {
  ref <- data.frame(a = rnorm(80), b = rnorm(80))
  g1 <- build_grid(ref, n_grid = 25, seed = 7)
  g2 <- build_grid(ref, n_grid = 25, seed = 7)
  expect_equal(g1, g2)
  expect_equal(nrow(g1), 25L)
  expect_equal(anyDuplicated(rownames(g1)), 0L)
})

test_that("build_grid rejects malformed inputs", {
  expect_error(build_grid("nope"), "matrix or data frame")
  expect_error(build_grid(data.frame(a = double(0))), "empty")
  expect_error(build_grid(data.frame(a = 1:5), n_grid = -1),
               "positive integer")
})

#' Choose d-fSPA denoising parameters by cross-validation
#'
#' At a fixed `K`, performs k-fold cross-validation over a grid of denoising
#' parameter pairs `(N, Delta)` for [dfspa()]. For each candidate pair and
#' each fold, the full MetaHunt pipeline is fit on the out-of-fold studies
#' and predicts the held-out studies' functions. The pair with the lowest
#' average prediction error is selected.
#'
#' @details
#' This is the cross-validated tuning of the denoising parameters discussed
#' in Section 3.1 of the paper. Joint tuning over `(K, N, Delta)` is not
#' supported because it scales poorly; if you also want to choose `K`, do
#' it first via [cv_error_curve()] and then call this function at the
#' selected `K`.
#'
#' Default candidate grids:
#' * `N_grid    = m * c(NA, NA, NA)` resolved at runtime to
#'   `c(0.2, 0.5, 1.0) * log(m)`.
#' * `Delta_grid = max_pairwise_dist * c(0.05, 0.10, 0.20, 0.30)`.
#'
#' Pass your own `N_grid` / `Delta_grid` (in original units) to override.
#'
#' @param F_hat An `m`-by-`G_grid` numeric matrix of study-level function
#'   evaluations.
#' @param W An `m`-by-`p` matrix or data frame of study-level covariates.
#' @param K Integer number of basis functions (fixed during this search).
#' @param N_grid Optional numeric vector of candidate `N` values. Defaults
#'   to `c(0.2, 0.5, 1.0) * log(nrow(F_hat))`.
#' @param Delta_grid Optional numeric vector of candidate `Delta` values.
#'   Defaults to `c(0.05, 0.10, 0.20, 0.30)` times the maximum pairwise
#'   \eqn{L^2(\mu)} distance among studies.
#' @param n_folds Integer number of folds (default `5`).
#' @param grid_weights Optional length-`G_grid` non-negative numeric vector.
#' @param dfspa_args Named list of additional arguments for [dfspa()]
#'   (e.g. `list()`; do not include `N`, `Delta`, or `denoise` here, they
#'   are set by the search).
#' @param weight_model_args Named list of additional arguments for
#'   [fit_weight_model()].
#' @param seed Optional integer seed for reproducible fold assignment.
#'
#' @return An object of class `metahunt_denoising_search`: a list with
#'   \describe{
#'     \item{`grid`}{A data frame with one row per `(N, Delta)` pair,
#'       columns `N`, `Delta`, `cv_error`, `cv_se`, `n_folds_ok`.}
#'     \item{`best`}{A list with the `(N, Delta)` minimising `cv_error`.}
#'     \item{`K`, `n_folds`, `grid_weights`}{Inputs echoed back for
#'       traceability.}
#'   }
#'
#' @seealso [cv_error_curve()] for selecting `K`, [dfspa()] for the
#'   underlying basis-hunting algorithm.
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' G <- 30; m <- 80
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)
#' W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
#' eta <- as.matrix(W) %*% cbind(c(1, -0.5), c(-0.4, 1), c(0, 0))
#' pi_true <- exp(eta) / rowSums(exp(eta))
#' F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
#'
#' tune <- select_denoising_params(F_hat, W, K = 3, n_folds = 4, seed = 1)
#' tune$grid
#' tune$best
#' }
#'
#' @export
select_denoising_params <- function(F_hat, W, K,
                                    N_grid = NULL,
                                    Delta_grid = NULL,
                                    n_folds = 5L,
                                    grid_weights = NULL,
                                    dfspa_args = list(),
                                    weight_model_args = list(),
                                    seed = NULL) {
  if (!is.matrix(F_hat) || !is.numeric(F_hat)) {
    stop("`F_hat` must be a numeric matrix with studies as rows.")
  }
  m <- nrow(F_hat); G_grid <- ncol(F_hat)
  if (is.matrix(W)) W <- as.data.frame(W)
  if (!is.data.frame(W)) stop("`W` must be a matrix or data frame.")
  if (nrow(W) != m) stop("`W` must have the same number of rows as `F_hat`.")
  if (!is.numeric(K) || length(K) != 1L || K < 1L || K != as.integer(K)) {
    stop("`K` must be a positive integer.")
  }
  K <- as.integer(K)
  if (!is.list(dfspa_args) || !is.list(weight_model_args)) {
    stop("`dfspa_args` and `weight_model_args` must be named lists.")
  }
  reserved <- c("N", "Delta", "denoise")
  bad <- intersect(names(dfspa_args), reserved)
  if (length(bad)) {
    stop("`dfspa_args` must not contain ", paste(bad, collapse = ", "),
         "; these are set by the search.")
  }
  if (!is.numeric(n_folds) || length(n_folds) != 1L ||
      n_folds < 2L || n_folds != as.integer(n_folds)) {
    stop("`n_folds` must be an integer >= 2.")
  }
  n_folds <- as.integer(n_folds)
  if (n_folds > m) stop("`n_folds` cannot exceed the number of studies.")

  if (is.null(N_grid)) {
    N_grid <- c(0.2, 0.5, 1.0) * log(m)
  }
  if (!is.numeric(N_grid) || any(!is.finite(N_grid)) || any(N_grid <= 0)) {
    stop("`N_grid` must be a vector of positive finite numbers.")
  }

  if (is.null(Delta_grid)) {
    if (is.null(grid_weights)) {
      gw <- rep(1 / G_grid, G_grid)
    } else {
      gw <- grid_weights
    }
    sqw <- sqrt(gw)
    F_tilde <- sweep(F_hat, 2, sqw, `*`)
    max_dist <- max(stats::dist(F_tilde))
    Delta_grid <- max_dist * c(0.05, 0.10, 0.20, 0.30)
  }
  if (!is.numeric(Delta_grid) || any(!is.finite(Delta_grid)) ||
      any(Delta_grid < 0)) {
    stop("`Delta_grid` must be a vector of non-negative finite numbers.")
  }

  combos <- expand.grid(N = N_grid, Delta = Delta_grid,
                        KEEP.OUT.ATTRS = FALSE)
  if (!is.null(seed)) withr::local_seed(seed)
  fold_id <- sample(rep(seq_len(n_folds), length.out = m))

  fold_errors <- matrix(NA_real_, nrow = nrow(combos), ncol = n_folds)
  failures <- character(0)

  for (j in seq_len(nrow(combos))) {
    Nj <- combos$N[j]; Dj <- combos$Delta[j]
    for (f in seq_len(n_folds)) {
      train <- which(fold_id != f)
      test  <- which(fold_id == f)
      result <- tryCatch({
        fit_f <- do.call(dfspa,
          c(list(F_hat = F_hat[train, , drop = FALSE], K = K,
                 grid_weights = grid_weights,
                 N = Nj, Delta = Dj, denoise = TRUE),
            dfspa_args))
        pi_tr <- project_to_simplex(F_hat[train, , drop = FALSE],
                                    fit_f$bases,
                                    grid_weights = grid_weights)
        wm <- do.call(fit_weight_model,
                      c(list(pi_hat = pi_tr,
                             W = W[train, , drop = FALSE]),
                        weight_model_args))
        f_pred <- predict_target(fit_f, wm, W[test, , drop = FALSE])
        mean(.row_l2(F_hat[test, , drop = FALSE] - f_pred, grid_weights))
      }, error = function(e) e)
      if (inherits(result, "error")) {
        failures <- c(failures,
                      sprintf("(N=%g, Delta=%g, fold %d): %s",
                              Nj, Dj, f, conditionMessage(result)))
      } else {
        fold_errors[j, f] <- result
      }
    }
  }
  if (length(failures)) {
    warning(sprintf("%d (N, Delta, fold) combinations failed:\n%s",
                    length(failures),
                    paste0("  ", failures, collapse = "\n")))
  }

  cv_error <- rowMeans(fold_errors, na.rm = TRUE)
  cv_sd    <- apply(fold_errors, 1, stats::sd, na.rm = TRUE)
  n_ok     <- rowSums(!is.na(fold_errors))
  cv_se    <- ifelse(n_ok > 1L, cv_sd / sqrt(n_ok), NA_real_)

  grid <- data.frame(N = combos$N, Delta = combos$Delta,
                     cv_error = cv_error, cv_se = cv_se,
                     n_folds_ok = n_ok)

  if (all(is.na(grid$cv_error))) {
    stop("All (N, Delta) combinations failed for every fold.")
  }
  best_idx <- which.min(grid$cv_error)
  best <- list(N = grid$N[best_idx], Delta = grid$Delta[best_idx],
               cv_error = grid$cv_error[best_idx])

  structure(
    list(grid = grid, best = best,
         K = K, n_folds = n_folds, grid_weights = grid_weights),
    class = "metahunt_denoising_search"
  )
}

#' Print method for d-fSPA denoising parameter search results
#'
#' @param x An object of class `metahunt_denoising_search` returned by
#'   [select_denoising_params()].
#' @param ... Unused; present for S3 generic compatibility.
#'
#' @return Invisibly returns `x`.
#'
#' @export
print.metahunt_denoising_search <- function(x, ...) {
  cat("Denoising parameter search (CV at fixed K)\n")
  cat("  K:           ", x$K, "\n")
  cat("  n_folds:     ", x$n_folds, "\n")
  n_N     <- length(unique(x$grid$N))
  n_Delta <- length(unique(x$grid$Delta))
  cat(sprintf("  grid size:    %d (%d N values x %d Delta values)\n",
              nrow(x$grid), n_N, n_Delta))
  cat(sprintf("  best:         N = %s, Delta = %s, cv_error = %s\n",
              signif(x$best$N, 4),
              signif(x$best$Delta, 4),
              signif(x$best$cv_error, 4)))
  cat("\n")
  top_n <- min(5L, nrow(x$grid))
  cat(sprintf("  Top %d (lowest cv_error):\n", top_n))
  ord <- order(x$grid$cv_error, na.last = TRUE)
  top_grid <- x$grid[ord[seq_len(top_n)], , drop = FALSE]
  for (col in c("N", "Delta", "cv_error", "cv_se")) {
    top_grid[[col]] <- signif(top_grid[[col]], 4)
  }
  print(top_grid, row.names = FALSE)
  invisible(x)
}

test_that("select_denoising_params returns the expected structure", {
  set.seed(1)
  G <- 25; m <- 60; K_true <- 3
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.5), c(-0.4, 1), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  res <- select_denoising_params(F_hat, W, K = 3,
                                 N_grid = c(0.2, 0.5) * log(m),
                                 Delta_grid = c(0.1, 0.2),
                                 n_folds = 3L, seed = 1)

  expect_type(res, "list")
  expect_s3_class(res, "metahunt_denoising_search")
  expect_named(res$grid,
               c("N", "Delta", "cv_error", "cv_se", "n_folds_ok"),
               ignore.order = TRUE)
  expect_equal(nrow(res$grid), 4L)
  expect_named(res$best, c("N", "Delta", "cv_error"))
  expect_true(res$best$cv_error == min(res$grid$cv_error, na.rm = TRUE))
})

test_that("print.metahunt_denoising_search runs and prints a summary", {
  set.seed(11)
  G <- 25; m <- 60
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
  eta <- as.matrix(W) %*% cbind(c(1, -0.5), c(-0.4, 1), c(0, 0))
  pi_true <- exp(eta) / rowSums(exp(eta))
  F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

  res <- select_denoising_params(F_hat, W, K = 3,
                                 N_grid = c(0.2, 0.5) * log(m),
                                 Delta_grid = c(0.1, 0.2),
                                 n_folds = 3L, seed = 1)

  expect_output(print(res), "Denoising parameter search")
  expect_invisible(print(res))
})

test_that("select_denoising_params honours custom grids", {
  set.seed(2)
  G <- 20; m <- 50
  F_hat <- matrix(rnorm(m * G), m, G)
  W <- data.frame(w1 = rnorm(m))

  res <- suppressWarnings(
    select_denoising_params(F_hat, W, K = 2,
                            N_grid = c(1, 2),
                            Delta_grid = c(0.5, 1, 2),
                            n_folds = 3L, seed = 1)
  )
  expect_setequal(unique(res$grid$N), c(1, 2))
  expect_setequal(unique(res$grid$Delta), c(0.5, 1, 2))
  expect_equal(nrow(res$grid), 6L)
})

test_that("select_denoising_params is reproducible with a seed", {
  set.seed(3)
  G <- 20; m <- 50
  F_hat <- matrix(rnorm(m * G), m, G)
  W <- data.frame(w1 = rnorm(m))

  a <- select_denoising_params(F_hat, W, K = 2,
                               N_grid = c(0.5, 1),
                               Delta_grid = c(1, 2),
                               n_folds = 3L, seed = 99)
  b <- select_denoising_params(F_hat, W, K = 2,
                               N_grid = c(0.5, 1),
                               Delta_grid = c(1, 2),
                               n_folds = 3L, seed = 99)
  expect_equal(a$grid, b$grid)
  expect_equal(a$best, b$best)
})

test_that("select_denoising_params rejects N/Delta inside dfspa_args", {
  G <- 15; m <- 30
  F_hat <- matrix(rnorm(m * G), m, G)
  W <- data.frame(w1 = rnorm(m))
  expect_error(
    select_denoising_params(F_hat, W, K = 2, n_folds = 2L,
                            dfspa_args = list(N = 1)),
    "must not contain N"
  )
})

#' Denoised functional Successive Projection Algorithm (d-fSPA)
#'
#' Recovers a set of `K` latent basis functions from a collection of
#' study-level function estimates under the low-rank cross-study heterogeneity
#' assumption of Shi, Imai, and Zhang. Implements Algorithm 1 of the paper
#' ("The d-fSPA Algorithm for basis hunting").
#'
#' @details
#' Each study-level function is represented by its evaluations on a shared
#' grid of `G` points. The (weighted) \eqn{L^2(\mu)} inner product is
#' \eqn{\langle f,g\rangle = \sum_{j=1}^G w_j f(x_j) g(x_j)}, where the
#' `grid_weights` `w_j` are proportional to the measure \eqn{\mu}. If not
#' supplied, uniform weights `1 / G` are used.
#'
#' Denoising follows Jin (2024): for each study `i`, let
#' \eqn{B_\Delta(\hat f^{(i)}) = \{j : \|\hat f^{(j)} - \hat f^{(i)}\| \le
#' \Delta\}}. If \eqn{|B_\Delta(\hat f^{(i)})| < N}, study `i` is discarded;
#' otherwise \eqn{\hat f^{(i)}} is replaced by the average of the functions
#' in \eqn{B_\Delta}.
#' After denoising, the functional SPA step iteratively selects, at each of
#' the `K` iterations, the remaining function with the largest norm after
#' projecting out the span of previously selected bases.
#'
#' Default tuning parameters follow the heuristics of the paper:
#' `N = 0.5 * log(m)` and
#' \eqn{\Delta = \max_{ij} \|\hat f^{(i)} - \hat f^{(j)}\| / 10}.
#'
#' @param F_hat An `m`-by-`G` numeric matrix where row `i` is the evaluation
#'   of the estimated function \eqn{\hat f^{(i)}} at `G` grid points.
#' @param K Integer number of basis functions to recover. Must satisfy
#'   `1 <= K <= m` after denoising.
#' @param grid_weights Optional length-`G` non-negative numeric vector of
#'   grid weights defining the \eqn{L^2(\mu)} inner product. Defaults to
#'   uniform weights `1 / G`.
#' @param N,Delta Optional numeric tuning parameters controlling denoising.
#'   See Details.
#' @param denoise Logical; if `FALSE`, the denoising step is skipped and
#'   plain fSPA is run. Defaults to `TRUE`.
#'
#' @return An object of class `"dfspa"`: a list containing
#' \describe{
#'   \item{`bases`}{A `K`-by-`G` matrix whose rows are the recovered basis
#'     functions evaluated on the grid (denoised, if applicable).}
#'   \item{`selected`}{Length-`K` integer vector of the selected row indices
#'     into the post-denoising function matrix.}
#'   \item{`original_indices`}{Length-`K` integer vector of the selected
#'     study indices in the original input `F_hat` (before any rows were
#'     dropped by denoising).}
#'   \item{`kept`}{Integer vector of row indices of `F_hat` that survived
#'     denoising.}
#'   \item{`F_denoised`}{The post-denoising function matrix
#'     (`length(kept)`-by-`G`).}
#'   \item{`grid_weights`}{Grid weights used.}
#'   \item{`N`, `Delta`}{Tuning parameters actually used (or `NA` when
#'     `denoise = FALSE`).}
#'   \item{`K`}{Number of bases requested.}
#'   \item{`call`}{The matched call.}
#' }
#'
#' @examples
#' set.seed(1)
#' G <- 50
#' x <- seq(0, 1, length.out = G)
#' basis <- rbind(sin(pi * x), cos(pi * x), x)          # 3 true bases
#' pi_mat <- rbind(diag(3),                              # 3 pure studies
#'                 c(0.5, 0.3, 0.2),
#'                 c(0.2, 0.5, 0.3),
#'                 c(0.3, 0.3, 0.4))
#' F_hat <- pi_mat %*% basis                             # m = 6, G = 50
#' fit <- dfspa(F_hat, K = 3, denoise = FALSE)
#' fit$original_indices    # should be a permutation of 1, 2, 3
#'
#' @export
dfspa <- function(F_hat, K, grid_weights = NULL,
                  N = NULL, Delta = NULL, denoise = TRUE) {
  call <- match.call()

  if (!is.matrix(F_hat) || !is.numeric(F_hat)) {
    stop("`F_hat` must be a numeric matrix with studies as rows.")
  }
  if (anyNA(F_hat)) {
    stop("`F_hat` must not contain NA or NaN values.")
  }
  m <- nrow(F_hat)
  G <- ncol(F_hat)
  if (m < 2L) stop("`F_hat` must have at least two rows (studies).")

  if (!is.numeric(K) || length(K) != 1L || !is.finite(K) ||
      K < 1 || K != as.integer(K)) {
    stop("`K` must be a positive integer.")
  }
  K <- as.integer(K)
  if (K > m) stop("`K` must not exceed the number of studies `m`.")

  if (is.null(grid_weights)) {
    grid_weights <- rep(1 / G, G)
  } else {
    if (!is.numeric(grid_weights) || length(grid_weights) != G ||
        any(!is.finite(grid_weights)) || any(grid_weights < 0)) {
      stop("`grid_weights` must be a length-`ncol(F_hat)` non-negative numeric vector.")
    }
    if (sum(grid_weights) <= 0) {
      stop("`grid_weights` must have strictly positive total mass.")
    }
  }

  sqw <- sqrt(grid_weights)
  F_tilde <- sweep(F_hat, 2, sqw, `*`)
  D <- as.matrix(stats::dist(F_tilde))

  if (denoise) {
    if (is.null(N))     N     <- 0.5 * log(m)
    if (is.null(Delta)) Delta <- max(D) / 10
    if (!is.numeric(N) || length(N) != 1L || !is.finite(N) || N <= 0) {
      stop("`N` must be a positive finite scalar.")
    }
    if (!is.numeric(Delta) || length(Delta) != 1L || !is.finite(Delta) ||
        Delta < 0) {
      stop("`Delta` must be a non-negative finite scalar.")
    }
    nbr_counts <- rowSums(D <= Delta)           # includes self
    kept <- which(nbr_counts >= N)
    if (length(kept) < K) {
      stop(sprintf(
        "Only %d studies survive denoising but K = %d; consider increasing `Delta`, decreasing `N`, or reducing `K`.",
        length(kept), K))
    }
    F_denoised <- matrix(0, nrow = length(kept), ncol = G)
    for (ii in seq_along(kept)) {
      nbrs <- which(D[kept[ii], ] <= Delta)
      F_denoised[ii, ] <- colMeans(F_hat[nbrs, , drop = FALSE])
    }
  } else {
    N <- NA_real_
    Delta <- NA_real_
    kept <- seq_len(m)
    F_denoised <- F_hat
  }

  m_kept <- length(kept)
  F_denoised_tilde <- sweep(F_denoised, 2, sqw, `*`)

  H <- F_denoised_tilde
  selected <- integer(K)
  tol <- .Machine$double.eps^0.5 * max(1, max(rowSums(H^2)))
  for (k in seq_len(K)) {
    norms_sq <- rowSums(H^2)
    s_k <- which.max(norms_sq)
    if (norms_sq[s_k] <= tol) {
      stop(sprintf(
        "d-fSPA halted at k = %d because residual norms are numerically zero; K may be too large.",
        k))
    }
    selected[k] <- s_k
    u <- H[s_k, ]
    u <- u / sqrt(sum(u^2))
    proj_coef <- as.numeric(H %*% u)
    H <- H - outer(proj_coef, u)
  }

  structure(
    list(
      bases            = F_denoised[selected, , drop = FALSE],
      selected         = selected,
      original_indices = kept[selected],
      kept             = kept,
      F_denoised       = F_denoised,
      grid_weights     = grid_weights,
      N                = N,
      Delta            = Delta,
      K                = K,
      m                = m,
      call             = call
    ),
    class = "dfspa"
  )
}

#' @export
print.dfspa <- function(x, ...) {
  cat("d-fSPA basis hunting result\n")
  cat("  studies kept after denoising:", length(x$kept), "of", x$m, "\n")
  cat("  K (bases recovered):         ", x$K, "\n")
  cat("  selected study indices:      ", paste(x$original_indices, collapse = ", "), "\n")
  if (is.finite(x$N))     cat("  N:                           ", signif(x$N, 4), "\n")
  if (is.finite(x$Delta)) cat("  Delta:                       ", signif(x$Delta, 4), "\n")
  invisible(x)
}

test_that("dfspa recovers pure vertices in the noiseless case", {
  set.seed(1)
  G <- 50
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x)
  pi_mat <- rbind(
    diag(3),
    c(0.5, 0.3, 0.2),
    c(0.2, 0.5, 0.3),
    c(0.3, 0.3, 0.4),
    c(0.1, 0.8, 0.1),
    c(0.1, 0.1, 0.8)
  )
  F_hat <- pi_mat %*% basis

  fit <- dfspa(F_hat, K = 3, denoise = FALSE)

  expect_s3_class(fit, "dfspa")
  expect_length(fit$selected, 3L)
  expect_setequal(fit$original_indices, c(1L, 2L, 3L))
  expect_equal(dim(fit$bases), c(3L, G))
})

test_that("dfspa with denoising still recovers bases under small noise", {
  set.seed(42)
  G <- 80
  x <- seq(0, 1, length.out = G)
  basis <- rbind(sin(pi * x), cos(pi * x), x^2)
  K_true <- 3L
  m <- 40L
  # generate random simplex weights concentrated near vertices
  pi_mat <- matrix(0, m, K_true)
  pi_mat[1:3, ] <- diag(3)
  pi_mat[4:6, ] <- diag(3)
  for (i in 7:m) {
    w <- stats::rgamma(K_true, shape = 0.3)
    pi_mat[i, ] <- w / sum(w)
  }
  F_true <- pi_mat %*% basis
  F_hat  <- F_true + matrix(stats::rnorm(m * G, sd = 0.02), m, G)

  fit <- dfspa(F_hat, K = K_true)

  expect_equal(fit$K, K_true)
  expect_equal(length(fit$original_indices), K_true)
  # recovered bases should match true bases (up to permutation) within
  # tolerance set by noise and purity level
  dists <- outer(
    seq_len(K_true), seq_len(K_true),
    Vectorize(function(i, j) sqrt(mean((fit$bases[i, ] - basis[j, ])^2)))
  )
  best_match <- apply(dists, 1, min)
  expect_true(all(best_match < 0.05))
})

test_that("dfspa errors on invalid inputs", {
  F_hat <- matrix(1:20, nrow = 5, ncol = 4)
  expect_error(dfspa(as.data.frame(F_hat), K = 2), "numeric matrix")
  expect_error(dfspa(F_hat, K = 0), "positive integer")
  expect_error(dfspa(F_hat, K = 10), "must not exceed")
  expect_error(dfspa(F_hat, K = 2, grid_weights = c(1, 2, 3)),
               "length-`ncol\\(F_hat\\)`")
  expect_error(dfspa(F_hat, K = 2, grid_weights = c(-1, 1, 1, 1)),
               "non-negative")
})

test_that("grid_weights reshape the norm and drive first-iteration selection", {
  G <- 40
  x <- seq(0, 1, length.out = G)
  f1 <- ifelse(x < 0.5, 2, 0)        # mass on left half
  f2 <- ifelse(x < 0.5, 0, 2)        # mass on right half
  f3 <- rep(0.1, G)                   # small everywhere
  F_hat <- rbind(f1, f2, f3)

  # uniform: tied norms for f1, f2; which.max breaks the tie by index = 1
  fit_u <- dfspa(F_hat, K = 1, denoise = FALSE)
  expect_equal(fit_u$original_indices, 1L)

  # weights concentrated on the right half should make f2 win
  w_right <- ifelse(x < 0.5, 0.01, 1)
  fit_w <- dfspa(F_hat, K = 1, denoise = FALSE, grid_weights = w_right)
  expect_equal(fit_w$original_indices, 2L)
})

---
title: "When to prefer minmax_regret"
output:
  rmarkdown::html_vignette: default
vignette: >
  %\VignetteIndexEntry{When to prefer minmax_regret}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r mm-knit-opts, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)
```

```{r setup}
library(MetaHunt)
set.seed(1)
```

## What `minmax_regret()` does

`minmax_regret()` is a second top-level method exported by MetaHunt
that does **not** use study-level covariates `W`. It treats each
source-site function as a candidate target and returns the convex
combination of source functions that minimises worst-case regret
across the source pool. The framing is adversarial: rather than
predicting for a specific new metadata profile, it produces a
single aggregator that is robust to the worst-case choice of true
target inside the source convex hull. See Zhang, Huang, & Imai
(2024) for the underlying theory.

## When to use which

| Use `metahunt()` | Use `minmax_regret()` |
|---|---|
| You have many studies (`m` $\gtrsim 20$) | Few studies and/or unreliable metadata |
| Study-level covariates `W` are informative | No `W`, or `W` doesn't predict heterogeneity |
| You want predictions for **specific** new populations characterized by `W_0` | You want a single **worst-case** aggregator across the source pool |
| Random effects driven by latent structure | Adversarial / robustness framing |

## A small standalone simulation

```{r mm-simulate}
m <- 30; G <- 20; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(as.matrix(W) %*% beta); pi_true <- pi_true / rowSums(pi_true)
F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
```

```{r mm-wnew}
W_new <- data.frame(w1 = 0, w2 = 0)
```

## Running `minmax_regret()`

```{r mm-fit}
mm <- minmax_regret(F_hat)
mm
```

## Visual comparison: `minmax_regret` vs `metahunt`

To compare, fit `metahunt()` and predict at the single new
metadata row.

```{r mm-fit-metahunt}
fit <- metahunt(F_hat, W, K = K_true, dfspa_args = list(denoise = FALSE))
f_pred <- predict(fit, newdata = W_new)
```

Now overlay the two predictions on the shared grid.

```{r mm-vs-metahunt}
plot(x, mm$prediction, type = "l", col = "tomato", lwd = 2,
     ylim = range(c(mm$prediction, f_pred[1, ])),
     xlab = "x", ylab = expression(tilde(f)(x)),
     main = "minimax-regret vs. MetaHunt (target W = (0,0))")
lines(x, f_pred[1, ], col = "steelblue", lwd = 2)
legend("topright", legend = c("minmax_regret", "metahunt (W = (0,0))"),
       col = c("tomato", "steelblue"), lty = 1, lwd = 2, bty = "n")
```

## Small-`m` users

> With `m < 10`, both methods can still run but conformal
> intervals around `metahunt()` will be wide because the
> calibration set is small. We recommend reporting both
> `metahunt()` and `minmax_regret()` and treating large
> disagreement between them as a signal that the metadata is not
> informative enough to support a covariate-specific prediction —
> which is exactly the regime where the worst-case aggregator
> is the more honest answer.

## See also

- `vignette("metahunt-intro", package = "MetaHunt")` — the broader
  pipeline and assumptions.
- `vignette("conformal-prediction", package = "MetaHunt")` — for
  the small-`m` calibration caveat referenced above.
- `?minmax_regret` — the worst-case-regret aggregator.

---
title: "Understanding grid_weights and the L2(mu) norm"
output:
  rmarkdown::html_vignette: default
vignette: >
  %\VignetteIndexEntry{Understanding grid_weights and the L2(mu) norm}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r gw-knit-opts, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)
```

```{r setup}
library(MetaHunt)
set.seed(1)
```

## What `grid_weights` does

MetaHunt represents every study-level function by its values on a
shared grid `x_1, ..., x_G`. Distances between two functions `f` and
`h` are computed in a **weighted L^2 sense**:

```
||f - h||^2  =  sum_j  w_j * ( f(x_j) - h(x_j) )^2
```

The vector `w = grid_weights` is exactly that vector of weights — one
non-negative number per grid point. It defines the measure `mu` that
the inner product

```
<f, h>_{L^2(mu)}  =  sum_j  w_j * f(x_j) * h(x_j)
```

is taken with respect to. Every routine in MetaHunt that talks about
"distance between functions" or "norm of a function" — basis hunting,
constrained projection, conformal pointwise scores — uses the same
`grid_weights` you supply.

## When the default is fine

The default is `grid_weights = NULL`, which inside the package becomes
the uniform vector `rep(1 / G, G)`. Every grid point counts equally.

This is the right choice when **the grid itself is a representative
sample** from the population whose functional you care about. For
example, if you sub-sampled the grid from a held-out target-population
cohort with `build_grid()`, the empirical distribution of the grid
already encodes the target measure, and uniform `grid_weights` simply
averages over that empirical distribution.

In short: if your grid was sampled from population `P` and you want
distances to reflect `P`, leave `grid_weights = NULL`.

## When non-uniform `grid_weights` matter

You should override the default when the **grid distribution does not
match the population whose distances you care about**. Two common
scenarios:

- The grid was sampled from a *source* population (or a public
  reference like NHANES) but you want distances to reflect a *target*
  population. Pass `grid_weights` proportional to the target density
  evaluated at each grid point.
- You care about a sub-region of covariate space (e.g. only patients
  with `age > 65`). Set `grid_weights[j]` to `0` outside the
  sub-region and to the target density inside.

`grid_weights` does **not** need to sum to 1 — only its relative
magnitudes matter inside `dfspa()` and `project_to_simplex()`. The
default wrapper in `apply_wrapper()` divides by `sum(grid_weights)`,
so weighted means come out on the natural scale regardless of
normalisation.

## A worked comparison: uniform vs non-uniform

Here we generate the same data twice — `m = 40` studies, `G = 30`
grid points — and run `dfspa()` with two different `grid_weights`.
The first run is uniform; the second up-weights the right half of the
grid by a factor of about 5.

```{r sim-data}
m <- 40
G <- 30
x <- seq(0, 1, length.out = G)

# True bases on the grid.
basis <- rbind(sin(pi * x), cos(pi * x), x)

# Mixing weights driven by two latent covariates.
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
eta  <- as.matrix(W) %*% beta
pi_true <- exp(eta) / rowSums(exp(eta))

F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
dim(F_hat)
```

Now the two `grid_weights` choices:

```{r two-weights}
gw_uniform <- rep(1 / G, G)

# Up-weight x > 0.5 by ~5x. Renormalisation is optional but tidy.
gw_right   <- ifelse(x > 0.5, 5, 1)
gw_right   <- gw_right / sum(gw_right) * G   # comparable scale, optional

fit_unif  <- dfspa(F_hat, K = 3, grid_weights = gw_uniform, denoise = FALSE)
fit_right <- dfspa(F_hat, K = 3, grid_weights = gw_right,   denoise = FALSE)

# Index sets selected can differ.
fit_unif$original_indices
fit_right$original_indices
```

A single overlay plot of the two recovered first basis functions
makes the qualitative effect visible — the right-weighted run pays
more attention to the right half of the grid when ranking which
study is the "extreme" one to anchor:

```{r overlay-plot}
matplot(
  x,
  cbind(fit_unif$bases[1, ], fit_right$bases[1, ]),
  type = "l", lty = 1, lwd = 2,
  col  = c("#0072B2", "#D55E00"),
  xlab = "x",
  ylab = "first recovered basis",
  main = expression(
    "Recovered first-vertex basis under uniform vs. target-weighted " ~ L^2 * (mu)
  )
)
abline(v = 0.5, lty = 2, col = "grey60")
legend("topright",
       legend = c("uniform", "up-weight x > 0.5", "x = 0.5 boundary"),
       col    = c("#0072B2", "#D55E00", "grey60"),
       lty = c(1, 1, 2), lwd = c(2, 2, 1), bty = "n")
```

The two curves usually agree closely on the well-sampled regions and
diverge where the weights differ most. The takeaway is *not* that one
choice is right and the other wrong — it is that `grid_weights`
quietly determines which features of the function get prioritised.

## Practical guidance

- Start with the default. Only switch if you have a concrete reason
  (target distribution, sub-region focus, deliberate down-weighting
  of unreliable grid points).
- Make `grid_weights` strictly positive on the support you care
  about. Zero weight at a grid point removes it from every distance,
  norm, and mean computation in the package. This includes the
  d-fSPA denoising step, which uses the same L²(μ) inner product to
  define neighbourhoods.
- Keep the same `grid_weights` across `dfspa()`,
  `project_to_simplex()`, and the conformal routines for a given
  analysis. Mixing different `grid_weights` between steps is rarely
  what you want — it changes the meaning of "distance" mid-pipeline.
- If you only care about a scalar functional (e.g. an ATE), passing a
  `wrapper` is often a cleaner solution than re-weighting the grid.
  See `vignette("wrapper-scalar")`.

## Functions that accept `grid_weights`

The argument has the same meaning everywhere it appears:

- `dfspa(F_hat, K, grid_weights = NULL, ...)`
- `project_to_simplex(F_hat, bases, grid_weights = NULL, ...)`
- `apply_wrapper(F_mat, wrapper = NULL, grid_weights = NULL)` — used
  for the default weighted-mean wrapper.
- `metahunt(F_hat, W, K, ..., grid_weights = NULL)` — passes the
  weights through to `dfspa()` and `project_to_simplex()`.
- `split_conformal(...)`, `cross_conformal(...)`,
  `conformal_from_fit(...)` — all accept `grid_weights` for the
  pointwise conformity score and for the default wrapper when none is
  supplied.
- `reconstruction_error_curve()`, `cv_error_curve()`,
  `select_denoising_params()` — propagate `grid_weights` to the
  underlying `dfspa()` calls.

If a function does not accept `grid_weights` directly, it is because
it operates on already-projected weights `pi_hat` (e.g.
`fit_weight_model()`) and therefore does not need the grid measure.

---
title: "Preparing your data: from fitted models to F_hat"
output:
  rmarkdown::html_vignette: default
vignette: >
  %\VignetteIndexEntry{Preparing your data: from fitted models to F_hat}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r dp-knit-opts, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)
```

```{r setup}
library(MetaHunt)
set.seed(1)
```

## Why this step matters

Most users arrive at MetaHunt with **one fitted model per study** (a
random forest from each centre, a causal forest from each trial site,
a linear model from each cohort), not with the `m`-by-`G` numeric
matrix `F_hat` that the rest of the package consumes. The two onramp
helpers in MetaHunt — `build_grid()` and `f_hat_from_models()` —
bridge that gap.

The conceptual picture is simple. Pick a finite set of "patient
profiles" (the **grid**); evaluate every centre's model at every
profile; stack the resulting numeric vectors as rows. The output is
`F_hat`, where `F_hat[i, g]` is the prediction of centre `i`'s model
at grid point `g`.

This tutorial walks through that two-step flow and covers:

- the dispatch table `f_hat_from_models()` uses,
- a fully-worked dependency-free example with `lm`,
- the `predict_fn` escape hatch for custom model classes,
- a multi-dimensional grid,
- the sanity checks worth running before plugging the result into
  `metahunt()`.

## `build_grid()`

`build_grid()` constructs a data frame of grid points from any
**reference patient-level dataset**. The reference data should have
the same columns and live on the same scale as the data each centre's
model was trained on. Common choices, in rough order of preference:

1. **A held-out sample from the target population.** Best if you have
   one — even 50–500 rows is enough. The grid then is a draw from the
   measure you care about and the default uniform `grid_weights` is
   exactly right.
2. **A public dataset of the same population.** For clinical work,
   e.g. NHANES, MEPS, or a national survey. For political-science
   replications, a public benchmark like the World Values Survey or
   ANES individual-level files works.
3. **A small individual-level sample from a single source site**, if
   your privacy regime allows one site to share its covariates (but
   not outcomes). This is common in multi-site clinical trials where
   one site can provide a covariate sample under a covariate-only DUA.
4. **A synthetic grid based on summary statistics.** If no
   individual-level data are available anywhere, construct a grid by
   sampling from marginal distributions whose moments you have agreed
   summary statistics for. This is the most permissive option but the
   least faithful — distances over the grid will reflect the
   independence assumption you imposed.

Whatever you pick, **the grid must use the same column names and
encodings the per-site models were trained with** — `f_hat_from_models()`
will call `predict(model, newdata = grid)` and silently produce
nonsense if (say) factor levels disagree.

```{r build-grid-1d}
ref <- data.frame(age = rnorm(500, 60, 10),
                  bp  = rnorm(500, 130, 15),
                  bmi = rnorm(500, 28, 4))
grid <- build_grid(ref, n_grid = 50, seed = 1)
dim(grid)
head(grid)
```

If `n_grid` is `NULL` (or at least `nrow(reference_data)`), the full
reference data is returned unchanged. Otherwise `n_grid` rows are
sampled uniformly without replacement.

The empirical distribution of the returned grid implicitly defines
the measure `mu` used by the `L^2(mu)` inner product downstream. If
your grid is itself a representative sample of the population you
care about, the default uniform `grid_weights` is appropriate. If the
grid was sampled from one population but you want distances to
reflect another, pass non-uniform `grid_weights` proportional to the
target density at each grid point. See `vignette("grid-weights")`.

## `f_hat_from_models()`

`f_hat_from_models(models, grid)` takes a list of fitted model
objects and the grid you just built, and returns the `m`-by-`G_grid`
numeric matrix `F_hat`.

### Dispatch table

Internally, `f_hat_from_models()` dispatches on each model's class:

| Class                                                  | Call form                                            |
|--------------------------------------------------------|------------------------------------------------------|
| `ranger`                                               | `predict(model, data = grid)$predictions`            |
| `grf::causal_forest`, `regression_forest`              | `predict(model, newdata = grid)$predictions`         |
| anything else                                          | `as.numeric(predict(model, newdata = grid))`         |

The default branch covers `lm`, `glm`, `randomForest`, and most other
R model objects whose `predict()` method returns a numeric vector
when called with `newdata`.

#### `ranger` example (not run)

```{r ranger-example, eval = FALSE}
library(ranger)
centre_models <- lapply(centre_data_list,
                        function(d) ranger(y ~ ., data = d))
F_hat <- f_hat_from_models(centre_models, grid)
```

#### `grf::causal_forest` example (not run)

```{r grf-example, eval = FALSE}
library(grf)
centre_models <- lapply(centre_data_list,
                        function(d) causal_forest(d$X, d$Y, d$W))
F_hat <- f_hat_from_models(centre_models, grid)
```

These chunks are `eval = FALSE` so this vignette does not depend on
`ranger` or `grf` — installing either is unnecessary just to read the
tutorial.

### Worked `lm` example

To keep the tutorial fully reproducible without external
dependencies, here is a complete onramp using `lm()` as the
per-centre model. (This is the same `lm-onramp` flow that previously
appeared in the introductory vignette, broken out and annotated.)

```{r lm-onramp}
m <- 8
centre_meta <- data.frame(
  region     = factor(sample(c("N", "S", "E", "W"), m, replace = TRUE)),
  mean_age   = round(runif(m, 50, 70)),
  pct_female = round(runif(m, 0.4, 0.6), 2)
)

# Each centre fits a quadratic on a single covariate `x`.
make_centre_data <- function(i) {
  x <- runif(80, -1, 1)
  beta <- centre_meta$mean_age[i] / 60      # toy effect of metadata
  data.frame(x = x, y = beta * x + 0.3 * x^2 + rnorm(80, sd = 0.2))
}
centre_models <- lapply(seq_len(m), function(i)
  stats::lm(y ~ poly(x, 2), data = make_centre_data(i)))

# A 1-D grid in the centres' covariate space.
grid_centres  <- data.frame(x = seq(-1, 1, length.out = 30))
F_hat_centres <- f_hat_from_models(centre_models, grid_centres)

dim(F_hat_centres)        # 8 x 30
F_hat_centres[1:3, 1:5]
```

You now have `F_hat_centres` (`m × G_grid`) and `centre_meta`
(`m × p`), which are everything `metahunt()` needs.

### `predict_fn` for custom S4 / bespoke classes

If your fitted models are S4 objects, ensembles, or otherwise need
custom handling, override the dispatcher with `predict_fn`. The
function must accept `(model, grid)` and return a length-`nrow(grid)`
numeric vector:

```{r predict-fn-demo}
# Toy "model" that is just a list with a slope. predict_fn evaluates it.
fake_models <- lapply(seq_len(4), function(i)
  list(slope = i / 4, intercept = 0))

custom_predict <- function(model, grid) {
  model$intercept + model$slope * grid$x
}

F_hat_custom <- f_hat_from_models(fake_models, grid_centres,
                                  predict_fn = custom_predict)
dim(F_hat_custom)         # 4 x 30
```

This same pattern extends to S4 model objects: write a one-line
adapter that pulls predictions out and coerces to a numeric vector.

## Multi-dimensional grids (3 covariates)

The grid does not need to be one-dimensional. With more than one
patient-level covariate, sample the grid from a multivariate
reference dataset and let `predict()` evaluate at each row.

```{r multi-d-grid}
# A 3-covariate reference dataset and a sub-sampled grid.
ref3   <- data.frame(age = rnorm(400, 60, 10),
                     bp  = rnorm(400, 130, 15),
                     bmi = rnorm(400, 28,  4))
grid3  <- build_grid(ref3, n_grid = 25, seed = 1)
dim(grid3)

# Each centre fits an lm on (age, bp, bmi); slopes vary across centres
# so there is genuine cross-centre heterogeneity to recover.
set.seed(2)
m3 <- 8
centre_data3 <- lapply(seq_len(m3), function(i) {
  n_i <- 60
  age <- rnorm(n_i, 60, 10)
  bp  <- rnorm(n_i, 130, 15)
  bmi <- rnorm(n_i, 28, 4)
  # slopes vary across centres
  beta_age <- 0.02 + 0.03 * (i / m3)
  beta_bp  <- -0.01 + 0.02 * cos(pi * i / m3)
  beta_bmi <- 0.05 - 0.04 * (i / m3)
  y <- beta_age * age + beta_bp * bp + beta_bmi * bmi + rnorm(n_i, sd = 0.3)
  data.frame(age = age, bp = bp, bmi = bmi, y = y)
})
centre_models3 <- lapply(centre_data3, function(d) stats::lm(y ~ age + bp + bmi, data = d))

F_hat3 <- f_hat_from_models(centre_models3, grid3)
dim(F_hat3)               # 8 x 25
```

The number of grid points `G_grid` is yours to choose. Larger grids
give finer-resolution function estimates; smaller grids run faster.
A few dozen to a few hundred is typical.

## Sanity checks

After building `F_hat`, run these quick checks before passing it
into the rest of the pipeline:

```{r sanity}
# Right shape: m studies x G grid points.
dim(F_hat3)

# Numeric, no NA.
is.numeric(F_hat3)
anyNA(F_hat3)

# Rows look like functions of similar magnitude (large outliers can
# dominate d-fSPA's `Delta`).
summary(apply(F_hat3, 1, function(r) c(min = min(r), max = max(r))))
```

A quick visual check on a 1-D grid is also worth the cost:

```{r sanity-plot}
matplot(grid_centres$x, t(F_hat_centres), type = "l", lty = 1,
        col = "grey50",
        xlab = "x", ylab = expression(hat(f)(x)),
        main = "Per-centre fitted functions on the shared grid")
```

If a single centre's curve is wildly far from the rest, that often
flags a bug (mis-scaled covariate, an `lm` with the wrong response)
rather than a real low-rank-violating outlier.

## Common pitfalls

- **Different columns across centres.** Every centre's model must
  accept rows of the same `grid` data frame. If centre A was fit on
  `(age, bp)` and centre B was fit on `(age, bmi)`, you cannot share
  a grid; harmonise the covariates first.
- **Categorical levels.** Factor levels in `grid` must match those
  the model saw at training. Pre-build levels explicitly with
  `factor(x, levels = ...)` if needed.
- **Wrong return shape from `predict_fn`.** Must be a length-`G_grid`
  numeric vector. `f_hat_from_models()` errors if the length is
  wrong or any value is `NA`.
- **Using a non-representative grid.** The grid implicitly defines
  the measure `mu`. A grid concentrated in one corner of covariate
  space will under-weight other regions. Either resample the grid
  more evenly, or use non-uniform `grid_weights` (see
  `vignette("grid-weights")`).
- **Confusing `m × p` study covariates `W` with the patient grid.**
  `W` lives at the *study* level (one row per centre, columns like
  region or year). The grid lives at the *patient* level (one row
  per profile, columns like age and BMI). They are different
  objects.

---
title: "Scalar summaries with wrapper"
output:
  rmarkdown::html_vignette: default
vignette: >
  %\VignetteIndexEntry{Scalar summaries with wrapper}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r ws-knit-opts, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)
```

```{r setup}
library(MetaHunt)
set.seed(1)
```

## Why scalar summaries

`metahunt()` predicts a function on the grid; conformal routines can
return a band at every grid point. Often, though, the inferential
target is a single number derived from that function:

- The average treatment effect (mean of a CATE function over a
  reference patient distribution).
- The treatment effect at a specific patient profile.
- The fraction of the population with a positive treatment effect.
- A contrast between two endpoints.

For all of these, MetaHunt accepts a `wrapper` argument that
**collapses the predicted function to a scalar** before any further
calculation. The same wrapper is applied identically to predictions
and to calibration residuals, so conformal coverage transfers
directly to the scalar summary.

## The wrapper protocol

`apply_wrapper(F_mat, wrapper, grid_weights)` defines the contract.

- `F_mat` is an `n`-by-`G_grid` numeric matrix; row `j` is one
  function on the grid.
- If `wrapper` is `NULL`, `apply_wrapper()` returns the weighted mean
  of each row using `grid_weights` (uniform `1/G_grid` by default),
  divided by `sum(grid_weights)`.
- If `wrapper` is a function, `apply_wrapper()` calls `apply(F_mat,
  1, wrapper)`, which means **the wrapper receives a single numeric
  vector of length `G_grid`** — one row of `F_mat` at a time — and
  must return a single numeric value.

The contract therefore is:

```
wrapper :: numeric vector of length G_grid  ->  numeric scalar
```

Any function satisfying that signature is a valid wrapper. The
package then enforces post-hoc that the result is numeric and has
exactly one entry per row.

## An ATE example with `grf::causal_forest`

We simulate a multi-site clinical trial with `m = 8` sites. Each
site has its own individual-level data $(Y, X, T)$ where $Y$ is a
continuous outcome, $X$ is a single patient covariate (`age`), and
$T$ is binary treatment. The site-level CATE function
$\tau^{(i)}(\text{age}) = E[Y(1) - Y(0) \mid \text{age}, \text{site} = i]$
varies across sites in a way that depends on the site's metadata.
Each site fits its own `grf::causal_forest` on its individual-level
data, and shares only the fitted model — not the patient data —
with us.

```{r ws-simulate-trials, eval = requireNamespace("grf", quietly = TRUE)}
m <- 8
n_per_site <- 200
G <- 30

W <- data.frame(
  year        = sample(2010:2020, m, replace = TRUE),
  pct_treated = round(runif(m, 0.3, 0.6), 2)
)

site_data_list <- lapply(seq_len(m), function(i) {
  age <- runif(n_per_site, 30, 80)
  T   <- rbinom(n_per_site, 1, W$pct_treated[i])
  site_eff <- (W$year[i] - 2015) / 5   # site-level shift in CATE
  tau_age  <- 0.02 * (age - 50) + site_eff
  Y0  <- 0.01 * age + rnorm(n_per_site, sd = 0.5)
  Y1  <- Y0 + tau_age
  Y   <- ifelse(T == 1, Y1, Y0)
  data.frame(Y = Y, age = age, T = T)
})

grid <- data.frame(age = seq(30, 80, length.out = G))
```

Each site fits its own `causal_forest`. We use `num.trees = 200` to
keep the vignette fast; in practice you would use the default 2000
or more.

```{r ws-fit-cf, eval = requireNamespace("grf", quietly = TRUE)}
cf_models <- lapply(site_data_list, function(d)
  grf::causal_forest(X = matrix(d$age, ncol = 1),
                     Y = d$Y,
                     W = d$T,
                     num.trees = 200))
```

We stack the per-site CATE estimates on the shared `age` grid into
the `m`-by-`G` matrix `F_hat`. Here we pass an explicit `predict_fn`
to illustrate the general pattern; the dispatch table inside
`f_hat_from_models()` already knows how to call `causal_forest`, so
for users on standard `grf::causal_forest`, the default `predict_fn`
is sufficient and you can omit the `predict_fn` argument.

```{r ws-build-fhat, eval = requireNamespace("grf", quietly = TRUE)}
cate_predict <- function(model, grid) {
  as.numeric(stats::predict(model, newdata = matrix(grid$age, ncol = 1))$predictions)
}
F_hat <- f_hat_from_models(cf_models, grid, predict_fn = cate_predict)
dim(F_hat)
```

We now fit `metahunt()` on `(F_hat, W)` and ask for the predicted
ATE at a hypothetical new site.

```{r ws-fit-metahunt, eval = requireNamespace("grf", quietly = TRUE)}
fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
W_new <- data.frame(year = 2018, pct_treated = 0.45)
ate_pred <- predict(fit, newdata = W_new, wrapper = mean)
ate_pred
```

The scalar `ate_pred` is the predicted average treatment effect for
a hypothetical new site with metadata `(year = 2018, pct_treated =
0.45)`, taking the unweighted mean over the 30-point age grid.

## Three custom wrappers

Below are three short, self-contained wrappers, each illustrating a
different idea. All three are applied to the `F_hat`, `fit`, and
`W_new` constructed in the previous section.

### Plain `mean`

`mean` is already a function `numeric -> numeric`, so it is a valid
wrapper. With a uniform grid this is just the unweighted average of
the function over the grid — i.e. the grid-uniform ATE.

```{r wrapper-mean, eval = requireNamespace("grf", quietly = TRUE)}
predict(fit, newdata = W_new, wrapper = mean)
```

### Restricted positive mean

Suppose we only credit treatment effects that are positive (for
example, in a cost-effectiveness setting). The wrapper averages
`max(f(x), 0)` over the grid:

```{r wrapper-restricted, eval = requireNamespace("grf", quietly = TRUE)}
restricted_pos_mean <- function(f) sum(pmax(f, 0)) / length(f)
predict(fit, newdata = W_new, wrapper = restricted_pos_mean)
```

Because every row of `F_mat` is passed in turn, `f` inside the
wrapper is just a numeric vector of length `G_grid`. `length(f)` is
therefore the grid size, and dividing by it gives a uniform-weighted
average.

### Endpoint contrast

The difference `f(x_G) - f(x_1)` is a useful summary when the grid
is ordered (e.g. age, dose, or time). For our age grid it is the
gap in CATE between an 80-year-old and a 30-year-old patient at the
new site:

```{r wrapper-endpoint, eval = requireNamespace("grf", quietly = TRUE)}
endpoint_contrast <- function(f) f[length(f)] - f[1]
predict(fit, newdata = W_new, wrapper = endpoint_contrast)
```

## Conformal coverage with a wrapper

When you pass `wrapper` into `split_conformal()` (or
`cross_conformal()`, or `conformal_from_fit()`), conformity scores
are computed *after* the wrapper, on a **single shared quantile**.
The interval covers the wrapped scalar with the nominal level — not
the underlying function pointwise.

With only `m = 8` sites, we hold out a single site (the 8th) and
use the other seven for training plus calibration. The calibration
set is small, so we use `alpha = 0.1` rather than `0.05`.

```{r ws-split-scalar, eval = requireNamespace("grf", quietly = TRUE)}
# Use 7 sites for training+calibration, predict for the held-out 8th
tr_cal <- 1:7; new <- 8
res <- split_conformal(
  F_hat[tr_cal, , drop = FALSE],
  W[tr_cal, , drop = FALSE],
  W[new, , drop = FALSE],
  K = 3, wrapper = mean, alpha = 0.1, cal_frac = 0.5, seed = 1,
  dfspa_args = list(denoise = FALSE)
)
data.frame(prediction = res$prediction,
           lower      = res$lower,
           upper      = res$upper)
```

With only 8 sites in this realistic example, an empirical-coverage
check on a single held-out site is not informative — for coverage
diagnostics, use a leave-one-out loop or simulate a larger study
count. See `?coverage` for the helper function and the
`conformal-prediction` vignette for split-conformal at scale.

## Pointwise vs scalar — quick reference

| Aspect              | Pointwise (`wrapper = NULL`)                       | Scalar (`wrapper` supplied)                         |
|---------------------|----------------------------------------------------|-----------------------------------------------------|
| Output shape        | `nrow(W_new)` x `G_grid` matrix                    | length-`nrow(W_new)` numeric vector                 |
| Conformal quantile  | one per grid point (length-`G_grid`)               | a single scalar                                     |
| Coverage guarantee  | per grid point, marginally (not joint over grid)   | for the scalar summary, marginally                  |
| Best for            | visualising the predicted function with a band     | reporting a single number with a valid CI           |
| Example call        | `split_conformal(F, W, W_new, K = 3)`              | `split_conformal(F, W, W_new, K = 3, wrapper = mean)` |

A pointwise band is a visualisation aid; a scalar interval is the
right object for an inferential claim about a specific functional.
Pick the wrapper that matches the question you actually want to
answer, and let the conformal machinery do the rest.

## See also

- `vignette("data-prep")` — building `F_hat` from per-site fitted
  models (including the `grf::causal_forest` dispatch and the
  `predict_fn` escape hatch used here).
- `vignette("conformal-prediction")` — split- and cross-conformal
  routines at scale, including empirical-coverage diagnostics that
  need more than a handful of held-out sites.

---
title: "Get started with MetaHunt"
output:
  rmarkdown::html_vignette: default
vignette: >
  %\VignetteIndexEntry{Get started with MetaHunt}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r gs-knitr-opts, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)
```

```{r setup}
library(MetaHunt)
set.seed(1)
```

## What this is

MetaHunt is a privacy-preserving meta-analysis tool. You start with
per-study estimated functions and a row of metadata for each study,
and end with a predicted function (or scalar summary) for a new study,
together with a conformal prediction interval. This vignette shows the
shortest possible end-to-end run; for the assumptions, the three-step
pipeline, and tuning, see
`vignette("metahunt-intro", package = "MetaHunt")`.

## Inputs in 30 seconds

Two ingredients drive every function in the package.

| Object | Description | Example |
|---|---|---|
| `F_hat` | An `m`-by-`G` numeric matrix. Row `i` is the function of study `i` evaluated on a *shared* set of `G` patient-level grid points. | `F_hat[i, g]` is the predicted treatment effect for centre `i` at "patient profile" `g`. |
| `W`     | An `m`-by-`p` matrix or data frame of study-level covariates (one row per study). | Region, mean age, percent female, year. |

> **Starting from fitted models?** If your inputs are one fitted model
> per study (e.g. `ranger`, `grf::causal_forest`, `lm`), turn them into
> `F_hat` first with `f_hat_from_models()` after building a shared
> evaluation grid via `build_grid()`. See
> `vignette("data-prep", package = "MetaHunt")` for a worked example.

## A minimal end-to-end run

Simulate a small problem so the vignette has no external data
dependency:

```{r gs-simulate}
m <- 30; G <- 20; K_true <- 2
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), x)                        # 2 true bases on the grid

W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.5), c(-0.5, 1))
pi_true <- exp(as.matrix(W) %*% beta)
pi_true <- pi_true / rowSums(pi_true)

F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
W_new <- data.frame(w1 = c(0, 1, -1), w2 = c(0, -0.5, 1))  # three new metadata profiles
```

Fit and get a conformal interval for the scalar wrapper `mean` (the
ATE under uniform grid weights):

```{r gs-fit-conformal}
res <- split_conformal(F_hat, W, W_new, K = K_true,
                       wrapper = mean, alpha = 0.1, seed = 1,
                       dfspa_args = list(denoise = FALSE))
data.frame(prediction = res$prediction,
           lower      = res$lower,
           upper      = res$upper)
```

## Visualising the result

`plot()` on the conformal object renders the prediction with its
band:

```{r gs-plot}
plot(res)
```

## Where to next

- `vignette("metahunt-intro", package = "MetaHunt")` — the four
  assumptions, the three-step pipeline, choosing `K`, denoising,
  conformal flavours, and `minmax_regret()`.
- `vignette("data-prep", package = "MetaHunt")` — turning per-centre
  fitted models into `(F_hat, W)`.
- `?metahunt` — the full pipeline reference.

---
title: "Choosing K and the denoising parameters"
output:
  rmarkdown::html_vignette: default
vignette: >
  %\VignetteIndexEntry{Choosing K and the denoising parameters}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r cdk-knit-opts, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)
```

```{r setup}
library(MetaHunt)
set.seed(1)
```

This vignette focuses on two practical knobs in the MetaHunt
pipeline: the latent rank `K` and the d-fSPA denoising parameters
`(N, Delta)`. For the broader setup — the four assumptions, the
three-step pipeline, and the running notation — see
`vignette("metahunt-intro", package = "MetaHunt")`.

## Why this matters

Choosing `K` is the single most consequential decision in a MetaHunt
fit. Picking `K` too small underfits: real cross-study heterogeneity
gets squashed into a low-rank approximation that cannot represent the
data, and downstream predictions are biased. Picking `K` too large
inflates variance and risks recovering spurious "bases" that fit
noise. The denoising step in d-fSPA controls finite-sample variance
in a complementary way: averaging each study with its near
neighbours before basis hunting smooths over per-study estimation
error, at the cost of a small smoothing bias.

## A small standalone simulation

```{r cdk-simulate}
m <- 30; G <- 20; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(as.matrix(W) %*% beta); pi_true <- pi_true / rowSums(pi_true)
F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
```

## Unsupervised diagnostic: reconstruction error vs K

The **elbow** plot tracks how well the recovered bases reconstruct
the observed `F_hat` as a function of `K`. It is unsupervised — it
does not use `W` — and is fast.

```{r cdk-elbow}
elbow <- reconstruction_error_curve(F_hat, K_range = 2:6,
                                    dfspa_args = list(denoise = FALSE))
plot(elbow$K, elbow$error, type = "b",
     xlab = "K", ylab = "reconstruction error",
     main = "Reconstruction error vs K",
     ylim = c(0, max(elbow$error, na.rm = TRUE) * 1.05))
```

## Supervised diagnostic: cross-validated prediction error vs K

The **CV prediction-error** curve uses the metadata `W` to predict
held-out studies' functions and reports the average prediction
error. This is supervised and tends to identify a tighter elbow when
the metadata is informative.

```{r cdk-cv}
cv <- cv_error_curve(F_hat, W, K_range = 2:6, n_folds = 4,
                     dfspa_args = list(denoise = FALSE), seed = 1)
plot(cv$K, cv$cv_error, type = "b",
     xlab = "K", ylab = "CV prediction error",
     main = "CV prediction error vs K",
     ylim = c(0, max(cv$cv_error, na.rm = TRUE) * 1.05))
```

Both curves should dip near `K = 3`, the true rank in this
simulation.

## The d-fSPA denoising knobs (N, Delta)

`dfspa()` averages each study with its near neighbours before
running the projection algorithm. Two parameters control this:
`N` (the neighbourhood size, in number of studies) and `Delta` (a
distance threshold). Larger `N` and `Delta` smooth more aggressively.

### Bypassing denoising

In clean simulations or with small `m`, the simplest choice is to
bypass denoising entirely. This avoids the small-sample failure mode
where aggressive denoising prunes too many studies.

```{r cdk-bypass}
fit_no <- metahunt(F_hat, W, K = K_true,
                   dfspa_args = list(denoise = FALSE))
fit_no
```

### Setting (N, Delta) by hand

If you have a sense of scale for the within-study estimation error,
pass `N` and `Delta` directly. These two calls illustrate a
hand-tuned and a near-default configuration on the same data.

```{r cdk-dfspa-knobs}
fit_no <- metahunt(F_hat, W, K = K_true,
                   dfspa_args = list(denoise = FALSE))

fit_manual <- metahunt(F_hat, W, K = K_true,
                       dfspa_args = list(N = 0.5 * log(nrow(F_hat)),
                                         Delta = 0.4))
```

### Tuning (N, Delta) by CV

`select_denoising_params()` cross-validates over a grid of
`(N, Delta)` combinations at fixed `K`. With small `m`, the search
will frequently warn that some combinations prune everything ("Only
0 studies survive denoising but K = 3..."). These warnings are
expected: aggressive `(N, Delta)` on small training folds is too
strong. The function records those folds as failures and returns
the best surviving combination.

```{r cdk-select-denoising, warning = FALSE}
tune <- select_denoising_params(F_hat, W, K = K_true, n_folds = 4, seed = 1)
tune$best
```

## Practical recipe

- Start with the elbow plot to get a rough range for `K`. Refine
  with the CV curve if `W` is informative.
- For very small `m` (say `m < 30`), bypass denoising
  (`denoise = FALSE`) and pick `K` from the CV curve.
- For larger `m`, leave the d-fSPA defaults on or tune
  `(N, Delta)` with `select_denoising_params()`.
- Treat warnings from `select_denoising_params()` as informative,
  not fatal. The reported `best` is the best surviving combination.
- Sanity-check the recovered bases visually with `plot(fit)`.
  Bases that look like noise are a sign of `K` set too high.

## See also

- `vignette("metahunt-intro", package = "MetaHunt")` — the full
  pipeline and key assumptions.
- `?metahunt` — the wrapper around the three pipeline steps.
- `?dfspa` — d-fSPA basis hunting and its denoising arguments.
- `?reconstruction_error_curve` — the unsupervised elbow diagnostic.
- `?cv_error_curve` — the supervised CV diagnostic.
- `?select_denoising_params` — cross-validating `(N, Delta)` at
  fixed `K`.

---
title: "An introduction to MetaHunt"
output:
  rmarkdown::html_vignette: default
vignette: >
  %\VignetteIndexEntry{An introduction to MetaHunt}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r intro-knitr-opts, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)
```

```{r setup}
library(MetaHunt)
set.seed(1)
```

## The problem: function-valued meta-analysis

You have several **studies** — trial centres, cohorts, sites — and each
one has produced a fitted model for some function of patient-level
covariates. That function might be a regression curve, a CATE, or a
dose-response curve. You also have **study-level metadata** for each
study: region, sample composition, year, and so on. You want to predict
that function for a *new* study, characterised only by its metadata,
together with a prediction band, and you want to do it without ever
pooling patient-level data across studies.

MetaHunt is designed for exactly that workflow. It assumes a small
number of latent **basis functions** drive cross-study heterogeneity,
recovers them from the per-study estimates, and learns how the mixing
weights depend on study-level metadata. The result is a single fitted
object that you can `predict()` for new metadata and combine with
conformal prediction for uncertainty quantification. If you would
rather see one short example before reading the full pipeline, start
with `vignette("get-started", package = "MetaHunt")`.

## Key assumptions

The MetaHunt pipeline rests on four assumptions. We state them in
applied-reader voice; the manuscript contains the formal statements
and proofs.

### A1. Low-rank cross-study heterogeneity

We assume the true study-level functions all live inside the convex
hull of a small number $K$ of shared *latent basis functions*. Each
study is a convex combination of the bases, with non-negative weights
that sum to one. Geometrically, every study is a point inside a
$(K-1)$-simplex whose vertices are the basis functions; the bases are
identifiable as the vertices of that hull.

Formally, there exist basis functions $g_1, \ldots, g_K$ with $K < m$ such that, for every study $i \in \{0, 1, \ldots, m\}$,
$$f^{(i)}(\boldsymbol{x}) \;=\; \sum_{k=1}^{K} \pi_{ik}\, g_k(\boldsymbol{x}), \qquad \boldsymbol{x} \in \mathcal{X},$$
with weight vector $\boldsymbol{\pi}_i = (\pi_{i1}, \ldots, \pi_{iK})^\top \in \Delta_{K-1}$, the $(K-1)$-simplex. The bases are non-degenerate, i.e. $g_2 - g_1, \ldots, g_K - g_1$ are linearly independent in $L^2(\mu)$.

*What this buys you:* the entire $m$-by-grid table of study functions
collapses to $K$ shared shapes plus an $m$-by-$K$ table of weights, so
heterogeneity becomes low-dimensional and shareable.

### A2. Weight model

The mixing weight $\boldsymbol{\pi}_i$ for study $i$ is drawn from
some conditional distribution given that study's covariates
$\boldsymbol{W}_i$. The distribution can be anything you can
estimate — a Dirichlet regression, an RKHS-Dirichlet, a multinomial
logit, a nearest-neighbour smoother. MetaHunt does not commit you to a
specific functional form.

Formally, for $i = 0, 1, \ldots, m$, the weight vectors are independent draws
$$\boldsymbol{\pi}_i \mid \boldsymbol{W}_i \;\stackrel{\text{ind.}}{\sim}\; \mathcal{P}_{\boldsymbol{\pi} \mid \boldsymbol{W}}(\,\cdot \mid \boldsymbol{W}_i),$$
where $\mathcal{P}_{\boldsymbol{\pi}\mid\boldsymbol{W}}$ is an arbitrary distributional map from the covariate space $\mathcal{W}$ to the simplex $\Delta_{K-1}$.

*What this buys you:* once you can map metadata to weights, predicting
the function for a brand-new target population reduces to predicting
its weights and re-mixing the recovered bases.

### A3. Exchangeability

The studies (including the target) are exchangeable units drawn from a
common data-generating process: their joint distribution is invariant
under reordering. This is much weaker than i.i.d. and is the standard
condition under which conformal prediction is valid.

Formally, the site-level covariates $\boldsymbol{W}_0, \boldsymbol{W}_1, \ldots, \boldsymbol{W}_m$ are exchangeable. Combined with A1–A2, this implies that the triples
$$\bigl(\boldsymbol{W}_i,\, \boldsymbol{\pi}_i,\, f^{(i)}\bigr), \qquad i = 0, 1, \ldots, m,$$
are jointly exchangeable across $i$.

*What this buys you:* the calibration step in conformal prediction
inherits a finite-sample, distribution-free coverage guarantee with
no need for parametric error models.

### A4. Estimation error control

Each study reports a noisy version of its true function,
$\hat f^{(i)} = f^{(i)} + \epsilon^{(i)}$. We assume the within-study
estimation error vanishes uniformly in the within-study sample size,
and that the number of studies $m$ does not grow too fast relative to
those sample sizes. In words: every study should be reasonably
well-estimated, and you should not have a large number of studies
whose individual sample sizes $n_i$ are small.

Formally, write $\hat f^{(i)}(\boldsymbol{x}) = f^{(i)}(\boldsymbol{x}) + \epsilon^{(i)}(\boldsymbol{x})$. We assume
$$\sup_{\boldsymbol{x} \in \mathcal{X}} \mathbb{E}\!\left[\epsilon^{(i)}(\boldsymbol{x})^2\right] = O\!\left(n_i^{-r}\right) \quad \text{for some } r > 0,$$
and that $m = o\bigl(\inf_i n_i^{\,a}\bigr)$ for some $0 < a < r$, where $n_i$ is study $i$'s sample size.

*What this buys you:* the noise in $\hat F$ does not accumulate
through the pipeline, basis recovery is consistent, and conformal
intervals retain their asymptotic coverage despite a multi-stage
estimator.

## The three-step pipeline

`metahunt()` is a thin wrapper around three exported steps. You can
also call them individually if you want fine-grained control.

### Step 1. Basis hunting via d-fSPA

`dfspa()` extends the Successive Projection Algorithm to functions.
It iteratively picks the study whose current residual norm is largest,
projects the rest onto the orthogonal complement, and repeats $K$
times. A denoising step averages each study with its near neighbours
before the search; this trades a small bias for substantially smaller
variance when the per-study estimates are noisy.

### Step 2. Fitting weight model

For each study, `project_to_simplex()` finds the convex weights that
best reconstruct its observed function from the recovered bases.
`fit_weight_model()` then regresses these weight vectors on the
study-level covariates `W` using a method of your choice (default:
Dirichlet regression).

### Step 3. Target prediction

`predict.metahunt()` takes a new metadata row $W_0$, predicts its
weight vector through the fitted weight model, and returns the convex
combination of the recovered bases. With `wrapper = mean` (or any
other reduction) it returns a scalar summary — for example, an ATE
under a uniform grid weighting.

## A worked end-to-end example

For the rest of the vignette we work with a simulated `(F_hat, W)` so
the truth is known. In your own data, replace this block with the
data-prep onramp described in `vignette("data-prep")`.

```{r simulate}
G <- 40; m <- 120; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)         # 3 true bases on the grid

W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))       # study-level covariates
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(as.matrix(W) %*% beta)
pi_true <- pi_true / rowSums(pi_true)

F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
dim(F_hat)
```

In real data, `F_hat[i, ]` would be `predict(model_i, newdata = grid)`
and `W[i, ]` would be that centre's metadata. The data-prep vignette
describes a one-line `lm`-based onramp.

Fit the full pipeline and inspect the recovered bases:

```{r fit}
fit <- metahunt(F_hat, W, K = 3)
fit
```

```{r plot-bases}
plot(fit, x_axis = x,
     col = c("#0072B2", "#D55E00", "#009E73"))
```

Predict the target functions for three new metadata profiles and plot
them:

```{r predict}
W_new <- data.frame(w1 = c(0, 1, -1), w2 = c(0, -0.5, 1),
                    row.names = c("baseline", "high w1, low w2", "low w1, high w2"))
f_pred <- predict(fit, newdata = W_new)
dim(f_pred)
oldpar <- par(mar = c(4, 4.5, 3, 1))
matplot(x, t(f_pred), type = "l", lty = 1,
        col = c("#0072B2", "#D55E00", "#009E73"),
        xlab = "x", ylab = expression(tilde(f)(x)),
        main = "Predicted target functions")
legend("topright", legend = rownames(W_new),
       col = 1:3, lty = 1, bty = "n")
par(oldpar)
```

Pass a `wrapper` for a scalar summary per target (an ATE under uniform
grid weights):

```{r predict-wrapped}
predict(fit, newdata = W_new, wrapper = mean)
```

## Where to next

- `vignette("data-prep")` — turning per-centre fitted models into
  `(F_hat, W)` via `build_grid()` and `f_hat_from_models()`,
  including a self-contained `lm` onramp.
- `vignette("grid-weights")` — choosing the `grid_weights` argument
  and the underlying $L^2(\mu)$ inner product.
- `vignette("choosing-k-denoising")` — picking the rank $K$ and the
  d-fSPA denoising knobs `(N, Delta)` via elbow and CV diagnostics.
- `vignette("conformal-prediction")` — split, cross, and from-fit
  conformal bands around the target function.
- `vignette("wrapper-scalar")` — using `wrapper = mean` and other
  reductions to turn function-valued predictions into ATEs and
  related scalar estimands.
- `vignette("minmax-baseline")` — when to prefer the covariate-free
  worst-case-regret aggregator and how it compares to MetaHunt.
- Function-level references: `?metahunt`, `?dfspa`,
  `?fit_weight_model`, `?predict_target`.

## References

Shi, W., Imai, K., and Zhang, Y. (2024). *Privacy-preserving
meta-analysis through low-rank basis hunting.*

Zhang, Y., Huang, M., and Imai, K. (2024). *Minimax regret estimation
for generalizing heterogeneous treatment effects with multisite data.*
arXiv:2412.11136.

---
title: "Conformal prediction with different choices"
output:
  rmarkdown::html_vignette: default
vignette: >
  %\VignetteIndexEntry{Conformal prediction with different choices}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r cp-knit-opts, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6, fig.height = 4,
  fig.align = "center"
)
```

```{r setup}
library(MetaHunt)
set.seed(1)
```

This vignette covers the three conformal-prediction interfaces
exported by MetaHunt. The validity of all three rests on the
exchangeability assumption A3 in
`vignette("metahunt-intro", package = "MetaHunt")` §"Key
assumptions"; we do not re-derive it here.

## Why conformal prediction here

Conformal prediction wraps any black-box prediction rule and
produces a band around its forecast that, on average across new
studies, will contain the truth at least `(1 - alpha)` of the time.
The key word is *marginal*: the guarantee is over the random draw
of the new study, not conditional on a specific covariate value.
All you need is for the calibration data to be exchangeable with
the new study (assumption A3) — no distributional assumptions on
the noise or on the weight model.

## A small standalone simulation

```{r cp-simulate}
# m = 80 is large enough that with cal_frac = 0.5 and alpha = 0.05 the conformal quantile is finite.
m <- 80; G <- 20; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(as.matrix(W) %*% beta); pi_true <- pi_true / rowSums(pi_true)
F_hat <- pi_true %*% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
```

```{r cp-wnew}
W_new <- data.frame(w1 = c(0, 1, -1), w2 = c(0, -0.5, 1))
```

## The three flavours

All three return an object you can plot directly with `plot()`.

| Function | When to use |
|---|---|
| `split_conformal()` | Default. One train/calibration split. Fastest; some variance from the random split. |
| `cross_conformal()` | Many studies, want lower split-induced variance. Refits the pipeline `n_folds + 1` times. |
| `conformal_from_fit()` | You've already fit a pipeline (e.g. after tuning `K`) and want intervals without refitting. See `?conformal_from_fit`. |

## Split conformal

`split_conformal()` does a single train / calibration split. With
small `m`, set `cal_frac = 0.5` so the calibration set is large
enough for the chosen `alpha`.

```{r cp-split-pointwise}
res_pw <- split_conformal(F_hat, W, W_new, K = K_true, alpha = 0.05,
                          cal_frac = 0.5, seed = 1,
                          dfspa_args = list(denoise = FALSE))
plot(res_pw, target_idx = 1, x_axis = x)
```

The shaded region is the pointwise 95% conformal band; it has finite width because `n_cal = 40` is large enough for `α = 0.05`.

```{r cp-split-scalar}
res_scalar <- split_conformal(F_hat, W, W_new, K = K_true,
                              wrapper = mean, alpha = 0.05,
                              cal_frac = 0.5, seed = 1,
                              dfspa_args = list(denoise = FALSE))
data.frame(prediction = res_scalar$prediction,
           lower      = res_scalar$lower,
           upper      = res_scalar$upper)
```

## Cross conformal

`cross_conformal()` reduces the variance of the band that comes
from the random split, at the cost of refitting `n_folds + 1`
times.

```{r cp-cross}
res_cross <- cross_conformal(F_hat, W, W_new, K = K_true, n_folds = 4,
                             wrapper = mean, alpha = 0.1, seed = 1,
                             dfspa_args = list(denoise = FALSE))
res_cross
```

## Pre-fit conformal

If you have already run `metahunt()` (for instance after tuning
`K`) and do not want to refit, `conformal_from_fit()` recycles the
existing fit to produce calibrated intervals. The example below
re-uses the training data as the calibration set *for demonstration
only*; in real use, hold out a separate calibration set so the
exchangeability argument applies to genuinely unseen studies.

```{r cp-prefit}
fit <- metahunt(F_hat, W, K = K_true, dfspa_args = list(denoise = FALSE))
pi_hat <- project_to_simplex(F_hat, fit$dfspa_fit$bases)
res_pre <- conformal_from_fit(
  dfspa_fit = fit$dfspa_fit, weight_model = fit$weight_model,
  F_cal = F_hat, W_cal = W, W_new = W_new,
  wrapper = mean, alpha = 0.1
)
res_pre
```

## Pointwise vs scalar bands

A pointwise band returns a `(1 - alpha)` interval at each grid
point but does not give a joint guarantee across grid points: the
probability that the truth lies inside the entire band
simultaneously is generally lower than `1 - alpha`. A scalar
wrapper (e.g. `wrapper = mean`) collapses the function to a single
number and gives one calibrated interval, which is the right tool
for joint inferential claims. If you need a joint coverage statement across the grid, either apply a multiple-testing correction (e.g. divide α by `G`) or replace the pointwise band with a scalar wrapper — see `vignette('wrapper-scalar', package = 'MetaHunt')`.

## Small-`m` warning on `cal_frac`

> With too-few calibration studies for the chosen `alpha`, the
> conformal quantile is `Inf` and intervals are unbounded. The
> finite-sample formula needs
> `n_cal >= ceiling((1 - alpha)(n_cal + 1))` calibration studies;
> below that threshold the package warns and the bands degenerate.
> Either raise `cal_frac`, raise `alpha`, or switch to
> `cross_conformal()`.

Below we deliberately reuse only the first 30 of our 80 studies so the calibration set is too small for `α = 0.05`. The package issues a warning and returns `Inf` quantiles; the corresponding intervals are unbounded. The fix is to either supply more studies, raise `α`, or raise `cal_frac`.

```{r cp-small-m-warning, warning = TRUE}
m_small <- 30  # too small for alpha = 0.05 with cal_frac = 0.5
F_small <- F_hat[1:m_small, , drop = FALSE]
W_small <- W[1:m_small, , drop = FALSE]
res_inf <- split_conformal(F_small, W_small, W_new, K = K_true,
                           alpha = 0.05, cal_frac = 0.5, seed = 1,
                           dfspa_args = list(denoise = FALSE))
res_inf$quantile        # Inf — quantile is unbounded
range(res_inf$lower)    # -Inf
range(res_inf$upper)    #  Inf
```

## See also

- `vignette("metahunt-intro", package = "MetaHunt")` — pipeline
  context and the A3 exchangeability assumption.
- `?split_conformal` — single-split conformal calibration.
- `?cross_conformal` — cross-fitting conformal calibration.
- `?conformal_from_fit` — calibration using an existing fit.
- `?coverage` — empirical coverage diagnostics for conformal bands.
- `?plot.metahunt_conformal` — plotting method for the returned
  objects.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight_estimation.R
\name{coef.metahunt_weight_model}
\alias{coef.metahunt_weight_model}
\title{Extract coefficients from a MetaHunt weight model}
\usage{
\method{coef}{metahunt_weight_model}(object, ...)
}
\arguments{
\item{object}{A fitted \code{"metahunt_weight_model"} from \code{\link[=fit_weight_model]{fit_weight_model()}}.}

\item{...}{Passed through to \code{\link[stats:coef]{stats::coef()}}.}
}
\value{
The coefficient vector / matrix returned by
\code{DirichletReg::DirichReg()}'s \code{coef()} method (numeric vector or matrix
depending on the parametrisation used by the underlying fit).
}
\description{
Returns the regression coefficients from the underlying weight-model fit.
For the default \code{"dirichlet"} method this delegates to
\code{\link[DirichletReg:dirichreg]{DirichletReg::DirichReg()}}'s \code{coef()} method.
}
\examples{
\donttest{
set.seed(1)
m <- 60; K <- 3
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- cbind(0.5 * W$w1, -0.3 * W$w2, rep(0, m))
pi_true <- exp(eta) / rowSums(exp(eta))
pi_hat <- pi_true + matrix(rnorm(m * K, sd = 0.01), m, K)
pi_hat <- pmax(pi_hat, 0); pi_hat <- pi_hat / rowSums(pi_hat)
model <- fit_weight_model(pi_hat, W)
coef(model)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_denoising_params.R
\name{select_denoising_params}
\alias{select_denoising_params}
\title{Choose d-fSPA denoising parameters by cross-validation}
\usage{
select_denoising_params(
  F_hat,
  W,
  K,
  N_grid = NULL,
  Delta_grid = NULL,
  n_folds = 5L,
  grid_weights = NULL,
  dfspa_args = list(),
  weight_model_args = list(),
  seed = NULL
)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G_grid} numeric matrix of study-level function
evaluations.}

\item{W}{An \code{m}-by-\code{p} matrix or data frame of study-level covariates.}

\item{K}{Integer number of basis functions (fixed during this search).}

\item{N_grid}{Optional numeric vector of candidate \code{N} values. Defaults
to \code{c(0.2, 0.5, 1.0) * log(nrow(F_hat))}.}

\item{Delta_grid}{Optional numeric vector of candidate \code{Delta} values.
Defaults to \code{c(0.05, 0.10, 0.20, 0.30)} times the maximum pairwise
\eqn{L^2(\mu)} distance among studies.}

\item{n_folds}{Integer number of folds (default \code{5}).}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector.}

\item{dfspa_args}{Named list of additional arguments for \code{\link[=dfspa]{dfspa()}}
(e.g. \code{list()}; do not include \code{N}, \code{Delta}, or \code{denoise} here, they
are set by the search).}

\item{weight_model_args}{Named list of additional arguments for
\code{\link[=fit_weight_model]{fit_weight_model()}}.}

\item{seed}{Optional integer seed for reproducible fold assignment.}
}
\value{
An object of class \code{metahunt_denoising_search}: a list with
\describe{
\item{\code{grid}}{A data frame with one row per \verb{(N, Delta)} pair,
columns \code{N}, \code{Delta}, \code{cv_error}, \code{cv_se}, \code{n_folds_ok}.}
\item{\code{best}}{A list with the \verb{(N, Delta)} minimising \code{cv_error}.}
\item{\code{K}, \code{n_folds}, \code{grid_weights}}{Inputs echoed back for
traceability.}
}
}
\description{
At a fixed \code{K}, performs k-fold cross-validation over a grid of denoising
parameter pairs \verb{(N, Delta)} for \code{\link[=dfspa]{dfspa()}}. For each candidate pair and
each fold, the full MetaHunt pipeline is fit on the out-of-fold studies
and predicts the held-out studies' functions. The pair with the lowest
average prediction error is selected.
}
\details{
This is the cross-validated tuning of the denoising parameters discussed
in Section 3.1 of the paper. Joint tuning over \verb{(K, N, Delta)} is not
supported because it scales poorly; if you also want to choose \code{K}, do
it first via \code{\link[=cv_error_curve]{cv_error_curve()}} and then call this function at the
selected \code{K}.

Default candidate grids:
\itemize{
\item \code{N_grid    = m * c(NA, NA, NA)} resolved at runtime to
\code{c(0.2, 0.5, 1.0) * log(m)}.
\item \code{Delta_grid = max_pairwise_dist * c(0.05, 0.10, 0.20, 0.30)}.
}

Pass your own \code{N_grid} / \code{Delta_grid} (in original units) to override.
}
\examples{
\donttest{
set.seed(1)
G <- 30; m <- 80
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- as.matrix(W) \%*\% cbind(c(1, -0.5), c(-0.4, 1), c(0, 0))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

tune <- select_denoising_params(F_hat, W, K = 3, n_folds = 4, seed = 1)
tune$grid
tune$best
}

}
\seealso{
\code{\link[=cv_error_curve]{cv_error_curve()}} for selecting \code{K}, \code{\link[=dfspa]{dfspa()}} for the
underlying basis-hunting algorithm.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metahunt.R
\name{summary.metahunt}
\alias{summary.metahunt}
\title{Summarise a MetaHunt fit}
\usage{
\method{summary}{metahunt}(object, ...)
}
\arguments{
\item{object}{A \code{"metahunt"} object from \code{\link[=metahunt]{metahunt()}}.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"summary.metahunt"}: a list with components
\describe{
\item{\code{m}}{Number of studies.}
\item{\code{G_grid}}{Grid size.}
\item{\code{K}}{Number of basis functions.}
\item{\code{weight_method}}{Method used by the weight model.}
\item{\code{predictor_names}}{Character vector of covariate names.}
\item{\code{pi_summary}}{A \code{K}-by-5 numeric matrix; each row gives \code{min},
\code{mean}, \code{median}, \code{max}, and \code{sd} of the corresponding column of
\code{object$pi_hat}.}
\item{\code{n_kept}}{Number of studies retained after denoising.}
\item{\code{n_dropped}}{Number of studies dropped (\code{m - n_kept}).}
\item{\code{denoising}}{List with \code{N} and \code{Delta} from the \code{dfspa} fit
(both \code{NA} when \code{denoise = FALSE}).}
}
}
\description{
Produces a compact summary of a \code{"metahunt"} object, including study/grid
sizes, the weight-model method, per-basis summary statistics for the
training simplex weights \code{pi_hat}, and denoising bookkeeping from the
underlying \code{\link[=dfspa]{dfspa()}} fit.
}
\examples{
set.seed(1)
G <- 25; m <- 40
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- as.matrix(W) \%*\% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
summary(fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conformal.R
\name{coverage}
\alias{coverage}
\title{Empirical coverage of a conformal prediction-interval object}
\usage{
coverage(object, F_obs, grid_weights = NULL)
}
\arguments{
\item{object}{A \code{"metahunt_conformal"} object from \code{\link[=split_conformal]{split_conformal()}},
\code{\link[=cross_conformal]{cross_conformal()}}, or \code{\link[=conformal_from_fit]{conformal_from_fit()}}.}

\item{F_obs}{An \code{n_target}-by-\code{G_grid} numeric matrix of observed
study-level functions for the target studies, in the same row order
as \code{object$prediction}.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric
vector. Only used in scalar mode (passed to \code{\link[=apply_wrapper]{apply_wrapper()}}).}
}
\value{
A list. In \strong{pointwise} mode the list contains
\describe{
\item{\code{pointwise}}{\code{n_target}-by-\code{G_grid} logical matrix of
coverage indicators.}
\item{\code{per_target}}{Length-\code{n_target} numeric vector of mean
coverage across the grid for each target.}
\item{\code{per_grid_point}}{Length-\code{G_grid} numeric vector of mean
coverage across targets at each grid point.}
\item{\code{overall}}{Scalar mean coverage across all entries.}
\item{\code{nominal}}{Nominal coverage \code{1 - object$alpha}.}
}
In \strong{scalar} mode the list contains
\describe{
\item{\code{pointwise}}{Length-\code{n_target} logical vector of coverage
indicators.}
\item{\code{overall}}{Scalar mean coverage.}
\item{\code{nominal}}{Nominal coverage \code{1 - object$alpha}.}
}
}
\description{
Computes empirical coverage indicators of a fitted
\code{"metahunt_conformal"} object against held-out observed study-level
functions. The held-out studies must correspond positionally to the
targets used to build \code{object} (i.e. \code{F_obs[i, ]} is the observed
function for the same target whose prediction is in
\code{object$prediction[i, ]} or \code{object$prediction[i]}).
}
\details{
In \strong{pointwise} mode each entry \verb{(i, g)} of \code{F_obs} is compared against
\verb{[object$lower[i, g], object$upper[i, g]]}. In \strong{scalar} mode \code{F_obs}
is first reduced to a length-\code{n_target} vector via
\code{\link[=apply_wrapper]{apply_wrapper()}} using \code{object$wrapper} and the supplied
\code{grid_weights}, and then compared against
\verb{[object$lower, object$upper]}.

Coverage is a finite-sample diagnostic. Nominal coverage is
\code{1 - object$alpha}; empirical coverage will fluctuate around this
value due to sampling.
}
\examples{
set.seed(1)
G <- 25; m <- 80
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m))
eta <- cbind(0.6 * W$w1, -0.3 * W$w1, rep(0, m))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

# held-out test set: same data-generating process, same W
test_idx <- 1:10
train_idx <- setdiff(seq_len(m), test_idx)
res <- split_conformal(F_hat[train_idx, ], W[train_idx, , drop = FALSE],
                       W[test_idx, , drop = FALSE], K = 3,
                       dfspa_args = list(denoise = FALSE), seed = 1)
cov <- coverage(res, F_obs = F_hat[test_idx, , drop = FALSE])
cov$overall

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MetaHunt-package.R
\docType{package}
\name{MetaHunt-package}
\alias{MetaHunt}
\alias{MetaHunt-package}
\title{MetaHunt: Privacy-Preserving Meta-Analysis via Low-Rank Basis Hunting}
\description{
Tools for privacy-preserving meta-analysis of function-valued quantities
(e.g. regression, conditional average treatment effect functions) across
heterogeneous studies. Implements the MetaHunt pipeline of Shi, Imai,
and Zhang: denoised functional Successive Projection Algorithm (d-fSPA)
for basis hunting, constrained projection of study functions onto the
recovered simplex, Dirichlet regression of mixing weights on study-level
covariates, target prediction, and split or cross conformal prediction
intervals.
}
\section{Main entry points}{

\itemize{
\item \code{\link[=metahunt]{metahunt()}} — fit the full pipeline end-to-end.
\item \code{\link[=predict.metahunt]{predict.metahunt()}} — predict target functions for new study-level covariates.
\item \code{\link[=split_conformal]{split_conformal()}}, \code{\link[=cross_conformal]{cross_conformal()}}, \code{\link[=conformal_from_fit]{conformal_from_fit()}} —
prediction intervals.
\item \code{\link[=minmax_regret]{minmax_regret()}} — covariate-free worst-case-regret aggregator
(Zhang, Huang, and Imai 2024) as a baseline.
\item \code{\link[=f_hat_from_models]{f_hat_from_models()}}, \code{\link[=build_grid]{build_grid()}} — onramp from fitted-model
lists and reference data to the package's matrix inputs.
}
}

\section{Pipeline building blocks}{

\code{\link[=dfspa]{dfspa()}}, \code{\link[=project_to_simplex]{project_to_simplex()}}, \code{\link[=fit_weight_model]{fit_weight_model()}},
\code{\link[=predict_target]{predict_target()}}, \code{\link[=apply_wrapper]{apply_wrapper()}},
\code{\link[=reconstruction_error_curve]{reconstruction_error_curve()}}, \code{\link[=cv_error_curve]{cv_error_curve()}},
\code{\link[=select_denoising_params]{select_denoising_params()}}.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/WShi18/MetaHunt}
  \item \url{https://wshi18.github.io/MetaHunt/}
  \item \url{https://arxiv.org/abs/2604.23847}
  \item Report bugs at \url{https://github.com/WShi18/MetaHunt/issues}
}

}
\author{
\strong{Maintainer}: Wenqi Shi \email{wenqishi18@gmail.com}

Authors:
\itemize{
  \item Kosuke Imai \email{imai@harvard.edu}
  \item Yi Zhang \email{yizhang0017@gmail.com}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onramp.R
\name{build_grid}
\alias{build_grid}
\title{Build a shared evaluation grid from a reference dataset}
\usage{
build_grid(reference_data, n_grid = NULL, seed = NULL)
}
\arguments{
\item{reference_data}{A data frame (or matrix) of reference patient-level
covariates. May be a held-out target-population sample, the pooled
source covariates, or any plausible reference distribution.}

\item{n_grid}{Optional integer giving the desired grid size. If \code{NULL}
or \verb{>= nrow(reference_data)}, the full reference data is returned.}

\item{seed}{Optional integer seed for reproducibility (used only when
sub-sampling).}
}
\value{
A data frame of grid points.
}
\description{
Constructs a data frame of grid points suitable for \code{\link[=f_hat_from_models]{f_hat_from_models()}}
from any reference patient-level dataset. This is convenient when the
patient-level covariate space is multidimensional and there is no
obvious one-dimensional grid.
}
\details{
If \code{n_grid} is \code{NULL} or at least \code{nrow(reference_data)}, the reference
data is returned unchanged. Otherwise \code{n_grid} rows are sampled uniformly
at random (without replacement). The reference data should be on the
same scale and have the same columns as the data each centre's model
was fitted on.

The empirical distribution of the returned grid implicitly defines the
\eqn{\mu} measure used downstream. Pass uniform \code{grid_weights} (the
default) to weight each grid point equally; pass non-uniform
\code{grid_weights} to weight by an external reference distribution.
}
\examples{
set.seed(1)
ref <- data.frame(age = rnorm(500, 60, 10),
                  bp  = rnorm(500, 130, 15),
                  sex = sample(c("F", "M"), 500, replace = TRUE))
grid <- build_grid(ref, n_grid = 50, seed = 1)
nrow(grid)
head(grid)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rank_selection.R
\name{reconstruction_error_curve}
\alias{reconstruction_error_curve}
\title{Reconstruction-error curve for basis-rank selection}
\usage{
reconstruction_error_curve(
  F_hat,
  K_range = NULL,
  grid_weights = NULL,
  dfspa_args = list()
)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G_grid} numeric matrix of study-level function
evaluations on the shared grid.}

\item{K_range}{Integer vector of candidate \code{K} values. Defaults to
\code{2:min(nrow(F_hat) - 1, 10)}.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector
used for the \eqn{L^2(\mu)} norm; defaults to uniform.}

\item{dfspa_args}{Named list of extra arguments passed to \code{\link[=dfspa]{dfspa()}},
e.g. \code{list(denoise = FALSE, N = 2)}.}
}
\value{
A data frame with columns \code{K} (integer) and \code{error} (numeric).
Rows where \code{dfspa()} or the projection fails are reported with
\code{error = NA} and a single warning summarising the failures.
}
\description{
For each candidate number of bases \code{K}, run \code{\link[=dfspa]{dfspa()}} followed by
\code{\link[=project_to_simplex]{project_to_simplex()}} and report the average projection residual
\deqn{\mathcal E(K) = \frac{1}{m}\sum_{i=1}^m
       \left\|\hat f^{(i)} - \sum_{k=1}^K \hat\pi_{ik}\hat g_k\right\|_{L^2(\mu)}.}
Plotting \code{error} against \code{K} typically shows an elbow.
}
\details{
This is the unsupervised rank-selection criterion of Section 3.2 of the
paper (Equation for \eqn{\mathcal E(K)}). It does not require
study-level covariates.
}
\examples{
set.seed(1)
G <- 40
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
m <- 50
pi_mat <- matrix(stats::rgamma(m * 3, shape = 0.5), m, 3)
pi_mat <- pi_mat / rowSums(pi_mat)
F_hat  <- pi_mat \%*\% basis + matrix(stats::rnorm(m * G, sd = 0.02), m, G)

elbow <- reconstruction_error_curve(F_hat, K_range = 2:6)
elbow

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conformal.R
\name{cross_conformal}
\alias{cross_conformal}
\title{Cross-conformal prediction intervals (pooled K-fold scores)}
\usage{
cross_conformal(
  F_hat,
  W,
  W_new,
  K,
  alpha = 0.05,
  n_folds = 5L,
  wrapper = NULL,
  grid_weights = NULL,
  dfspa_args = list(),
  weight_model_args = list(),
  seed = NULL
)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G_grid} numeric matrix of study-level function
evaluations.}

\item{W}{An \code{m}-by-\code{p} matrix or data frame of study-level covariates.}

\item{W_new}{A matrix or data frame of new target covariates. Must
contain columns matching \code{W}.}

\item{K}{Integer number of basis functions.}

\item{alpha}{Miscoverage level; interval has nominal coverage
\eqn{1-\alpha}. Default \code{0.05}.}

\item{n_folds}{Integer number of folds (>= 2). Default \code{5}.}

\item{wrapper}{Optional reduction function (see \code{\link[=apply_wrapper]{apply_wrapper()}}). If
\code{NULL}, intervals are constructed pointwise at every grid point.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector
used for the \eqn{L^2(\mu)} norm and for the default wrapper.}

\item{dfspa_args, weight_model_args}{Named lists passed to \code{\link[=dfspa]{dfspa()}} and
\code{\link[=fit_weight_model]{fit_weight_model()}} respectively.}

\item{seed}{Optional integer seed for reproducible train/calibration
splits.}
}
\value{
An object of class \code{"metahunt_conformal"} (see
\code{\link[=split_conformal]{split_conformal()}} for fields). \code{method} is \code{"cross"} and \code{n_cal} is
the number of pooled scores.
}
\description{
Computes K-fold split conformal intervals in which the calibration scores
are pooled across folds, while point predictions for new targets are
produced from a final pipeline fit on all studies. Equivalent to running
\code{\link[=split_conformal]{split_conformal()}} \code{n_folds} times with different calibration sets and
pooling all conformity scores into a single empirical distribution.
}
\details{
For each fold, the MetaHunt pipeline is fit on the out-of-fold studies,
and conformity scores are computed on the in-fold studies. After all
folds complete, the pooled scores yield a single
\eqn{(1-\alpha)}-quantile (or one per grid point). Point predictions for
\code{W_new} use a pipeline refit on the full dataset. The interval at grid
point \code{g} for target \code{j} is
\eqn{[\tilde f^{(j)}(x_g) - q_g, \tilde f^{(j)}(x_g) + q_g]} (or the
scalar analogue when \code{wrapper} is supplied).

This differs from Vovk's original cross-conformal predictor for
classification. For regression, pooling scores across folds is a
common practical extension of split conformal and reduces the variance
due to the single calibration split. Exact finite-sample coverage is
not guaranteed; see Barber et al. (2021, Jackknife+) for more
conservative alternatives.
}
\examples{
\donttest{
set.seed(1)
G <- 30; m <- 60; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m))
eta <- cbind(0.8 * W$w1, -0.3 * W$w1, rep(0, m))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
W_new <- data.frame(w1 = c(0, 1))

res <- cross_conformal(F_hat, W, W_new, K = 3, n_folds = 4,
                       dfspa_args = list(denoise = FALSE), seed = 1)
res
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dfspa.R
\name{dfspa}
\alias{dfspa}
\title{Denoised functional Successive Projection Algorithm (d-fSPA)}
\usage{
dfspa(F_hat, K, grid_weights = NULL, N = NULL, Delta = NULL, denoise = TRUE)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G} numeric matrix where row \code{i} is the evaluation
of the estimated function \eqn{\hat f^{(i)}} at \code{G} grid points.}

\item{K}{Integer number of basis functions to recover. Must satisfy
\verb{1 <= K <= m} after denoising.}

\item{grid_weights}{Optional length-\code{G} non-negative numeric vector of
grid weights defining the \eqn{L^2(\mu)} inner product. Defaults to
uniform weights \code{1 / G}.}

\item{N, Delta}{Optional numeric tuning parameters controlling denoising.
See Details.}

\item{denoise}{Logical; if \code{FALSE}, the denoising step is skipped and
plain fSPA is run. Defaults to \code{TRUE}.}
}
\value{
An object of class \code{"dfspa"}: a list containing
\describe{
\item{\code{bases}}{A \code{K}-by-\code{G} matrix whose rows are the recovered basis
functions evaluated on the grid (denoised, if applicable).}
\item{\code{selected}}{Length-\code{K} integer vector of the selected row indices
into the post-denoising function matrix.}
\item{\code{original_indices}}{Length-\code{K} integer vector of the selected
study indices in the original input \code{F_hat} (before any rows were
dropped by denoising).}
\item{\code{kept}}{Integer vector of row indices of \code{F_hat} that survived
denoising.}
\item{\code{F_denoised}}{The post-denoising function matrix
(\code{length(kept)}-by-\code{G}).}
\item{\code{grid_weights}}{Grid weights used.}
\item{\code{N}, \code{Delta}}{Tuning parameters actually used (or \code{NA} when
\code{denoise = FALSE}).}
\item{\code{K}}{Number of bases requested.}
\item{\code{call}}{The matched call.}
}
}
\description{
Recovers a set of \code{K} latent basis functions from a collection of
study-level function estimates under the low-rank cross-study heterogeneity
assumption of Shi, Imai, and Zhang. Implements Algorithm 1 of the paper
("The d-fSPA Algorithm for basis hunting").
}
\details{
Each study-level function is represented by its evaluations on a shared
grid of \code{G} points. The (weighted) \eqn{L^2(\mu)} inner product is
\eqn{\langle f,g\rangle = \sum_{j=1}^G w_j f(x_j) g(x_j)}, where the
\code{grid_weights} \code{w_j} are proportional to the measure \eqn{\mu}. If not
supplied, uniform weights \code{1 / G} are used.

Denoising follows Jin (2024): for each study \code{i}, let
\eqn{B_\Delta(\hat f^{(i)}) = \{j : \|\hat f^{(j)} - \hat f^{(i)}\| \le
\Delta\}}. If \eqn{|B_\Delta(\hat f^{(i)})| < N}, study \code{i} is discarded;
otherwise \eqn{\hat f^{(i)}} is replaced by the average of the functions
in \eqn{B_\Delta}.
After denoising, the functional SPA step iteratively selects, at each of
the \code{K} iterations, the remaining function with the largest norm after
projecting out the span of previously selected bases.

Default tuning parameters follow the heuristics of the paper:
\code{N = 0.5 * log(m)} and
\eqn{\Delta = \max_{ij} \|\hat f^{(i)} - \hat f^{(j)}\| / 10}.
}
\examples{
set.seed(1)
G <- 50
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)          # 3 true bases
pi_mat <- rbind(diag(3),                              # 3 pure studies
                c(0.5, 0.3, 0.2),
                c(0.2, 0.5, 0.3),
                c(0.3, 0.3, 0.4))
F_hat <- pi_mat \%*\% basis                             # m = 6, G = 50
fit <- dfspa(F_hat, K = 3, denoise = FALSE)
fit$original_indices    # should be a permutation of 1, 2, 3

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/onramp.R
\name{f_hat_from_models}
\alias{f_hat_from_models}
\title{Build the \code{F_hat} matrix from a list of fitted study-level models}
\usage{
f_hat_from_models(models, grid, predict_fn = NULL)
}
\arguments{
\item{models}{A non-empty list of fitted model objects, one per study.}

\item{grid}{A data frame (or matrix) of grid points at which to evaluate
each model. Same columns as the data each model was fitted on.}

\item{predict_fn}{Optional \verb{function(model, grid)} returning a numeric
vector; defaults to the class-aware dispatcher above.}
}
\value{
An \code{length(models)}-by-\code{nrow(grid)} numeric matrix; row \code{i} is
model \code{i} evaluated at every row of \code{grid}.
}
\description{
Many users arrive at MetaHunt with one fitted model per study (e.g. a
\code{ranger::ranger()} random forest or a \code{grf::causal_forest()}) and a
chosen evaluation grid. \code{f_hat_from_models()} evaluates each model on
the shared grid and stacks the predictions into the \code{m}-by-\code{G_grid}
matrix the rest of the package expects.
}
\details{
By default the function dispatches on the model class:
\itemize{
\item \code{ranger} objects are evaluated as \code{predict(model, data = grid)$predictions}.
\item Objects inheriting from \code{causal_forest} or \code{grf} are evaluated as
\code{predict(model, newdata = grid)$predictions}.
\item All other classes fall through to
\code{as.numeric(predict(model, newdata = grid))}, which works for \code{lm},
\code{glm}, \code{randomForest}, and most other R model objects.
}

Override the dispatch with \code{predict_fn = function(model, grid) ...} if
your models need bespoke handling. The function must return a length-\code{G_grid}
numeric vector for each model.

All rows of the returned matrix must have the same length (\code{G_grid})
and contain no \code{NA} values; the function errors otherwise.
}
\examples{
# Toy example: each "centre" fits a polynomial regression
set.seed(1)
make_centre_data <- function(slope) {
  x <- runif(60)
  data.frame(x = x, y = slope * x + rnorm(60, sd = 0.1))
}
models <- lapply(c(-1, 0, 1, 0.5, -0.5), function(s)
  stats::lm(y ~ poly(x, 2), data = make_centre_data(s)))

grid  <- data.frame(x = seq(0, 1, length.out = 30))
F_hat <- f_hat_from_models(models, grid)
dim(F_hat)   # 5 x 30

}
\seealso{
\code{\link[=build_grid]{build_grid()}} to construct \code{grid} from a reference dataset.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conformal.R
\name{plot.metahunt_conformal}
\alias{plot.metahunt_conformal}
\title{Plot a conformal prediction-interval object}
\usage{
\method{plot}{metahunt_conformal}(
  x,
  target_idx = 1L,
  x_axis = NULL,
  fill = grDevices::adjustcolor("steelblue", 0.2),
  line_col = "steelblue",
  ...
)
}
\arguments{
\item{x}{A \code{"metahunt_conformal"} object.}

\item{target_idx}{For pointwise objects, the integer index of the
target study to plot (default \code{1}). Ignored for scalar objects.}

\item{x_axis}{Optional numeric vector of length \code{G_grid} giving the
x-axis values for pointwise plotting. Defaults to \code{seq_len(G_grid)}.}

\item{fill}{Polygon fill colour for the band. Default semi-transparent
steel blue.}

\item{line_col}{Line colour for the predicted function (or points in
scalar mode). Default steel blue.}

\item{...}{Additional graphical parameters passed to the underlying
plotting calls.}
}
\value{
Invisibly returns \code{x}.
}
\description{
For pointwise objects (no \code{wrapper} was used), draws the predicted
function for one target study together with its pointwise band.
For scalar objects, draws point predictions with whisker error bars
over all targets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conformal.R
\name{conformal_from_fit}
\alias{conformal_from_fit}
\title{Split conformal intervals from a pre-fit MetaHunt pipeline}
\usage{
conformal_from_fit(
  dfspa_fit,
  weight_model,
  F_cal,
  W_cal,
  W_new,
  alpha = 0.05,
  wrapper = NULL,
  grid_weights = NULL
)
}
\arguments{
\item{dfspa_fit}{A \code{"dfspa"} object from \code{\link[=dfspa]{dfspa()}}.}

\item{weight_model}{A \code{"metahunt_weight_model"} object from
\code{\link[=fit_weight_model]{fit_weight_model()}}. Must satisfy
\code{weight_model$K == nrow(dfspa_fit$bases)}.}

\item{F_cal}{An \code{n_cal}-by-\code{G_grid} numeric matrix of observed study-level
function evaluations for the calibration set. Calibration studies must
be independent of both \code{dfspa_fit} and \code{weight_model} for valid
coverage.}

\item{W_cal}{A matrix or data frame of study-level covariates for the
calibration set, with the same columns used to fit \code{weight_model}.}

\item{W_new}{A matrix or data frame of study-level covariates for new
target studies.}

\item{alpha}{Miscoverage level; default \code{0.05}.}

\item{wrapper}{Optional reduction function (see \code{\link[=apply_wrapper]{apply_wrapper()}}). If
\code{NULL}, intervals are pointwise at every grid point.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector
used for the default wrapper and for weighted inner products.}
}
\value{
An object of class \code{"metahunt_conformal"}; see \code{\link[=split_conformal]{split_conformal()}}
for a description of its fields.
}
\description{
Lower-level entry point that builds split conformal intervals for new
target covariates using an already-fitted d-fSPA basis decomposition,
an already-fitted weight model, and a user-supplied calibration set.
Use this when you have independently tuned \code{K} or want to reuse a
pipeline fit; otherwise the high-level \code{\link[=split_conformal]{split_conformal()}} is usually
more convenient.
}
\examples{
set.seed(1)
G <- 30; m <- 80
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m))
eta <- cbind(0.8 * W$w1, -0.4 * W$w1, rep(0, m))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

# user-controlled split and fit
tr  <- 1:50; cal <- 51:70; new <- 71:80
fit <- dfspa(F_hat[tr, ], K = 3)
pih <- project_to_simplex(F_hat[tr, ], fit$bases)
wm  <- fit_weight_model(pih, W[tr, , drop = FALSE])

res <- conformal_from_fit(
  fit, wm,
  F_cal = F_hat[cal, ], W_cal = W[cal, , drop = FALSE],
  W_new = W[new, , drop = FALSE],
  wrapper = mean
)
res

}
\seealso{
\code{\link[=split_conformal]{split_conformal()}} for the high-level version that splits and
fits internally, \code{\link[=cross_conformal]{cross_conformal()}} for the K-fold variant.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metahunt.R
\name{metahunt}
\alias{metahunt}
\title{Fit the full MetaHunt pipeline}
\usage{
metahunt(
  F_hat,
  W,
  K,
  grid_weights = NULL,
  dfspa_args = list(),
  weight_model_args = list()
)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G_grid} numeric matrix of study-level function
evaluations on a shared grid; row \code{i} is \eqn{\hat f^{(i)}}.}

\item{W}{An \code{m}-by-\code{p} matrix or data frame of study-level covariates.}

\item{K}{Integer number of basis functions.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector
defining the \eqn{L^2(\mu)} inner product; defaults to uniform.}

\item{dfspa_args}{Named list of extra arguments for \code{\link[=dfspa]{dfspa()}}.}

\item{weight_model_args}{Named list of extra arguments for
\code{\link[=fit_weight_model]{fit_weight_model()}}.}
}
\value{
An object of class \code{"metahunt"}: a list with the \code{dfspa_fit},
\code{weight_model}, training \code{pi_hat}, \code{K}, and a stored copy of
\code{grid_weights}.
}
\description{
End-to-end convenience wrapper that runs the three training-time steps
of the MetaHunt pipeline in sequence:
(1) \code{\link[=dfspa]{dfspa()}} for basis hunting,
(2) \code{\link[=project_to_simplex]{project_to_simplex()}} for per-study weight recovery, and
(3) \code{\link[=fit_weight_model]{fit_weight_model()}} for modelling the weight-to-covariate map.
The result supports \code{\link[=predict.metahunt]{predict.metahunt()}} for generating target-function
predictions on new study-level covariates.
}
\details{
For uncertainty quantification, pair a \code{"metahunt"} fit with
\code{\link[=conformal_from_fit]{conformal_from_fit()}} (requires a separate calibration set). The
high-level \code{\link[=split_conformal]{split_conformal()}} and \code{\link[=cross_conformal]{cross_conformal()}} functions
perform their own fitting and do not consume a pre-fit \code{"metahunt"}
object.
}
\examples{
set.seed(1)
G <- 40; m <- 80
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- as.matrix(W) \%*\% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

fit <- metahunt(F_hat, W, K = 3)
fit
f_pred <- predict(fit, newdata = W[1:3, ])
dim(f_pred)                                  # 3 x G: predicted functions
predict(fit, newdata = W[1:3, ], wrapper = mean)  # scalar summaries

}
\seealso{
\code{\link[=predict.metahunt]{predict.metahunt()}}, \code{\link[=split_conformal]{split_conformal()}},
\code{\link[=cross_conformal]{cross_conformal()}}, \code{\link[=conformal_from_fit]{conformal_from_fit()}},
\code{\link[=reconstruction_error_curve]{reconstruction_error_curve()}}, \code{\link[=cv_error_curve]{cv_error_curve()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metahunt.R
\name{print.summary.metahunt}
\alias{print.summary.metahunt}
\title{Print a \code{summary.metahunt} object}
\usage{
\method{print}{summary.metahunt}(x, ...)
}
\arguments{
\item{x}{A \code{"summary.metahunt"} object from \code{\link[=summary.metahunt]{summary.metahunt()}}.}

\item{...}{Ignored.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print a \code{summary.metahunt} object
}
\examples{
set.seed(1)
G <- 25; m <- 40
x_grid <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x_grid), cos(pi * x_grid), x_grid)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- as.matrix(W) \%*\% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
print(summary(fit))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight_estimation.R
\name{fit_weight_model}
\alias{fit_weight_model}
\title{Fit a weight model mapping study-level covariates to simplex weights}
\usage{
fit_weight_model(
  pi_hat,
  W,
  method = c("dirichlet"),
  boundary_eps = 1e-04,
  formula = NULL,
  ...
)
}
\arguments{
\item{pi_hat}{An \code{m}-by-\code{K} numeric matrix of simplex weights; rows must
be non-negative and sum to 1 (up to tolerance \code{1e-6}).}

\item{W}{An \code{m}-by-\code{p} matrix or data frame of study-level covariates.}

\item{method}{Weight-model method. Currently only \code{"dirichlet"} is
supported.}

\item{boundary_eps}{Small positive scalar used to shrink weights away
from the simplex boundary before Dirichlet fitting. Defaults to \code{1e-4}.}

\item{formula}{Optional RHS-only formula (e.g. \code{~ x1 + I(x2^2)}) describing
the covariate part of the Dirichlet regression. Defaults to \code{~ .}
(all columns of \code{W}).}

\item{...}{Passed through to \code{\link[DirichletReg:dirichreg]{DirichletReg::DirichReg()}}.}
}
\value{
An object of class \code{"metahunt_weight_model"}: a list with the
fitted model, formula, method, \code{K}, and training covariate names.
}
\description{
Given a matrix of simplex-valued weights \eqn{\hat\pi_1,\ldots,\hat\pi_m}
(e.g. from \code{\link[=project_to_simplex]{project_to_simplex()}}) and associated study-level covariates
\eqn{\mathbf W_1,\ldots,\mathbf W_m}, fit a model
\eqn{\widehat{\mathcal M}:\mathbf W \mapsto \boldsymbol\pi}.
The default method is Dirichlet regression via the \code{DirichletReg} package.
}
\details{
Dirichlet regression cannot handle weights exactly at the simplex boundary
(\code{0} or \code{1}), which frequently arise after constrained projection. Before
fitting, rows of \code{pi_hat} are shrunk toward the barycenter via
\eqn{\tilde\pi = (\pi + \varepsilon) / (1 + K\varepsilon)}, with
\eqn{\varepsilon} set by \code{boundary_eps}.
}
\examples{
\donttest{
set.seed(1)
m <- 80; K <- 3; p <- 2
W <- matrix(rnorm(m * p), m, p); colnames(W) <- c("w1", "w2")
# generate simplex weights driven by W
eta <- cbind(0.5 * W[, 1], -0.3 * W[, 2], rep(0, m))
pi_true <- exp(eta) / rowSums(exp(eta))
pi_hat <- pi_true + matrix(rnorm(m * K, sd = 0.01), m, K)
pi_hat <- pmax(pi_hat, 0); pi_hat <- pi_hat / rowSums(pi_hat)
model <- fit_weight_model(pi_hat, W)
predict(model, newdata = matrix(c(0, 0), 1, 2, dimnames = list(NULL, c("w1","w2"))))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_denoising_params.R
\name{print.metahunt_denoising_search}
\alias{print.metahunt_denoising_search}
\title{Print method for d-fSPA denoising parameter search results}
\usage{
\method{print}{metahunt_denoising_search}(x, ...)
}
\arguments{
\item{x}{An object of class \code{metahunt_denoising_search} returned by
\code{\link[=select_denoising_params]{select_denoising_params()}}.}

\item{...}{Unused; present for S3 generic compatibility.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print method for d-fSPA denoising parameter search results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight_estimation.R
\name{project_to_simplex}
\alias{project_to_simplex}
\title{Project study-level functions onto the simplex spanned by basis functions}
\usage{
project_to_simplex(F_hat, bases, grid_weights = NULL, ridge = 1e-10)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G_grid} numeric matrix; row \code{i} is the study
function \eqn{\hat f^{(i)}} evaluated on the shared grid.}

\item{bases}{A \code{K}-by-\code{G_grid} numeric matrix of basis functions on the
same grid, typically the \code{bases} slot of a \code{\link[=dfspa]{dfspa()}} result.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector of
grid weights defining the \eqn{L^2(\mu)} inner product. Defaults to
uniform weights \code{1 / G_grid}.}

\item{ridge}{Small non-negative scalar added to the diagonal of the QP
Hessian for numerical stability. Defaults to \code{1e-10}.}
}
\value{
An \code{m}-by-\code{K} numeric matrix of simplex weights; rows sum to 1
and entries are non-negative.
}
\description{
For each study \code{i}, solves the constrained projection
\deqn{\hat\pi_i = \arg\min_{\pi \in \Delta_{K-1}}
  \left\| \hat f^{(i)} - \sum_{k=1}^K \pi_k \hat g_k \right\|_{L^2(\mu)}}
where the norm is the weighted \eqn{L^2} norm defined by \code{grid_weights}.
This is Equation (3) of the paper and yields the study-specific weights
\verb{hat pi_i} used downstream for weight-model fitting and prediction.
}
\details{
The projection reduces to the quadratic program
\deqn{\min_{\pi \in \mathbb R^K}\ \pi^\top D\,\pi - 2\,d^\top \pi
  \quad \text{s.t.} \quad \mathbf 1^\top \pi = 1,\ \pi \ge 0}
with \eqn{D = G W G^\top} and \eqn{d = G W f^{(i)}}, where
\eqn{G} is the \code{K}-by-\code{G_grid} basis matrix, \eqn{W = \mathrm{diag}(}\code{grid_weights}\eqn{)},
and \eqn{f^{(i)}} is the \code{i}-th row of \code{F_hat}. Solved via
\code{\link[quadprog:solve.QP]{quadprog::solve.QP()}}. A tiny ridge is added to \code{D} for numerical stability.
}
\examples{
set.seed(1)
G <- 40
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
true_pi <- rbind(diag(3), c(0.4, 0.3, 0.3), c(0.1, 0.7, 0.2))
F_hat <- true_pi \%*\% basis
fit <- dfspa(F_hat, K = 3, denoise = FALSE)
pi_hat <- project_to_simplex(F_hat, fit$bases)
round(pi_hat, 3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metahunt.R
\name{predict.metahunt}
\alias{predict.metahunt}
\title{Predict target functions (or scalar summaries) from a MetaHunt fit}
\usage{
\method{predict}{metahunt}(object, newdata, wrapper = NULL, grid_weights = NULL, ...)
}
\arguments{
\item{object}{A \code{"metahunt"} object from \code{\link[=metahunt]{metahunt()}}.}

\item{newdata}{A matrix or data frame of new study-level covariates.}

\item{wrapper}{Optional reduction function. If \code{NULL}, returns the full
predicted function matrix (\code{nrow(newdata)}-by-\code{G_grid}). If a function,
applied to each predicted function to return a scalar per new target;
see \code{\link[=apply_wrapper]{apply_wrapper()}}.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector
used only when \code{wrapper = NULL} with the default weighted-mean path.
Defaults to the \code{grid_weights} stored in \code{object}.}

\item{...}{Ignored.}
}
\value{
Either an \code{nrow(newdata)}-by-\code{G_grid} matrix of predicted
functions (when \code{wrapper = NULL}) or a length-\code{nrow(newdata)} numeric
vector of scalar summaries.
}
\description{
Predict target functions (or scalar summaries) from a MetaHunt fit
}
\examples{
set.seed(1)
G <- 25; m <- 40
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- as.matrix(W) \%*\% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

fit <- metahunt(F_hat, W, K = 3, dfspa_args = list(denoise = FALSE))
f_pred <- predict(fit, newdata = W[1:3, ])
dim(f_pred)                                       # 3 x G
predict(fit, newdata = W[1:3, ], wrapper = mean)  # scalar summaries

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minmax_regret.R
\name{minmax_regret}
\alias{minmax_regret}
\title{Minimax-regret aggregator for multisite function-valued estimands}
\usage{
minmax_regret(F_hat, grid_weights = NULL, ridge = 1e-10, wrapper = NULL)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G_grid} numeric matrix of source-site function
evaluations on a shared grid; row \code{i} is \eqn{\hat f^{(i)}}.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector
defining the target measure used to compute \eqn{\hat\Gamma} and
\eqn{\hat d}. Defaults to uniform weights \code{1 / G_grid}.}

\item{ridge}{Non-negative scalar; replaces \eqn{\hat\Gamma} with
\eqn{\hat\Gamma + \text{ridge}\cdot I} for numerical stability.
Defaults to \code{1e-10}.}

\item{wrapper}{Optional reduction function (see \code{\link[=apply_wrapper]{apply_wrapper()}}). If
\code{NULL}, \code{prediction} is the length-\code{G_grid} predicted target function;
if a function, \code{prediction} is the scalar \code{wrapper(prediction)}
(e.g. an ATE when \code{wrapper = mean} and rows of \code{F_hat} are CATE
evaluations).}
}
\value{
An object of class \code{"minmax_regret"}: a list with
\describe{
\item{\code{prediction}}{Predicted target. Length-\code{G_grid} vector when
\code{wrapper = NULL}; scalar otherwise.}
\item{\code{q}}{Length-\code{m} simplex weights from the minimax-regret QP.}
\item{\code{Gamma}}{The \code{m}-by-\code{m} Gram matrix used (post-ridge).}
\item{\code{d}}{Length-\code{m} linear coefficient vector.}
\item{\code{grid_weights}}{Grid weights used.}
\item{\code{ridge}}{Ridge value used.}
\item{\code{wrapper}}{Wrapper function or \code{NULL}.}
}
}
\description{
Implements the minimax-regret estimator of Zhang, Huang, and Imai
(\href{https://arxiv.org/abs/2412.11136}{arXiv:2412.11136}) for
aggregating site-level function estimates. Given study-level functions
\eqn{\hat f^{(i)}}, the estimator is
\deqn{\hat q = \arg\min_{q \in \Delta_{m-1}}\
  q^\top \hat\Gamma\, q - \hat d^\top q,
  \qquad
  \hat\Gamma_{ij} = \sum_g w_g\, \hat f^{(i)}(x_g)\hat f^{(j)}(x_g),\ \
  \hat d_i = \sum_g w_g\, (\hat f^{(i)}(x_g))^2,}
yielding the predicted target function
\eqn{\tilde f(x) = \sum_{i=1}^m \hat q_i\, \hat f^{(i)}(x)}.
Unlike \code{\link[=metahunt]{metahunt()}}, this method does not use study-level covariates;
the target is the worst-case-regret aggregator over the convex hull of
source functions.
}
\details{
The simplex-constrained QP is solved with \code{\link[quadprog:solve.QP]{quadprog::solve.QP()}}.
A small ridge is added to \eqn{\hat\Gamma} for numerical stability when
source functions are highly collinear. The resulting \code{q} is clipped to
be non-negative and renormalised to sum to 1 to absorb floating-point
drift.
}
\examples{
set.seed(1)
G <- 30; m <- 6
x <- seq(0, 1, length.out = G)
F_hat <- rbind(
  sin(pi * x),
  cos(pi * x),
  x,
  0.5 * sin(pi * x) + 0.5 * x,
  0.3 * cos(pi * x) + 0.7 * x,
  0.4 * sin(pi * x) + 0.4 * cos(pi * x) + 0.2 * x
)

fit <- minmax_regret(F_hat)
fit$q                                        # simplex weights over sources
length(fit$prediction)                       # G: predicted target function

# ATE-style scalar via wrapper
minmax_regret(F_hat, wrapper = mean)$prediction

}
\references{
Zhang, Y., Huang, M., and Imai, K. (2024). Minimax regret estimation
for generalizing heterogeneous treatment effects with multisite data.
\href{https://arxiv.org/abs/2412.11136}{arXiv:2412.11136}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_target.R
\name{apply_wrapper}
\alias{apply_wrapper}
\title{Reduce predicted functions to scalars via a user-supplied wrapper}
\usage{
apply_wrapper(F_mat, wrapper = NULL, grid_weights = NULL)
}
\arguments{
\item{F_mat}{An \code{n}-by-\code{G_grid} numeric matrix; each row a function
on a shared grid (e.g. the output of \code{\link[=predict_target]{predict_target()}}).}

\item{wrapper}{Either \code{NULL} (default, weighted mean) or a function that
takes a single numeric vector of length \code{G_grid} and returns a scalar.
Examples: \code{mean}, \code{median}, \code{max}, \code{function(f) f[17]} (point
evaluation), \code{function(f) sum(f^2)}.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector.
Used only when \code{wrapper = NULL}. Defaults to uniform \code{1 / G_grid}.}
}
\value{
A length-\code{n} numeric vector of scalar summaries.
}
\description{
Many downstream quantities of interest (average treatment effect,
pointwise predictions, or other functionals of \eqn{f^{(0)}}) are scalar
summaries of the predicted function. \code{apply_wrapper()} applies any
user-supplied reduction to each row of a function matrix, with a default
of the weighted mean with respect to \code{grid_weights} (which coincides with
\eqn{\int f\,d\mu} when \code{grid_weights} represents \eqn{\mu}).
}
\examples{
F_mat <- matrix(1:12, nrow = 3, byrow = TRUE)  # 3 "functions" on a 4-point grid
apply_wrapper(F_mat)                            # row means (uniform grid weights)
apply_wrapper(F_mat, wrapper = max)             # row maxes
apply_wrapper(F_mat, wrapper = function(f) f[2]) # point evaluation at grid idx 2

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rank_selection.R
\name{cv_error_curve}
\alias{cv_error_curve}
\title{Cross-validated prediction-error curve for basis-rank selection}
\usage{
cv_error_curve(
  F_hat,
  W,
  K_range = NULL,
  n_folds = 5L,
  grid_weights = NULL,
  dfspa_args = list(),
  weight_model_args = list(),
  seed = NULL
)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G_grid} numeric matrix of study-level function
evaluations.}

\item{W}{An \code{m}-by-\code{p} matrix or data frame of study-level covariates.}

\item{K_range}{Integer vector of candidate \code{K} values. Defaults to
\code{2:min(nrow(F_hat) - 1, 10)}.}

\item{n_folds}{Integer number of CV folds (default \code{5}).}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector.}

\item{dfspa_args}{Named list of extra arguments for \code{\link[=dfspa]{dfspa()}}.}

\item{weight_model_args}{Named list of extra arguments for
\code{\link[=fit_weight_model]{fit_weight_model()}}.}

\item{seed}{Optional integer seed for reproducible fold assignment; if
\code{NULL} no seeding is performed.}
}
\value{
A data frame with columns \code{K}, \code{cv_error} (mean over folds),
and \code{cv_se} (standard error across folds). The per-fold error matrix
is attached as the attribute \code{"fold_errors"}
(\code{length(K_range)}-by-\code{n_folds}). Folds where the pipeline fails
contribute \code{NA} and are summarised in a single warning.
}
\description{
For each candidate \code{K}, perform k-fold cross-validation at the study
level. Within each fold, the full MetaHunt pipeline (d-fSPA + constrained
projection + weight model) is refit on the training studies, and the
held-out studies' functions are predicted from their study-level
covariates. The prediction error is
\eqn{\|\hat f^{(i)} - \tilde f^{(i)}\|_{L^2(\mu)}} averaged over
held-out studies and then over folds.
}
\details{
This is the supervised rank-selection criterion of Section 3.2 of the
paper. Each held-out study is excluded from both basis hunting and
weight-model fitting.
}
\examples{
\donttest{
set.seed(1)
G <- 40; m <- 80; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- as.matrix(W) \%*\% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.02), m, G)
cv <- cv_error_curve(F_hat, W, K_range = 2:5, n_folds = 4, seed = 1)
cv
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_target.R
\name{predict_target}
\alias{predict_target}
\title{Predict the target function for new study-level covariates}
\usage{
predict_target(dfspa_fit, weight_model, W_new)
}
\arguments{
\item{dfspa_fit}{A \code{"dfspa"} object produced by \code{\link[=dfspa]{dfspa()}}. Its \code{bases}
slot (\code{K}-by-\code{G_grid} matrix) provides the basis functions on the grid.}

\item{weight_model}{A \code{"metahunt_weight_model"} object produced by
\code{\link[=fit_weight_model]{fit_weight_model()}}. Must have \code{weight_model$K == nrow(dfspa_fit$bases)}.}

\item{W_new}{A matrix or data frame of study-level covariates for the new
target studies, with columns matching those used to fit \code{weight_model}.}
}
\value{
An \code{nrow(W_new)}-by-\code{G_grid} numeric matrix; row \code{j} is the
predicted target function on the grid for the \code{j}-th new study.
}
\description{
Given a fitted d-fSPA basis decomposition and a fitted weight model,
compute the predicted target function on the shared grid as
\deqn{\tilde f^{(0)}(\cdot) = \sum_{k=1}^{\hat K} \tilde\pi_{0k}\, \hat g_k(\cdot),
  \qquad \tilde\pi_0 = \widehat{\mathcal M}(\mathbf W_0).}
}
\examples{
set.seed(1)
G <- 40; m <- 60; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)

# generate study-level covariates and softmax weights
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
beta <- cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
eta  <- as.matrix(W) \%*\% beta
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.02), m, G)

fit    <- dfspa(F_hat, K = K_true)
pi_hat <- project_to_simplex(F_hat, fit$bases)
wm     <- fit_weight_model(pi_hat, W)

W_new   <- data.frame(w1 = c(0, 1), w2 = c(0, -1))
f_pred  <- predict_target(fit, wm, W_new)
dim(f_pred)   # 2 x G

}
\seealso{
\code{\link[=apply_wrapper]{apply_wrapper()}} to reduce predicted functions to scalars.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conformal.R
\name{split_conformal}
\alias{split_conformal}
\title{Split conformal prediction intervals for target-function predictions}
\usage{
split_conformal(
  F_hat,
  W,
  W_new,
  K,
  alpha = 0.05,
  cal_frac = 0.3,
  wrapper = NULL,
  grid_weights = NULL,
  calibration_idx = NULL,
  dfspa_args = list(),
  weight_model_args = list(),
  seed = NULL
)
}
\arguments{
\item{F_hat}{An \code{m}-by-\code{G_grid} numeric matrix of study-level function
evaluations.}

\item{W}{An \code{m}-by-\code{p} matrix or data frame of study-level covariates.}

\item{W_new}{A matrix or data frame of new target covariates. Must
contain columns matching \code{W}.}

\item{K}{Integer number of basis functions.}

\item{alpha}{Miscoverage level; interval has nominal coverage
\eqn{1-\alpha}. Default \code{0.05}.}

\item{cal_frac}{Numeric in \verb{(0, 1)} giving the fraction of studies in
the calibration set. Default \code{0.3}. Ignored if \code{calibration_idx} is
supplied.}

\item{wrapper}{Optional reduction function (see \code{\link[=apply_wrapper]{apply_wrapper()}}). If
\code{NULL}, intervals are constructed pointwise at every grid point.}

\item{grid_weights}{Optional length-\code{G_grid} non-negative numeric vector
used for the \eqn{L^2(\mu)} norm and for the default wrapper.}

\item{calibration_idx}{Optional integer vector of row indices in \code{F_hat}
to use as the calibration set. If supplied, \code{cal_frac} is ignored.}

\item{dfspa_args, weight_model_args}{Named lists passed to \code{\link[=dfspa]{dfspa()}} and
\code{\link[=fit_weight_model]{fit_weight_model()}} respectively.}

\item{seed}{Optional integer seed for reproducible train/calibration
splits.}
}
\value{
An object of class \code{"metahunt_conformal"}: a list with
\describe{
\item{\code{prediction}}{Point predictions for \code{W_new}. A numeric vector
of length \code{nrow(W_new)} in the scalar case, or an
\code{nrow(W_new)}-by-\code{G_grid} matrix in the pointwise case.}
\item{\code{lower}, \code{upper}}{Interval endpoints, same shape as
\code{prediction}.}
\item{\code{alpha}}{Miscoverage level used.}
\item{\code{method}}{\code{"split"}.}
\item{\code{n_cal}}{Calibration sample size.}
\item{\code{quantile}}{The conformal quantile: a scalar (scalar case)
or a length-\code{G_grid} vector (pointwise case).}
\item{\code{wrapper}}{The wrapper used, or \code{NULL}.}
}
}
\description{
Implements Algorithm 2 of the paper (split conformal prediction) over
the MetaHunt pipeline. Studies are partitioned into training and
calibration sets. The training set is used to fit d-fSPA, the constrained
projection, and the weight model; the calibration set supplies
conformity scores, which determine the width of the intervals.
}
\details{
Given a target function, one can either construct intervals \strong{pointwise
at every grid point} (when \code{wrapper = NULL}) or for a \strong{scalar summary}
of the target function (when \code{wrapper} is a function).
\itemize{
\item \strong{Pointwise} (\code{wrapper = NULL}): for each grid point \code{g} the
conformity score is \eqn{R_{i,g} = |\hat f^{(i)}(x_g) - \tilde
  f^{(i)}(x_g)|} across calibration studies \code{i}. A separate
\eqn{(1-\alpha)}-quantile \eqn{q_g} is computed per grid point, and
the interval at grid point \code{g} for target \code{j} is
\eqn{[\tilde f^{(j)}(x_g)-q_g,\ \tilde f^{(j)}(x_g)+q_g]}.
\item \strong{Scalar} (\code{wrapper} supplied): conformity scores are
\eqn{R_i = |s(\hat f^{(i)}) - s(\tilde f^{(i)})|} with
\code{s = wrapper}, and the interval for each target is
\eqn{[s(\tilde f^{(j)})-q,\ s(\tilde f^{(j)})+q]} with a single
shared quantile \eqn{q}.
}

The finite-sample quantile is
\eqn{q = R_{(k)}} with \eqn{k = \lceil(1-\alpha)(n_\mathrm{cal}+1)\rceil};
if \eqn{k > n_\mathrm{cal}}, \code{q = Inf} and intervals are
\eqn{(-\infty, \infty)}.
}
\examples{
set.seed(1)
G <- 40; m <- 80; K_true <- 3
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- as.matrix(W) \%*\% cbind(c(1, -0.8), c(-0.5, 1.2), c(0, 0))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)

W_new <- data.frame(w1 = c(0, 1), w2 = c(0, -1))
# pointwise intervals at every grid point
pi_grid <- split_conformal(F_hat, W, W_new, K = 3, seed = 1)
dim(pi_grid$lower)  # 2 x 40
# scalar intervals for the grid-weighted mean (ATE-style)
pi_ate  <- split_conformal(F_hat, W, W_new, K = 3, wrapper = mean, seed = 1)
pi_ate$prediction

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conformal.R
\name{summary.metahunt_conformal}
\alias{summary.metahunt_conformal}
\title{Summarise a conformal prediction-interval object}
\usage{
\method{summary}{metahunt_conformal}(object, ...)
}
\arguments{
\item{object}{A \code{"metahunt_conformal"} object from \code{\link[=split_conformal]{split_conformal()}},
\code{\link[=cross_conformal]{cross_conformal()}}, or \code{\link[=conformal_from_fit]{conformal_from_fit()}}.}

\item{...}{Unused; present for S3 generic consistency.}
}
\value{
A list of class \code{"summary.metahunt_conformal"}. In pointwise mode
(no wrapper) the list contains \code{n_targets}, \code{G_grid}, \code{n_cal}, \code{alpha},
\code{method}, \code{mean_interval_width}, \code{frac_finite_quantile},
\code{quantile_summary}, and \code{wrapper}. In scalar mode (wrapper supplied)
the list contains \code{n_targets}, \code{n_cal}, \code{alpha}, \code{method},
\code{mean_interval_width}, \code{quantile}, \code{quantile_finite}, and \code{wrapper}.
}
\description{
Produces a small list of descriptive statistics about a
\code{"metahunt_conformal"} object: interval widths, quantile summaries, and
calibration diagnostics. Returns an object of class
\code{"summary.metahunt_conformal"} with a matching \code{print} method.
}
\examples{
set.seed(1)
G <- 25; m <- 60
x <- seq(0, 1, length.out = G)
basis <- rbind(sin(pi * x), cos(pi * x), x)
W <- data.frame(w1 = rnorm(m))
eta <- cbind(0.6 * W$w1, -0.3 * W$w1, rep(0, m))
pi_true <- exp(eta) / rowSums(exp(eta))
F_hat <- pi_true \%*\% basis + matrix(rnorm(m * G, sd = 0.05), m, G)
W_new <- data.frame(w1 = c(0, 1))
res <- split_conformal(F_hat, W, W_new, K = 3,
                       dfspa_args = list(denoise = FALSE), seed = 1)
summary(res)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight_estimation.R
\name{predict.metahunt_weight_model}
\alias{predict.metahunt_weight_model}
\title{Predict simplex weights for new study-level covariates}
\usage{
\method{predict}{metahunt_weight_model}(object, newdata, ...)
}
\arguments{
\item{object}{A fitted \code{"metahunt_weight_model"} from \code{\link[=fit_weight_model]{fit_weight_model()}}.}

\item{newdata}{A matrix or data frame of new study-level covariates with
the same columns used at fitting.}

\item{...}{Ignored.}
}
\value{
An \code{nrow(newdata)}-by-\code{K} numeric matrix of predicted simplex
weights (component means); rows sum to 1.
}
\description{
Predict simplex weights for new study-level covariates
}
\examples{
\donttest{
set.seed(1)
m <- 40; K <- 3
W <- data.frame(w1 = rnorm(m), w2 = rnorm(m))
eta <- cbind(0.5 * W$w1, -0.3 * W$w2, rep(0, m))
pi_true <- exp(eta) / rowSums(exp(eta))
pi_hat <- pi_true + matrix(rnorm(m * K, sd = 0.01), m, K)
pi_hat <- pmax(pi_hat, 0); pi_hat <- pi_hat / rowSums(pi_hat)
model <- fit_weight_model(pi_hat, W)
predict(model, newdata = data.frame(w1 = c(0, 1), w2 = c(0, -1)))
}

}
